/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_COMMON_REFERENCEVECTOR_HH
#define DUNE_FEM_COMMON_REFERENCEVECTOR_HH

#include <cmath>
#include <complex>
#include <cstddef>
#include <cstdlib>
#include <cstring>
#include <limits>
#include <utility>
#include <vector>

#include <dune/common/exceptions.hh>
#include <dune/common/densevector.hh>
#include <dune/common/nullptr.hh>

namespace Dune
{

  namespace Fem
  {
    template< class K, class A = std::allocator< K* > >
    class DynamicReferenceVector;
  }


  template< class K, class A >
  struct DenseMatVecTraits< Fem::DynamicReferenceVector< K, A > >
  {
    typedef Fem::DynamicReferenceVector< K, A > derived_type;
    typedef std::vector< K*, typename A::template rebind< K* >::other > container_type;

    typedef K value_type;
    typedef typename container_type::size_type size_type;
  };

  template< class K, class A >
  struct FieldTraits< Fem::DynamicReferenceVector< K, A > >
  {
    typedef typename FieldTraits< K >::field_type field_type;
    typedef typename FieldTraits< K >::real_type real_type;
  };

  namespace Fem
  {

    /** @addtogroup DenseMatVec
        @{
     */

    /*! \file
     * \brief This file implements a dense vector with a dynamic size.
     */


    /** \brief Construct a vector with a dynamic size.
     *
     * \tparam K is the field type (use float, double, complex, etc)
     */
    template< class K, class A >
    class DynamicReferenceVector : public DenseVector< DynamicReferenceVector< K, A > >
    {
      typedef DynamicReferenceVector< K, A > This;
      typedef DenseVector< DynamicReferenceVector < K, A > > Base;

      typename DenseMatVecTraits< This > :: container_type data_;
    public:
      typedef typename Base::size_type size_type;
      typedef typename Base::value_type value_type;

      //! Constructor making uninitialized vector
      explicit DynamicReferenceVector ( const A &a = A() )
      : data_( a )
      {}

      //! Constructor making vector with identical coordinates
      explicit DynamicReferenceVector ( size_type n, const A &a = A() )
      : data_( n, nullptr, a )
      {}

      DynamicReferenceVector ( const This &other )
      : data_( other.data_ )
      {}

      DynamicReferenceVector ( This &&other )
      : data_( std::move( other.data_ ) )
      {}

      //! assignment
      using Base::operator=;

      template< class V >
      This & operator= ( const DenseVector< V > &other )
      {
        assert( data_.size() == other.size() );
        std::copy( other.begin(), other.end(), Base::begin() );
        return *this;
      }

      This & operator= ( const This &other )
      {
        assert( data_.size() == other.size() );
        std::copy( other.begin(), other.end(), Base::begin() );
        return *this;
      }

      This & operator= ( This &&other )
      {
        data_ = std::move( other.data_ );
        return *this;
      }

      //==== forward some methods of std::vector
      /** \brief Number of elements for which memory has been allocated.

          capacity() is always greater than or equal to size().
       */
      size_type capacity () const { return data_.capacity(); }

      void resize ( size_type n ) { data_.resize( n, nullptr ); }
      void reserve ( size_type n ) { data_.reserve( n ); }

      //! bind i-th entry to a reference
      void bind ( size_type i, K& u ) { assert( i < data_.size() ); data_[ i ] = &u; }
      void unbind ( size_type i ) { asssert( i < data_.size() ); data_[ i ] = nullptr; }

      //==== make this thing a vector
      size_type vec_size () const { return data_.size(); }

      K &vec_access ( size_type i ) { return *data_[ i ]; }
      const K &vec_access ( size_type i ) const { return *data_[ i ]; }
    };

    /** @} end documentation */

  } // namespace Fem


} // namespace Dune

#endif //#ifndef DUNE_FEM_COMMON_REFERENCEVECTOR_HH
