/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_COMMON_STACKALLOCATOR_HH
#define DUNE_FEM_COMMON_STACKALLOCATOR_HH

#include <cassert>
#include <new>
#include <stack>
#include <utility>

#include <dune/common/nullptr.hh>

namespace Dune
{

  namespace Fem
  {

    // UninitializedObjectStack
    // ------------------------

    struct UninitializedObjectStack
      : public std::stack< void * >
    {
      explicit UninitializedObjectStack ( std::size_t objectSize )
        : objectSize_( objectSize )
      {}

      UninitializedObjectStack ( const UninitializedObjectStack &other )
        : objectSize_( other.objectSize_ )
      {}

      UninitializedObjectStack &operator= ( const UninitializedObjectStack &other )
      {
        if( objectSize_ != other.objectSize_ )
          clear();
        objectSize_ = other.objectSize_;
        return *this;
      }

      ~UninitializedObjectStack () { clear(); }

      void clear ()
      {
        for( ; !empty(); pop() )
          ::operator delete( top() );
      }

      std::size_t objectSize () const { return objectSize_; }

      void resize ( std::size_t newSize ) { clear(); objectSize_ = newSize; }

    private:
      std::size_t objectSize_;
    };



    // StackAllocator
    // --------------

    template< class T, class S = UninitializedObjectStack * >
    struct StackAllocator
    {
      typedef T value_type;

      typedef T *pointer;
      typedef const T*const_pointer;

      typedef T &reference;
      typedef const T &const_reference;

      typedef std::size_t size_type;
      typedef std::ptrdiff_t difference_type;

      template< class U >
      struct rebind { typedef StackAllocator< U, S > other; };

      typedef UninitializedObjectStack Stack;
      typedef S StackPtr;

      explicit StackAllocator ( StackPtr stack ) : stack_( stack ) {}

      template< class U >
      StackAllocator ( const StackAllocator< U, S > &other ) : stack_( other.stack_ ) {}

      template< class U >
      StackAllocator ( StackAllocator< U, S > &&other ) : stack_( std::move( other.stack_ ) ) {}

      StackAllocator ( const StackAllocator &other ) : stack_( other.stack_ ) {}
      StackAllocator ( StackAllocator && other ) : stack_( other.stack_ ) {}

      pointer address ( reference x ) const { return &x; }
      const_pointer address ( const_reference x ) const { return &x; }

      pointer allocate ( size_type n, typename rebind< void >::other::const_pointer hint = nullptr )
      {
        assert( n <= max_size() );
        if( !stack().empty() )
        {
          pointer p = (pointer) stack().top();
          stack().pop();
          return p;
        }
        else
          return (pointer) ::operator new( stack().objectSize() );
      }

      void deallocate ( pointer p, size_type n )
      {
        assert( n <= max_size() );
        stack().push( p );
      }

      size_type max_size () const { return stack().objectSize() / sizeof( T ); }

      template< class... Args >
      void construct ( pointer p, Args &&... args )
      {
        assert( p );
        new( p ) T( std::forward< Args >( args )... );
      }

      void destroy ( pointer p ) { p->~T(); }

    private:
      template< class, class >
      friend struct StackAllocator;

      const Stack &stack () const { return *stack_; }
      Stack &stack () { return *stack_; }

      StackPtr stack_;
    };


    template<class S>
    struct StackAllocator<void, S>
    {
      typedef void value_type;

      typedef void *pointer;
      typedef const void*const_pointer;

      typedef std::size_t size_type;
      typedef std::ptrdiff_t difference_type;

      template< class U >
      struct rebind { typedef StackAllocator< U, S > other; };

      typedef UninitializedObjectStack Stack;
    };

  } // namespace

} //namespace Dune

#endif // #ifndef DUNE_FEM_COMMON_STACKALLOCATOR_HH
