/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_COMMON_SUBVECTOR_HH
#define DUNE_FEM_COMMON_SUBVECTOR_HH

#include <cmath>
#include <complex>
#include <cstddef>
#include <cstdlib>
#include <cstring>
#include <limits>
#include <vector>

#include <dune/common/exceptions.hh>
#include <dune/common/densevector.hh>
#include <dune/common/nullptr.hh>

namespace Dune
{

  namespace Fem
  {
    template< class V >
    class DenseSubVector;
  }


  template< class V >
  struct DenseMatVecTraits< Fem::DenseSubVector< V > >
  {
    typedef Fem::DenseSubVector< V > derived_type;
    typedef V container_type;

    typedef typename V::value_type value_type;
    typedef typename V::size_type size_type;
  };

  template< class V >
  struct FieldTraits< Fem::DenseSubVector< V > >
  {
    typedef typename FieldTraits< typename V::value_type >::field_type field_type;
    typedef typename FieldTraits< typename V::value_type >::real_type real_type;
  };

  namespace Fem
  {

    // DenseSubVector
    // ---------

    template< class V >
    class DenseSubVector : public DenseVector< DenseSubVector< V > >
    {
      typedef DenseSubVector< V > This;
      typedef DenseVector< DenseSubVector < V > > Base;


    public:
      typedef typename Base::size_type size_type;
      typedef typename Base::value_type value_type;

      //! Constructor making uninitialized vector
      explicit DenseSubVector ( V& v, size_type size, size_type offset )
      : v_( v ),
        size_( size ),
        offset_( offset )
      {}

      DenseSubVector ( const This &other )
      : v_( other.v_ ),
        size_( other.size_ ),
        offset_( other.offset_ )
      {}

      using Base::operator=;

      //==== forward some methods of std::vector
      /** \brief Number of elements for which memory has been allocated.

          capacity() is always greater than or equal to size().
       */

      void resize ( size_type ) {}

      //==== make this thing a vector
      size_type vec_size () const { return size_; }

      value_type &vec_access ( size_type i ) { return v_[ i + offset_ ]; }
      const value_type &vec_access ( size_type i ) const { return v_[ i + offset_ ]; }
    private:
      typename DenseMatVecTraits< This > :: container_type &v_;
      const size_type size_, offset_;
    };

    /** @} end documentation */

  } // namespace Fem


} // namespace Dune

#endif //#ifndef DUNE_FEM_COMMON_SUBVECTOR_HH
