/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_PASS_COMMON_TUPLETYPETRAITS_HH
#define DUNE_FEM_PASS_COMMON_TUPLETYPETRAITS_HH

#include <dune/common/tuples.hh>
#include <dune/common/tupleutility.hh>
#include <dune/common/typetraits.hh>

#include "tupleutility.hh"

namespace
{

  // CheckAllElements
  // ----------------

  template< class Tuple,
            template< class > class Predicate,
            int N = std::tuple_size< Tuple >::value
          >
  struct CheckAllElements
  {
    static const bool value = ( Predicate< typename std::tuple_element< N-1, Tuple >::type >::value
                                && CheckAllElements< Tuple, Predicate, (N-1) >::value );
  };

  template< class Tuple,
            template< class > class Predicate
          >
  struct CheckAllElements< Tuple, Predicate, 0 >
  {
    static const bool value = true;
  };

} // namespace



namespace Dune
{

  // SingleTypeTuple
  // ---------------

  /*
   * \brief Make tuple of given length and single type.
   */
  template< class T,
            int size,
            class Seed = std::tuple<>,
            int index = 0
          >
  struct SingleTypeTuple
  {
    typedef typename SingleTypeTuple< T, size, typename Dune::PushBackTuple< Seed, T >::type, (index+1) >::Type Type;
  };

  template< class T,
            int size,
            class Seed
          >
  struct SingleTypeTuple< T, size, Seed, size >
  {
    typedef Seed Type;
  };



  // EmptyTuple
  // ----------

  /*
   * \brief a tuple consisting of Dune::Empty
   */
  template< int size >
  struct EmptyTuple
  : public SingleTypeTuple< Dune::Empty, size >
  {};



  // TupleTypeTraits
  // ---------------

  /*
   * \brief Mimicks Dune::TypeTraits (see dune/common/typetraits.hh) for tuples.
   */
  template< class Tuple >
  class TupleTypeTraits
  {
    template< class T >
    struct IsPointer
    {
      static const bool value = Dune::TypeTraits< T >::isPointer;
    };

    template< class T >
    struct IsReference
    {
      static const bool value = Dune::TypeTraits< T >::isReference;
    };

    template< class T >
    struct PointeeTypeEvaluator
    {
      typedef typename Dune::TypeTraits< T >::PointeeType Type;
    };

    template< class T >
    struct ReferredTypeEvaluator
    {
      typedef typename Dune::TypeTraits< T >::ReferredType Type;
    };

  public:
    static const bool isPointerTuple = CheckAllElements< Tuple, IsPointer >::value;

    typedef typename Dune::conditional< isPointerTuple,
                                        typename Dune::ForEachType< PointeeTypeEvaluator, Tuple >::Type,
                                        EmptyTuple< std::tuple_size< Tuple >::value >
                                      >::type PointeeTupleType;

    static const bool isReferenceTuple = CheckAllElements< Tuple, IsReference >::value;

    typedef typename Dune::ForEachType< ReferredTypeEvaluator, Tuple >::Type ReferredTupleType;
  };



  // PointerTuple
  // ------------

  /**
   * \brief Convert a tuple to a tuple of pointer types.
   *
   * \tparam  Tuple  none of the types in Tuple should be of pointer type
   */
  template< class Tuple >
  class PointerTuple
  {
    template< class T >
    struct PointerEvaluator
    {
      typedef typename Dune::TypeTraits< T >::PointeeType * Type;
    };

  public:
    typedef typename Dune::ForEachType< PointerEvaluator, Tuple >::Type Type;
  };



  // ReferenceTuple
  // --------------

  /**
   * \brief Convert a tuple to a tuple of references.
   *
   * \tparam  Tuple  tuple to convert
   */
  template< class Tuple >
  class ReferenceTuple
  {
    template< class T >
    struct ReferenceEvaluator
    {
      typedef typename Dune::TypeTraits< T >::ReferredType & Type;
    };

  public:
    typedef typename Dune::ForEachType< ReferenceEvaluator, Tuple >::Type Type;
  };



  // ConstTuple
  // ----------

  /**
   * \brief Add const qualifier to all tuple elements.
   *
   * \tparam  Tuple  tuple to convert
   */
  template< class Tuple >
  class ConstTuple
  {
    template< class T >
    struct ConstEvaluator
    {
      typedef typename Dune::ConstantVolatileTraits< T >::ConstType Type;
    };

  public:
    typedef typename Dune::ForEachType< ConstEvaluator, Tuple >::Type Type;
  };



  // RemoveConstTuple
  // ----------------

  /**
   * \brief Remove const qualifiers from tuple.
   */
  template< class Tuple >
  class RemoveConstTuple
  {
    template< class T >
    struct RemoveConstEvaluator
    {
      typedef typename Dune::ConstantVolatileTraits< T >::UnqualifiedType Type;
    };

  public:
    typedef typename Dune::ForEachType< RemoveConstEvaluator, Tuple >::Type Type;
  };



  // tuple_remove_const
  // ------------------

  /*
   * \brief Convenience structure mimicking Dune::remove_const.
   */
  template< class Tuple >
  struct tuple_remove_const
  {
    typedef typename RemoveConstTuple< Tuple >::Type type;
  };



  // ValidPointerTupleCheck
  // ----------------------

  /**
   * \brief Check whether a pointer tuple can be dereferenced.
   *
   * \tparam  Tuple  tuple of pointer types
   */
  template< class Tuple >
  class ValidPointerTupleCheck
  {
    static_assert( TupleTypeTraits< Tuple >::isPointerTuple,
                   "Can not check non-pointer tuple." );

    struct ValidPointer
    {
      ValidPointer () : v_( true ) {}

      template< class Ptr >
      void visit ( Ptr &ptr )
      {
        v_ &= bool( ptr );
      }

      operator bool() const { return v_; }

    private:
      bool v_;
    };

  public:
    static bool apply ( const Tuple &tuple )
    {
      Dune::ForEachValue< Tuple > forEach( tuple );
      ValidPointer check;
      forEach.apply( check );
      return check;
    }
  };



  // DereferenceTuple
  // ----------------

  /**
   * \brief Dereference pointer tuple.
   */
  template< class Tuple,
            class Seed = std::tuple<>,
            int index = 0,
            int size = std::tuple_size< Tuple >::value
          >
  class DereferenceTuple
  {
    template< class, class, int, int > friend class DereferenceTuple;

    typedef typename Dune::TypeTraits< typename std::tuple_element< index, Tuple >::type >::PointeeType & AppendType;
    typedef typename Dune::PushBackTuple< Seed, AppendType >::type AccumulatedType;

    typedef DereferenceTuple< Tuple, AccumulatedType, (index+1), size > NextType;

  public:
    typedef typename Dune::ReferenceTuple<
        typename Dune::TupleTypeTraits< Tuple >::PointeeTupleType
      >::Type Type;

    static Type apply ( Tuple &tuple )
    {
      Seed seed;
      return append( tuple, seed );
    }

  private:
    static Type append ( Tuple &tuple, Seed &seed )
    {
      typename std::tuple_element< index, Tuple >::type pointer = std::get< index >( tuple );
      AppendType append = *pointer;
      AccumulatedType next = Dune::tuple_push_back< AppendType >( seed, append );
      return NextType::append( tuple, next );
    }
  };

  template< class Tuple,
            class Seed,
            int size
          >
  class DereferenceTuple< Tuple, Seed, size, size >
  {
    template< class, class, int, int > friend class DereferenceTuple;

  public:
    typedef typename Dune::ReferenceTuple<
        typename Dune::TupleTypeTraits< Tuple >::PointeeTupleType
      >::Type Type;

    static_assert( (std::is_same< Seed, Type >::value), "Failed to dereference pointer tuple." );

    static Type apply ( Tuple & )
    {
      return Type();
    }

  private:
    static Seed append ( Tuple &tuple, Seed &seed ) { return seed; }
  };

} // namespace Dune

#endif // #ifndef DUNE_FEM_PASS_COMMON_TUPLETYPETRAITS_HH
