/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_RANGEGENERATORS_HH
#define DUNE_FEM_RANGEGENERATORS_HH

#include <dune/common/iteratorrange.hh>

namespace Dune
{
  namespace Fem
  {
    // only make the implementations available on new compilers
    #if HAVE_RANGE_BASED_FOR
    /**
     * \brief Iterator ranges for entities and DOFs to support iteration with range-based for loops.
     *
     * <h2>Range-based for loop</h2>
     *
     * Range-based for loops are availabe in GCC 4.6+, Clang 3.2+ and Intel ICC 13+.
     *
     * For further details, see e.g. http://en.cppreference.com/w/cpp/language/range-for.
     *
     * <h2>Entities</h2>
     *
     * Assuming you have a DiscreteFunction `df`, you can iterate over the grid entities like this:
     *
     * \code
     * // iterate over entities
     * for (auto&& entity : entities(df))
     * {
     *   // do stuff
     * }
     * \endcode
     *
     * <h2>DOFs</h2>
     *
     * If you want to iterate over the DOFs of a DiscreteFunction `df`, you can use the function dofs()
     * to obtain a range that is suitable for a range-based for loop. As before you can do like this:
     *
     * \code
     * for (auto&& dof : dofs(df))
     * {
     *   // do stuff
     * }
     * \endcode
     *
     */

    //! Iterates over all entities.
    /**
     * This functions returns an object representing the range of entities. The main purpose of this function
     * is to enable iteration over those entities by means of a range-based for loop.
     *
     * \since      GCC 4.6
     * \param df   a DiscreteFunction.
     * \returns    an unspecified object that is guaranteed to fulfil the interface
     *             of IteratorRange and that can be iterated over using a range-based
     *             for loop.
     */
    template<typename DF>
    inline IteratorRange<typename DF::DiscreteFunctionSpaceType::IteratorType> entities(const DF& df)
    {
      typedef IteratorRange<typename DF::DiscreteFunctionSpaceType::IteratorType> ReturnType;
      return ReturnType(df.space().begin(),df.space().end());
    }
    //! \}

    //! Iterates over all DOFs.
    /**
     * This functions returns an object representing the range of DOFs. The main purpose of this function
     * is to enable iteration over those DOFs by means of a range-based for loop.
     *
     * \since      GCC 4.6
     * \param df   a DiscreteFunction.
     * \returns    an unspecified object that is guaranteed to fulfil the interface
     *             of IteratorRange and that can be iterated over using a range-based
     *             for loop.
     */
    template<typename DF>
    inline IteratorRange<typename DF::DofIteratorType> dofs(DF& df)
    {
      typedef IteratorRange<typename DF::DofIteratorType> ReturnType;
      return ReturnType(df.dbegin(),df.dend());
    }
    //! \}

    //! Iterates over all DOFs.
    /**
     * This functions returns an object representing the range of DOFs. The main purpose of this function
     * is to enable iteration over those DOFs by means of a range-based for loop.
     *
     * \since      GCC 4.6
     * \param df   a DiscreteFunction.
     * \returns    an unspecified object that is guaranteed to fulfil the interface
     *             of IteratorRange and that can be iterated over using a range-based
     *             for loop.
     */
    template<typename DF>
    inline IteratorRange<typename DF::ConstDofIteratorType> dofs(const DF& df)
    {
      typedef IteratorRange<typename DF::ConstDofIteratorType> ReturnType;
      return ReturnType(df.dbegin(),df.dend());
    }
    //! \}


    #endif // HAVE_RANGE_BASED_FOR

  } // end namespace Fem

} // end namespace Dune

#endif // DUNE_FEM_RANGEGENERATORS_HH
