/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_FUNCTION_LOCALFUNCTION_AVERAGE_HH
#define DUNE_FEM_FUNCTION_LOCALFUNCTION_AVERAGE_HH

#include <dune/fem/gridpart/common/capabilities.hh>
#include <dune/fem/quadrature/cachingquadrature.hh>
#include <dune/fem/space/common/discretefunctionspace.hh>
#include <dune/fem/space/discontinuousgalerkin/declaration.hh>
#include <dune/fem/space/finitevolume/declaration.hh>

namespace Dune
{

  namespace Fem
  {

    // Internal forward declaration
    // ----------------------------

    template< class LocalFunction, class GridPart >
    class LocalAverage;



    // LocalAverageHelper
    // ------------------

    struct LocalAverageHelper
    {
      template< class LocalFunction, class Quadrature >
      static void applyQuadrature ( const LocalFunction &localFunction,
                                    const typename LocalFunction::EntityType::Geometry &geometry,
                                    const Quadrature &quadrature,
                                    typename LocalFunction::RangeType &value )
      {
        typedef typename LocalFunction::RangeType RangeType;
        typedef typename RangeType::value_type RangeFieldType;

        value = RangeType( 0 );
        RangeFieldType volume( 0 );
        const int nop = quadrature.nop();
        for( int qp = 0; qp < nop; ++qp )
        {
          RangeType tmp;
          localFunction.evaluate( quadrature[ qp ], tmp );
          RangeFieldType weight = quadrature.weight( qp )*geometry.integrationElement( quadrature.point( qp ) );
          volume += weight;
          value.axpy( weight, tmp );
        }
        value /= volume;
      }
    };



    // LocalAverageImpl
    // ----------------

    template< class LocalFunction, class GridPart, class DiscreteFunctionSpace >
    struct LocalAverageImpl
    {
      static void apply ( const LocalFunction &localFunction, typename LocalFunction::RangeType &average )
      {
        // get entity
        typedef typename LocalFunction::EntityType EntityType;
        const EntityType &entity = localFunction.entity();

        // create quadrature
        typedef CachingQuadrature< GridPart, EntityType::codimension > QuadratureType;
        QuadratureType quadrature( entity, localFunction.order() );
        LocalAverageHelper::applyQuadrature( localFunction, entity.geometry(), quadrature, average );
      }
    };



    // LocalAverageImpl for DiscontinuousGalerkinSpace
    // -----------------------------------------------

    template< class LocalFunction, class GridPart, class FunctionSpace, int order, template< class > class Storage >
    class LocalAverageImpl< LocalFunction, GridPart, DiscontinuousGalerkinSpace< FunctionSpace, GridPart, order, Storage > >
    {
      static const bool cartesian = GridPartCapabilities::isCartesian< GridPart >::v;

      static void applyAffine ( const LocalFunction &localFunction, typename LocalFunction::RangeType &average )
      {
        // otherwise use direct computation
        const int dimRange = LocalFunction::dimRange;
        for( int i = 0; i < dimRange; ++i )
          average[ i ] = localFunction[ i ];

        const typename LocalFunction::BasisFunctionSetType &basisFunctionSet = localFunction.basisFunctionSet();
        average /= std::sqrt( basisFunctionSet.referenceElement().volume() );
      }

    public:
      static void apply ( const LocalFunction &localFunction, typename LocalFunction::RangeType &average )
      {
        if( cartesian )
          return applyAffine( localFunction, average );

        // get entity and geometry
        typedef typename LocalFunction::EntityType EntityType;
        const EntityType &entity = localFunction.entity();
        const typename EntityType::Geometry geometry = entity.geometry();

        if( geometry.affine() )
          return applyAffine( localFunction, average );

        // for not affine elements, use quadrature
        typedef CachingQuadrature< GridPart, EntityType::codimension > QuadratureType;
        QuadratureType quadrature( entity, localFunction.order() );
        LocalAverageHelper::applyQuadrature( localFunction, geometry, quadrature, average );
      }
    };



    // LocalAverageImpl for FiniteVolumeSpace
    // --------------------------------------

    template< class LocalFunction, class GridPart, class FunctionSpace, int codim, template< class > class Storage >
    class LocalAverageImpl< LocalFunction, GridPart, FiniteVolumeSpace< FunctionSpace, GridPart, codim, Storage > >
    {
    public:
      static void apply ( const LocalFunction &localFunction, typename LocalFunction::RangeType &average )
      {
        localFunction.evaluate( typename LocalFunction::DomainType( 0 ), average );
      }
    };



    // LocalAverage
    // ------------

    template< class LocalFunction, class GridPart >
    class LocalAverage
    {
      typedef typename ExportsDiscreteFunctionSpaceType< LocalFunction >::Type DiscreteFunctionSpaceType;
      typedef LocalAverageImpl< LocalFunction, GridPart, DiscreteFunctionSpaceType > Impl;

    public:
      static void apply ( const LocalFunction &localFunction, typename LocalFunction::RangeType &average )
      {
        Impl::apply( localFunction, average );
      }

      void operator() ( const LocalFunction &localFunction, typename LocalFunction::RangeType &average ) const
      {
        Impl::apply( localFunction, average );
      }
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_FUNCTION_LOCALFUNCTION_AVERAGE_HH
