/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_FUNCTION_LOCALFUNCTION_LOCALFUNCTIONSET_HH
#define DUNE_FEM_FUNCTION_LOCALFUNCTION_LOCALFUNCTIONSET_HH

#include <cassert>
#include <cstddef>

#include <dune/common/nullptr.hh>

#include <dune/fem/space/common/functionspace.hh>

namespace Dune
{

  namespace Fem
  {

    // LocalFunctionSet
    // ----------------

    /** \class LocalFunctionSet
     *
     *  \brief Local basis functions
     *
     *  This class documents the local basis functions interface.
     *
     *  \tparam  FunctionSpace  function space
     */
    template< class Entity, class Range >
    struct LocalFunctionSet
    {
      //! \brief entity type
      typedef Entity EntityType;

      //! \brief function space type
      typedef FunctionSpace< typename Entity::Geometry::ctype, typename Range::value_type,
                             Entity::Geometry::coorddimension, Range::dimension > FunctionSpaceType;

      //! \brief domain type
      typedef typename FunctionSpaceType::DomainType DomainType;
      //! \brief range type
      typedef typename FunctionSpaceType::RangeType RangeType;
      //! \brief jacobian range type
      typedef typename FunctionSpaceType::JacobianRangeType JacobianRangeType;
      //! \brief hessian range type
      typedef typename FunctionSpaceType::HessianRangeType HessianRangeType;

      //! \brief return order of basis functions
      int order () const;

      //! \brief return entity
      const EntityType &entity () const;

      //! \brief return number of basis functions
      std::size_t size () const;

      /**
       * \brief evalute each basis function
       *
       *  \param[in]  x        local coordinate or quadrature point
       *  \param[in]  functor  functor call for evaluating each basis function
       *
       *  The functor has to be a copyable object satisfying the following
       *  interface:
\code
  struct Functor
  {
    template< class Value >
    void operator() ( const int basisFunction, const Value &value );
  };
\endcode
       */
      template< class Point, class Functor >
      void evaluateEach ( const Point &x, Functor functor ) const;

      /**
       * \brief evalute jacobian of each basis function
       *
       *  \param[in]  x        local coordinate or quadrature point
       *  \param[in]  functor  functor call for evaluating the jacobian of each basis function
       *
       *  The functor has to be a copyable object satisfying the following
       *  interface:
\code
  struct Functor
  {
    template< class Jacobian >
    void operator() ( const int basisFunction, const Jacobian &jacobian );
  };
\endcode
       */
      template< class Point, class Functor >
      void jacobianEach ( const Point &x, Functor functor ) const;

      /**
       * \brief evalute hessian of each basis function
       *
       *  \param[in]  x        local coordinate or quadrature point
       *  \param[in]  functor  functor call for evaluating the hessian of each basis function
       *
       *  The functor has to be a copyable object satisfying the following
       *  interface:
\code
  struct Functor
  {
    template< class Hessian >
    void operator() ( const int basisFunction, const Hessian &hessian );
  };
\endcode
       */
      template< class Point, class Functor >
      void hessianEach ( const Point &x, Functor functor ) const;
    };



    // LocalFunctionSetProxy
    // ---------------------

    /** \class LocalFunctionSetProxy
     *
     *  \brief Proxy for a LocalBasisFunctionSet
     *
     *  \tparam  LocalFunctionSet  local function set
     */
    template< class LocalFunctionSet >
    class LocalFunctionSetProxy
    {
    public:
      typedef LocalFunctionSet ImplementationType;
      const ImplementationType &impl () const
      {
        assert( localFunctionSet_ );
        return *localFunctionSet_;
      }

      typedef typename LocalFunctionSet::EntityType EntityType;
      typedef typename LocalFunctionSet::FunctionSpaceType FunctionSpaceType;

      typedef typename LocalFunctionSet::DomainType DomainType;
      typedef typename LocalFunctionSet::RangeType RangeType;
      typedef typename LocalFunctionSet::JacobianRangeType JacobianRangeType;
      typedef typename LocalFunctionSet::HessianRangeType HessianRangeType;

      LocalFunctionSetProxy () : localFunctionSet_( nullptr ) {}

      LocalFunctionSetProxy ( const LocalFunctionSet *localFunctionSet )
      : localFunctionSet_( localFunctionSet )
      {}

      int order () const { return impl().order(); }

      const EntityType &entity () const { return impl().entity(); }

      std::size_t size () const { return impl().size(); }

      template< class Point, class Functor >
      void evaluateEach ( const Point &x, Functor functor ) const
      {
        impl().evaluateEach( x, functor );
      }

      template< class Point, class Functor >
      void jacobianEach ( const Point &x, Functor functor ) const
      {
        impl().jacobianEach( x, functor );
      }

      template< class Point, class Functor >
      void hessianEach ( const Point &x, Functor functor ) const
      {
        impl().hessianEach( x, functor );
      }

    private:
      const LocalFunctionSet *localFunctionSet_;
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_FUNCTION_LOCALFUNCTION_LOCALFUNCTIONSET_HH
