/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_FUNCTION_LOCALFUNCTION_LOCALFUNCTIONSETADAPTER_HH
#define DUNE_FEM_FUNCTION_LOCALFUNCTION_LOCALFUNCTIONSETADAPTER_HH

#include <cassert>
#include <cstddef>

#include <dune/fem/common/coordinate.hh>

namespace Dune
{

  namespace Fem
  {

    // LocalFunctionSetAdapter
    // -----------------------

    /** \class LocalFunctionSetAdapter
     *
     *  \brief convert (global) function set to local function set
     *
     *  \tparam  Entity       entity type
     *  \tparam  FunctionSet  implementation of FunctionSet
     */
    template< class Entity, class FunctionSet >
    struct LocalFunctionSetAdapter
    {
      //! \brief entity type
      typedef Entity EntityType;
      //! \brief function set type
      typedef FunctionSet FunctionSetType;

      //! \brief function space type
      typedef typename FunctionSet::FunctionSpaceType FunctionSpaceType;

      //! \brief domain type
      typedef typename FunctionSpaceType::DomainType DomainType;
      //! \brief range type
      typedef typename FunctionSpaceType::RangeType RangeType;
      //! \brief jacobian range type
      typedef typename FunctionSpaceType::JacobianRangeType JacobianRangeType;
      //! \brief hessian range type
      typedef typename FunctionSpaceType::HessianRangeType HessianRangeType;

      explicit LocalFunctionSetAdapter ( const Entity &entity,
                                         const FunctionSet &functionSet = FunctionSet() )
      : entity_( &entity ),
        functionSet_( functionSet )
      {}

      /** \copydoc Dune::Fem::LocalFunctionSet::order() const */
      int order () const { return functionSet_.order(); }

      /** \copydoc Dune::Fem::LocalFunctionSet::entity() const */
      const EntityType &entity () const
      {
        assert( entity_ );
        return *entity_;
      }

      /** \copydoc Dune::Fem::LocalFunctionSet::size() const */
      std::size_t size () const { return functionSet().size(); }

      /** \copydoc Dune::Fem::LocalFunctionSet::evaluateEach() const */
      template< class Point, class Functor >
      void evaluateEach ( const Point &x, Functor functor ) const
      {
        DomainType y = geometry().global( coordinate( x ) );
        functionSet().evaluateEach( y, functor );
      }

      /** \copydoc Dune::Fem::LocalFunctionSet::jacobianEach() const */
      template< class Point, class Functor >
      void jacobianEach ( const Point &x, Functor functor ) const
      {
        DomainType y = geometry().global( coordinate( x ) );
        functionSet().jacobianEach( y, functor );
      }

      /** \copydoc Dune::Fem::LocalFunctionSet::hessianEach() const */
      template< class Point, class Functor >
      void hessianEach ( const Point &x, Functor functor ) const
      {
        DomainType y = geometry().global( coordinate( x ) );
        functionSet().hessianEach( y, functor );
      }


      // Non-interface methods
      // ---------------------

      //! \brief return function set
      const FunctionSetType functionSet () const { return functionSet_; }

    private:
      typename EntityType::Geometry geometry () const { return entity().geometry(); }

      const EntityType *entity_;
      FunctionSetType functionSet_;
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_FUNCTION_LOCALFUNCTION_LOCALFUNCTIONSETADAPTER_HH
