/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_FUNCTION_LOCALFUNCTION_MUTABLE_HH
#define DUNE_FEM_FUNCTION_LOCALFUNCTION_MUTABLE_HH

//-s system includes
#include <cassert>
#include <utility>

//- Dune includes
#include <dune/fem/function/localfunction/localfunction.hh>

namespace Dune
{

  namespace Fem
  {

    template< class >
    struct DiscreteFunctionTraits;

    template< class >
    class ConstLocalFunction;

    //**************************************************************************
    //
    //  --MutableLocalFunction
    //
    //**************************************************************************
    //! Manages the getting and deleting of local function pointers and
    //! acts like a local functions
    template < class DiscreteFunction >
    class MutableLocalFunction
    : public LocalFunction< typename DiscreteFunctionTraits< DiscreteFunction > :: DiscreteFunctionSpaceType ::  BasisFunctionSetType,
      typename DiscreteFunctionTraits< DiscreteFunction > :: LocalDofVectorType >
    {
      typedef MutableLocalFunction< DiscreteFunction > ThisType;
      typedef LocalFunction< typename DiscreteFunctionTraits< DiscreteFunction > :: DiscreteFunctionSpaceType ::  BasisFunctionSetType,
      typename DiscreteFunctionTraits< DiscreteFunction > :: LocalDofVectorType > BaseType;

    public:
      //! type of DiscreteFunction
      typedef DiscreteFunction DiscreteFunctionType;

      //! type of the entity, the local function lives on is given by the space
      typedef typename BaseType::EntityType EntityType;

      //! type of local Dof vector object
      typedef typename BaseType::LocalDofVectorType LocalDofVectorType;

      //! type of BasisFunctionSet
      typedef typename BaseType::BasisFunctionSetType BasisFunctionSetType;

      //! cast from ConstLocalFunction
      MutableLocalFunction( const ConstLocalFunction< DiscreteFunctionType > &constLocalFunction ) DUNE_DEPRECATED
      : BaseType( constLocalFunction.basisFunctionSet(), LocalDofVectorType( constLocalFunction.discreteFunction_.localDofVectorAllocator() ) ),
        discreteFunction_( &const_cast< DiscreteFunctionType& >( constLocalFunction.discreteFunction() ) )
      {
        discreteFunction().getLocalDofs( constLocalFunction.entity(), localDofVector() );
      }

      //! Constructor creating empty local function from given discrete function
      explicit MutableLocalFunction ( DiscreteFunctionType &discreteFunction )
      : BaseType( LocalDofVectorType( discreteFunction.localDofVectorAllocator() ) ),
        discreteFunction_( &discreteFunction )
      {}

      //! Constructor creating empty local function from given discrete function
      explicit MutableLocalFunction ( const DiscreteFunctionType &discreteFunction )
      : BaseType( LocalDofVectorType( discreteFunction.localDofVectorAllocator() ) ),
        discreteFunction_( &const_cast<DiscreteFunctionType &>( discreteFunction ) )
      {}

      //! Constructor creating local function from given discrete function and entity, not empty
      explicit MutableLocalFunction ( DiscreteFunctionType &discreteFunction, const EntityType &entity )
      : BaseType( discreteFunction.space().basisFunctionSet( entity ), LocalDofVectorType( discreteFunction.localDofVectorAllocator() ) ),
        discreteFunction_( &discreteFunction )
      {
        discreteFunction.getLocalDofs( entity, localDofVector() );
      }

      //! Constructor creating local function from given discrete function and entity, not empty
      explicit MutableLocalFunction ( const DiscreteFunctionType &dFunction, const EntityType &entity )
      : BaseType( dFunction.space().basisFunctionSet( entity ), LocalDofVectorType( dFunction.localDofVectorAllocator() ) ),
        discreteFunction_( &const_cast<DiscreteFunctionType &>( dFunction ) )
      {
        discreteFunction().getLocalDofs( entity, localDofVector() );
      }

      //! copy constructor
      MutableLocalFunction ( const ThisType &other )
      : BaseType( static_cast< const BaseType& > ( other ) ),
        discreteFunction_( other.discreteFunction_ )
      {}

      //! move constructor
      MutableLocalFunction ( ThisType &&other )
      : BaseType( static_cast< BaseType&& > ( other ) ),
        discreteFunction_( other.discreteFunction_ )
      {}

      // prohibit assignment
      ThisType &operator= ( const ThisType & ) = delete;
      ThisType &operator= ( ThisType && ) = delete;

      using BaseType::localDofVector;

      void init ( const EntityType &entity )
      {
        BaseType::init( discreteFunction().space().basisFunctionSet( entity ) );
        discreteFunction().getLocalDofs( entity, localDofVector() );
      }

      const DiscreteFunctionType &discreteFunction () const { return *discreteFunction_; }
      DiscreteFunctionType &discreteFunction () { return *discreteFunction_; }

    private:
      DiscreteFunctionType *discreteFunction_;
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_FUNCTION_LOCALFUNCTION_MUTABLE_HH
