/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_FUNCTION_LOCALFUNCTION_TEMPORARY_HH
#define DUNE_FEM_FUNCTION_LOCALFUNCTION_TEMPORARY_HH

#include <dune/common/dynvector.hh>
#include <dune/fem/function/localfunction/localfunction.hh>


namespace Dune
{

  namespace Fem
  {

    // internal forward declarations
    // -----------------------------

    template< class DiscreteFunctionSpace, class DoFVector >
    class BasicTemporaryLocalFunction;

    template< class DiscreteFunctionSpace, class Dof >
    class TemporaryLocalFunction;


    /** \ingroup LocalFunction
        \class BasicTemporaryLocalFunction
        \brief A temporary function carrying values for one entity

        A BasicTemporaryLocalFunction is a LocalFunction which is not associated with
        any DiscreteFunction. It can be used when generating discrete functions
        to temporarily store values for one entity.

        \param DiscreteFunctionSpace type of the discrete function space, the
                                     local function shall belong to
        \param DofVector type of vector the degrees of freedom for an entity are stored in
     */
    template< class DiscreteFunctionSpace, class DofVector >
    class BasicTemporaryLocalFunction
    : public LocalFunction < typename DiscreteFunctionSpace :: BasisFunctionSetType, DofVector >
    {
      typedef BasicTemporaryLocalFunction< DiscreteFunctionSpace, DofVector > ThisType;
      typedef LocalFunction < typename DiscreteFunctionSpace :: BasisFunctionSetType, DofVector > BaseType;

    public:
      //! type of the discrete function space
      typedef DiscreteFunctionSpace  DiscreteFunctionSpaceType;

      //! type of Entity
      typedef typename BaseType :: EntityType EntityType;

      //! type of BasisFunctionSet
      typedef typename BaseType :: BasisFunctionSetType BasisFunctionSetType;

      //! type of LocalDofVector
      typedef typename BaseType :: LocalDofVectorType LocalDofVectorType;

      /* \copydoc Dune::Fem::LocalFunction :: localDofVector */
      using BaseType::localDofVector;

      /** \brief constructor creating a local function without binding it to an
                 entity

          Creates the local function without initializing the fields depending on
          the current entity.

          \note Before using the local function it must be initilized by
          \code
          localFunction.init( entity );
          \endcode

          \param[in] dfSpace discrete function space the local function shall
                             belong to
       */
      explicit BasicTemporaryLocalFunction ( const DiscreteFunctionSpaceType &dfSpace,
                                             const LocalDofVectorType &dofVector = LocalDofVectorType() )
      : BaseType( dofVector ),
        dfSpace_( dfSpace )
      {
        localDofVector().reserve( DiscreteFunctionSpaceType::localBlockSize * dfSpace_.blockMapper().maxNumDofs() );
      }

      /** \brief constructor creating a local function and binding it to an
                 entity

          Creates the local function and initilizes the fields depending on the
          current entity. It is not necessary, though allowed, to call init
          before using the discrete function.

          \note The degrees of freedom are not initialized by this function.

          \param[in] dfSpace discrete function space the local function shall
                             belong to
          \param[in] entity  entity for initialize the local function to
       */
      BasicTemporaryLocalFunction ( const DiscreteFunctionSpaceType &dfSpace, const EntityType &entity,
                                    const LocalDofVectorType &dofVector = LocalDofVectorType() )
      : BaseType( dofVector ),
        dfSpace_( dfSpace )
      {
        localDofVector().reserve( DiscreteFunctionSpaceType::localBlockSize * dfSpace_.blockMapper().maxNumDofs() );
        init( entity );
      }

      /** \brief initialize the local function for an entity
       *
          Binds the local function to an entity.

          \note A local function must be initialized to an entity before it can
                be used.

          \note This function can be called multiple times to use the local
                function for more than one entity.

          \param[in] entity entity to bind the local function to
       */
      void init ( const EntityType &entity )
      {
        BaseType::init( dfSpace_.basisFunctionSet( entity ) );
      }

    protected:
      const DiscreteFunctionSpaceType &dfSpace_;
    };


    /** \ingroup LocalFunction
        \class TemporaryLocalFunction
        \brief A temporary function carrying values for one entity

        A TemporaryLocalFunction is a LocalFunction which is not associated with
        any DiscreteFunction. It can be used when generating discrete functions
        to temporarily store values for one entity.

        \note Local DoF numbers correspond directly to array indices. Hence it
        may be more cache efficient to generate a TemporaryLocalFunction and then
        do only one update step on the discrete function's LocalFunction.

        \param DiscreteFunctionSpaceImp type of the discrete function space, the
                                        local function shall belong to
     */
    template< class DiscreteFunctionSpace, class Dof = typename DiscreteFunctionSpace::RangeFieldType >
    class TemporaryLocalFunction
    : public BasicTemporaryLocalFunction< DiscreteFunctionSpace, Dune::DynamicVector< Dof > >
    {
      typedef TemporaryLocalFunction< DiscreteFunctionSpace, Dof > ThisType;
      typedef BasicTemporaryLocalFunction< DiscreteFunctionSpace, Dune::DynamicVector< Dof > > BaseType;

    public:
      //! type of Entity
      typedef typename BaseType :: EntityType EntityType;

      //! type of the discrete function space
      typedef typename BaseType :: DiscreteFunctionSpaceType DiscreteFunctionSpaceType;

      /** \brief constructor creating a local function without binding it to an
                 entity

          Creates the local function without initializing the fields depending on
          the current entity.

          \note Before using the local function it must be initilized by
          \code
          localFunction.init( entity );
          \endcode

          \param[in] dfSpace discrete function space the local function shall
                             belong to
       */
      explicit TemporaryLocalFunction ( const DiscreteFunctionSpaceType &dfSpace )
      : BaseType( dfSpace ) {}

      /** \brief constructor creating a local function and binding it to an
                 entity

          Creates the local function and initilizes the fields depending on the
          current entity. It is not necessary, though allowed, to call init
          before using the discrete function.

          \note The degrees of freedom are not initialized by this function.

          \param[in] dfSpace discrete function space the local function shall
                             belong to
          \param[in] entity  entity for initialize the local function to
       */
      TemporaryLocalFunction ( const DiscreteFunctionSpaceType &dfSpace, const EntityType &entity )
      : BaseType( dfSpace, entity ){}
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_FUNCTION_LOCALFUNCTION_TEMPORARY_HH
