/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_VECTORFUNCTION_HH
#define DUNE_FEM_VECTORFUNCTION_HH

#include <dune/common/typetraits.hh>

#include <dune/fem/common/referencevector.hh>
#include <dune/fem/common/stackallocator.hh>
#include <dune/fem/function/common/discretefunction.hh>
#include <dune/fem/function/localfunction/mutable.hh>
#include <dune/fem/storage/envelope.hh>
#include <dune/fem/storage/vector.hh>

#include <dune/fem/function/blockvectors/defaultblockvectors.hh>

namespace Dune
{

  namespace Fem
  {

    // Internal Forward Declarations
    // -----------------------------

    template < class DiscreteFunctionSpace, class Vector >
    class VectorDiscreteFunction;

    // VectorDiscreteFunctionTraits
    // ----------------------------

    template< typename DiscreteFunctionSpace, typename Vector >
    struct DiscreteFunctionTraits< VectorDiscreteFunction< DiscreteFunctionSpace, Vector > >
      : public DefaultDiscreteFunctionTraits< DiscreteFunctionSpace,
                  SimpleBlockVector< Vector, DiscreteFunctionSpace::localBlockSize > >
    {
      typedef VectorDiscreteFunction< DiscreteFunctionSpace, Vector >  DiscreteFunctionType;
      typedef MutableLocalFunction< DiscreteFunctionType > LocalFunctionType;
    };


    // VectorDiscreteFunction
    // ----------------------

    template < class DiscreteFunctionSpace, class Vector >
    class VectorDiscreteFunction
    : public DiscreteFunctionDefault<
          VectorDiscreteFunction< DiscreteFunctionSpace, Vector > >
    {
      typedef VectorDiscreteFunction< DiscreteFunctionSpace, Vector > ThisType;
      typedef DiscreteFunctionDefault< ThisType > BaseType;

    public:
      typedef Vector VectorType;
      typedef typename BaseType :: DiscreteFunctionSpaceType  DiscreteFunctionSpaceType;
      typedef typename BaseType :: DofVectorType              DofVectorType;
      typedef typename DofVectorType :: DofContainerType      DofContainerType;
      typedef typename BaseType :: DofType                    DofType;

      using BaseType::assign;

      VectorDiscreteFunction( const std::string &name,
                              const DiscreteFunctionSpaceType &space,
                              VectorType& vector )
        : BaseType( name, space ),
          vec_(),
          dofVector_( vector )
      {
      }

      VectorDiscreteFunction( const VectorDiscreteFunction& other )
        : BaseType( "copy of " + other.name(), other.space() ),
          vec_(),
          dofVector_( allocateDofVector( other.space() ) )
      {
        assign( other );
      }

      DofVectorType& dofVector() { return dofVector_; }
      const DofVectorType& dofVector() const { return dofVector_; }

    protected:
      // allocate managed dof storage
      VectorType& allocateDofVector ( const DiscreteFunctionSpaceType& space )
      {
        vec_.reset( new VectorType( space.size() ) );
        return *vec_;
      }

      // pointer to DofContainer
      std::unique_ptr< VectorType > vec_;
      // dof vector that stores referenc to vector
      DofVectorType dofVector_;
    };

    // Capabilibies
    // ------------

    namespace Capabilities
    {

      template< class DiscreteFunctionSpace, class DofVector >
      struct HasLeakPointer
        < Fem :: VectorDiscreteFunction< DiscreteFunctionSpace, DofVector > >
      : public HasLeakPointer< DofVector >
      {};

    }

  } // namespace Fem

} // namespace Dune

#include "managedvectorfunction.hh"

#endif // #ifndef DUNE_FEM_VECTORFUNCTION_HH
