/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_GRIDPART_COMMON_INDEXSETWRAPPER_HH
#define DUNE_FEM_GRIDPART_COMMON_INDEXSETWRAPPER_HH

// #warning "Using Dune::Fem::NonAdaptiveIndexSet wrapper around grid part index sets"

#include <vector>

#include <dune/fem/gridpart/common/indexset.hh>

namespace Dune
{

  namespace Fem
  {

    // NonAdaptiveIndexSet
    // -------------------

    template< class IndexSet >
    class NonAdaptiveIndexSet;



    // NonAdaptiveIndexSetTraits
    // -------------------------

    template< class IndexSet >
    struct NonAdaptiveIndexSetTraits
    {
      typedef NonAdaptiveIndexSet< IndexSet > IndexSetType;

      static const int dimension = IndexSet::dimension;

      template< int codim >
      struct Codim
        : public IndexSet::template Codim< codim >
      {};

      typedef typename IndexSet::IndexType IndexType;

      typedef typename IndexSet::Types Types;
    };



    // NonAdaptiveIndexSet
    // -------------------

    /** \brief wrapper for (adaptive) index sets that disables all support for
     *         adaptivity
     */
    template< class IndexSet >
    class NonAdaptiveIndexSet
      : public AdaptiveIndexSet< NonAdaptiveIndexSetTraits< IndexSet > >
    {
      typedef AdaptiveIndexSet< NonAdaptiveIndexSetTraits< IndexSet > > BaseType;

    public:
      /** \copydoc Dune::Fem::ConsecutiveIndexSet::IndexType */
      typedef typename BaseType::IndexType IndexType;

      explicit NonAdaptiveIndexSet ( const IndexSet &indexSet )
        : indexSet_( indexSet )
      {}

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::types */
      typename BaseType::Types types ( int codim ) const
      {
        return indexSet().types( codim );
      }

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::geomTypes */
      const std::vector< GeometryType > &geomTypes ( int codim ) const
      {
        return indexSet().geomTypes( codim );
      }

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::contains */
      template< class Entity >
      bool contains ( const Entity &entity ) const
      {
        return indexSet().contains( entity );
      }

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::size */
      IndexType size ( GeometryType type ) const
      {
        return indexSet().size( type );
      }

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::zie */
      IndexType size ( int codim ) const
      {
        return indexSet().size( codim );
      }

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::index */
      template< class Entity >
      IndexType index ( const Entity &entity ) const
      {
        return index< Entity::codimension >( entity );
      }

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::index */
      template< int codim >
      IndexType index ( const typename BaseType::template Codim< codim >::Entity &entity ) const
      {
        return indexSet().template index< codim >( entity );
      }

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::subIndex */
      template< class Entity >
      IndexType subIndex ( const Entity &entity, int i, unsigned int cd ) const
      {
        return subIndex< Entity::codimension >( entity, i, cd );
      }

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::subIndex */
      template< int codim >
      IndexType subIndex ( const typename BaseType::template Codim< codim >::Entity &entity, int i, unsigned int cd ) const
      {
        return indexSet().template subIndex< codim >( entity, i, cd );
      }

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::consecutive */
      static constexpr bool consecutive () noexcept
      // DUNE_DEPRECATED_MSG("IndexSet::consecutive() is deprecated, use Capabilities::isConsecutiveIndexSet<IndexSet>::v instead")
      {
        return false;
      }

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::resize */
      static void resize () {}

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::compress */
      static constexpr bool compress () noexcept { return false; }

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::insertEntity */
      static void insertEntity ( const typename BaseType::template Codim< 0 >::Entity & )
      {}

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::removeEntity */
      static void removeEntity ( const typename BaseType::template Codim< 0 >::Entity & )
      {}

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::backup */
      void backup () const {}

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::restore */
      void restore () {}

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::write */
      template< class T >
      void write ( OutStreamInterface< T > &stream ) const
      {}

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::read */
      template< class T >
      void read ( InStreamInterface< T > &stream )
      {}

      /** \copydoc Dune::Fem::AdaptiveIndexSet::numberOfHoles */
      static constexpr int numberOfHoles ( ... ) noexcept
      {
        return 0;
      }

      /** \copydoc Dune::Fem::AdaptiveIndexSet::oldIndex */
      static constexpr int oldIndex ( ... ) noexcept
      {
        return 0;
      }

      /** \copydoc Dune::Fem::AdaptiveIndexSet::newIndex */
      static constexpr int newIndex ( ... ) noexcept
      {
        return 0;
      }

    private:
      const IndexSet &indexSet () const { return indexSet_; }

      const IndexSet &indexSet_;
    };



    namespace Capabilities
    {

      template< class IndexSet >
      struct isConsecutiveIndexSet< NonAdaptiveIndexSet< IndexSet > >
      {
        static const bool v = false;
      };

      template< class IndexSet >
      struct isAdaptiveIndexSet< NonAdaptiveIndexSet< IndexSet > >
      {
        static const bool v = false;
      };

    } // namespace Capabilities

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_GRIDPART_COMMON_INDEXSETWRAPPER_HH
