/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_GRIDPART_TEST_PERSISTENTINDEXSET_HH
#define DUNE_FEM_GRIDPART_TEST_PERSISTENTINDEXSET_HH

#include <type_traits>
#include <utility>

#include <dune/fem/gridpart/common/indexset.hh>
#include <dune/fem/io/file/persistencemanager.hh>
#include <dune/fem/io/streams/streams.hh>
#include <dune/fem/space/common/dofmanager.hh>

namespace Dune
{

  namespace Fem
  {

    // Internal forward declaration
    // ----------------------------

    template< class Traits >
    class PersistentConsecutiveIndexSet;
    template< class Traits >
    class PersistentAdaptiveIndexSet;



    // PersistentIndexSetInterface
    // ---------------------------

    /** \brief virtual base class for persistent index sets
     *
     */
    struct PersistentIndexSetInterface
    {
      virtual ~PersistentIndexSetInterface () {}

      /** \brief please doc me */
      virtual void addBackupRestore () = 0;

      /** \brief please doc me */
      virtual void removeBackupRestore () = 0;
    };



    namespace Capabilities
    {

      // isPersistentIndexSet
      // --------------------

#ifndef DOXYGEN

      template< class IndexSet, bool value = std::is_base_of< PersistentIndexSetInterface, IndexSet >::type::value >
      struct __isPersistentIndexSet;

      template< class IndexSet >
      struct __isPersistentIndexSet< IndexSet, true >
      {
        static const bool v = true;

        static constexpr PersistentIndexSetInterface* map ( IndexSet &indexSet )
        {
          return static_cast< PersistentIndexSetInterface * >( &indexSet );
        }
      };

      template< class IndexSet >
      struct __isPersistentIndexSet< IndexSet, false >
      {
        static const bool v = false;

        static constexpr PersistentIndexSetInterface* map ( IndexSet & ) noexcept
        {
          return nullptr;
        }
      };

#endif // #ifndef DOXYGEN

      /** \brief capability for persistent index sets
       *
       *  \tparam  IndexSet  an index set type
       *
       *  \note default value is \b true if IndexSet is derived from
       *        PersistentIndexSetInterface
       */
      template< class IndexSet >
      struct isPersistentIndexSet
        : public __isPersistentIndexSet< IndexSet >
      {
      private:
        typedef __isPersistentIndexSet< IndexSet > BaseType;

      public:
        /** \brief please doc me */
        static const bool v = BaseType::v;

        /** \brief please doc me */
        static constexpr PersistentIndexSetInterface* map ( IndexSet &indexSet ) noexcept
        {
          return BaseType::map( indexSet );
        }
      };

#ifndef DOXYGEN

      template< class IndexSet >
      struct isPersistentIndexSet< const IndexSet >
        : public isPersistentIndexSet< IndexSet >
      {};

#endif // #ifndef DOXYGEN

    } // namespace Capabilities



    // PersistentIndexSet
    // ------------------

    /** \brief please doc me
     *
     *  \tparam  Grid  grid type
     *  \tparam  Implementation  index set type
     */
    template< class Traits, template< class > class Base >
    class PersistentIndexSet
      : public Base< Traits >,
        public PersistentIndexSetInterface
    {
      typedef Base< Traits > BaseType;

    protected:
      using BaseType::impl;

      /** \brief grid type */
      typedef typename Traits::GridType GridType;
      /** \brief dof manager type */
      typedef DofManager< GridType > DofManagerType;

      explicit PersistentIndexSet ( const GridType &grid )
        : grid_( grid ),
          dofManager_( DofManagerType::instance( grid ) ),
          counter_( 0 )
      {
        dofManager_.addIndexSet( impl() );
      }

    public:
      using BaseType::read;
      using BaseType::write;

      ~PersistentIndexSet ()
      {
        dofManager_.removeIndexSet( impl() );
      }

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::backup */
      void backup () const
      {
        if( needsBackupRestore() )
          write( PersistenceManager::backupStream() );
      }

      /** \copydoc Dune::Fem::ConsecutiveIndexSet::backup */
      void restore ()
      {
        if( needsBackupRestore() )
          read( PersistenceManager::restoreStream() );
      }

      /** \copydoc Dune::Fem::PersistentIndexSetInterface::addBackupRestore */
      void addBackupRestore () override final { ++counter_; }

      /** \copydoc Dune::Fem::PersistentIndexSetInterface::removeBackupRestore */
      void removeBackupRestore () override final { --counter_; }

    private:
      bool needsBackupRestore () const { return counter_ > 0; }

    protected:
      const GridType &grid_;
      DofManagerType &dofManager_;

    private:
      int counter_ ;
    };



    // PersistentConsecutiveIndexSet
    // -----------------------------

    template< class Traits >
    class PersistentConsecutiveIndexSet
      : public PersistentIndexSet< Traits, ConsecutiveIndexSet >
    {
      typedef PersistentIndexSet< Traits, ConsecutiveIndexSet > BaseType;

    protected:
      explicit PersistentConsecutiveIndexSet ( const typename BaseType::GridType &grid )
        : BaseType( grid )
      {}
    };



    // PersistentAdaptiveIndexSet
    // --------------------------

    template< class Traits >
    class PersistentAdaptiveIndexSet
      : public PersistentIndexSet< Traits, AdaptiveIndexSet >
    {
      typedef PersistentIndexSet< Traits, AdaptiveIndexSet > BaseType;

    protected:
      explicit PersistentAdaptiveIndexSet ( const typename BaseType::GridType &grid )
        : BaseType( grid )
      {}
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_GRIDPART_TEST_PERSISTENTINDEXSET_HH
