/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_THEADFILTER_HH
#define DUNE_FEM_THEADFILTER_HH

//- system includes
#include <algorithm>

//- dune-fem includes
#include <dune/fem/gridpart/filter/filter.hh>
#include <dune/fem/space/common/arrays.hh>

namespace Dune
{
  namespace Fem
  {

    // forward declarations
    // --------------------

    template< class > class FilterDefaultImplementation;
    template< class > class ThreadFilter;


    // ThreadFilterTraits
    // ------------------------

    template< class GridPartImp >
    struct ThreadFilterTraits
    {
      //! \brief grid part type
      typedef GridPartImp GridPartType;

      //! \brief filter type
      typedef ThreadFilter< GridPartType > FilterType;

      //! \brief entity types
      template < int cd >
      struct Codim
      {
        //! \brief entity type for given codimension
        typedef typename GridPartType::template Codim< cd >::EntityType EntityType;
      };

      //! \brief entity type for codimension 0
      typedef typename Codim< 0 >::EntityType EntityType;
    };


    // ThreadFilter
    // ------------------

    template< class GridPartImp >
    class ThreadFilter
    : public FilterDefaultImplementation< ThreadFilterTraits< GridPartImp > >
    {
      // type of grid part
      typedef GridPartImp GridPartType;

      // type of traits
      typedef ThreadFilterTraits< GridPartType > Traits;

      // this type
      typedef ThreadFilter< GridPartType > ThisType;

      // base type
      typedef FilterDefaultImplementation< Traits > BaseType;

      static const int dimension = GridPartType::GridType::dimension;

      static const int nCodim = dimension+1;

    public:
      //! \brief type of the filter implementation
      typedef typename Traits::FilterType FilterType;

      //! \brief type of index set
      typedef typename GridPartType :: IndexSetType IndexSetType;

      template< int cd >
      struct Codim
      {
        typedef typename Traits::template Codim< cd >::EntityType EntityType;
      };

      //! \brief type of codim 0 entity
      typedef typename Traits::EntityType EntityType;

      typedef MutableArray< int >  ThreadArrayType ;

      //! \brief constructor
      ThreadFilter ( const GridPartType & gridPart,
                     const ThreadArrayType& threadNum,
                     const int thead )
      : indexSet_( gridPart.indexSet() ),
        threadNum_( threadNum ),
        thread_( thead )
      { }

      //! \brief copy constructor
      ThreadFilter ( const ThisType & other )
      : indexSet_( other.indexSet_ ),
        threadNum_( other.threadNum_ ),
        thread_( other.thread_ )
      {
      }

      //! \brief return false since all interior intersections should be skipped
      template< class Intersection >
      bool interiorIntersection( const Intersection &intersection ) const
      {
        return false;
      }

      //! \brief returns true if the given entity has the correct thread number
      //! for higher codims false is returned
      template< int cd >
      bool contains ( const typename Codim< cd >::EntityType & entity ) const
      {
        if( cd == 0 )
        {
          return (thread_ == threadNum_[ indexSet_.index( entity ) ]);
        }
        else
          return false ;
      }

      //! \brief returns true if the given entity has the correct thread number
      //! for higher codims false is returned
      template< class Entity >
      bool contains ( const Entity & entity ) const
      {
        enum { cc = Entity::codimension };
        return contains< cc >( entity );
      }

      //! \brief returns true if an intersection is a boundary intersection
      template< class Intersection >
      bool intersectionBoundary( const Intersection & intersection ) const
      {
        return intersection.boundary();
      }

      //! \brief returns the boundary id for an intersection
      template< class Intersection >
      int intersectionBoundaryId ( const Intersection & intersection ) const
      {
        return intersection.boundaryId();
      }

      //! \brief returns true if for an intersection a neighbor exsits
      template< class Intersection >
      bool intersectionNeighbor ( const Intersection & intersection ) const
      {
        return intersection.neighbor();
      }

    protected:
      const IndexSetType& indexSet_;
      const ThreadArrayType& threadNum_ ;
      const int thread_;
    };

  }  // end namespace Fem

}  // end namespace Dune

#endif // #ifndef DUNE_FEM_THREADFILTER_HH
