/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_GRIDPART_FILTEREDGRIDPART_ITERATOR_HH
#define DUNE_FEM_GRIDPART_FILTEREDGRIDPART_ITERATOR_HH

//- system includes
#include <cassert>

//- dune-grid includes
#include <dune/grid/common/gridenums.hh>

namespace Dune
{

  namespace Fem
  {

    // FilteredGridPartIterator
    // ------------------------

    template< int codim, PartitionIteratorType pitype, class GridPartImp >
    class FilteredGridPartIterator
    {
      typedef FilteredGridPartIterator< codim, pitype, GridPartImp > ThisType;

      typedef GridPartImp GridPartType;
      typedef typename GridPartType::HostGridPartType HostGridPartType;
      typedef typename HostGridPartType::template Codim< codim >::template Partition< pitype >::IteratorType HostIteratorType;

      typedef typename GridPartType::template Codim< codim >::EntityPointerType EntityPointerType;

    public:
      // type of entity
      typedef typename HostIteratorType::Entity Entity;

      static const int codimension = codim;

      //! \brief constructor
      FilteredGridPartIterator ( const GridPartType &gridPart, const HostIteratorType &hostIterator )
      : gridPart_( gridPart ),
        hostIterator_( hostIterator ),
        hostEnd_( gridPart.hostGridPart().template end< codim, pitype >() )
      {
        if( done() )
          return;

        if( !gridPart.contains( *hostIterator_ ) )
          increment();
      }

      //! \brief constructor
      FilteredGridPartIterator ( const ThisType &other )
      : gridPart_( other.gridPart_ ),
        hostIterator_( other.hostIterator_ ),
        hostEnd_( other.hostEnd_ )
      {}

      //! \brief assignment operator
      ThisType &operator= ( const ThisType &other )
      {
        assert( &gridPart_ == &other.gridPart_ );
        hostIterator_ = other.hostIterator_;
        hostEnd_ = other.hostEnd_;
        return *this;
      }

      //! \brief increment
      void increment ()
      {
        assert( !done() );
        do { ++hostIterator_; } while ( !done() && !contains() );
      }

      //! \brief return level
      int level () const { return hostIterator_.level(); }

      //! return reference to entity object
      Entity dereference () const { return *hostIterator_; }

      /** \brief cast to entity pointer implementation
       *
       *  This cast allows to cast the iterator into an entity pointer.
       */
      operator typename EntityPointerType::Implementation () const { return hostIterator_.impl(); }

      //! \brief check for equality
      bool equals ( const ThisType &other ) const
      {
        return hostIterator_ == other.hostIterator_;
      }

    private:
      bool done () const
      {
        return (hostIterator_ == hostEnd_);
      }

      bool contains () const
      {
        assert( !done() );
        return gridPart().contains( *hostIterator_ );
      }

      // reference to grid part
      const GridPartType &gridPart () const { return gridPart_; }

      const GridPartType &gridPart_;
      HostIteratorType hostIterator_;
      HostIteratorType hostEnd_;
    };

  }  // namespace Fem

}  // namespace Dune

#endif // #ifndef DUNE_FEM_GRIDPART_FILTEREDGRIDPART_ITERATOR_HH
