/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_GRIDPART_IDGRIDPART_GEOMETRY_HH
#define DUNE_FEM_GRIDPART_IDGRIDPART_GEOMETRY_HH

#include <type_traits>

#include <dune/grid/common/geometry.hh>

namespace Dune
{

  namespace Fem
  {

    // Internal Forward Declarations
    // -----------------------------

    template< int, int, class > class IdGeometry;
    template< int, int, class > class IdLocalGeometry;


    // IdBasicGeometry
    // ---------------

    template< class Traits >
    struct IdBasicGeometry
    {
      typedef typename Traits::HostGeometryType HostGeometryType;

      static const int mydimension = HostGeometryType::mydimension;
      static const int coorddimension = HostGeometryType::coorddimension;

      typedef typename HostGeometryType::ctype ctype;
      typedef FieldVector< ctype, mydimension > LocalVector;
      typedef FieldVector< ctype, coorddimension > GlobalVector;

      typedef typename HostGeometryType::JacobianTransposed JacobianTransposed;
      typedef typename HostGeometryType::JacobianInverseTransposed JacobianInverseTransposed;

      IdBasicGeometry ( const HostGeometryType &hostGeometry )
      : hostGeometry_( hostGeometry )
      {}

      operator bool () const { return bool( hostGeometry_ ); }

      GeometryType type () const { return hostGeometry_.type(); }
      bool affine () const { return hostGeometry_.affine(); }

      int corners () const { return hostGeometry_.corners(); }
      GlobalVector corner ( const int i ) const { return hostGeometry_.corner( i ); }
      GlobalVector center () const { return hostGeometry_.center(); }

      GlobalVector global ( const LocalVector &local ) const { return hostGeometry_.global( local ); }
      LocalVector local ( const GlobalVector &global ) const { return hostGeometry_.local( global ); }

      ctype integrationElement ( const LocalVector &local ) const { return hostGeometry_.integrationElement( local ); }
      ctype volume () const { return hostGeometry_.volume(); }

      JacobianTransposed jacobianTransposed ( const LocalVector &local ) const
      {
        return hostGeometry_.jacobianTransposed( local );
      }

      JacobianInverseTransposed jacobianInverseTransposed ( const LocalVector &local ) const
      {
        return hostGeometry_.jacobianInverseTransposed( local );
      }

    private:
      HostGeometryType hostGeometry_;
    };



    // IdGeometryTraits
    // ----------------

    template< int mydim, class GridFamily >
    struct IdGeometryTraits
    {
      typedef typename std::remove_const< GridFamily >::type::Traits::HostGridPartType HostGridPartType;

      static const int dimension = HostGridPartType::dimension;
      static const int mydimension = mydim;
      static const int codimension = dimension - mydimension;

      typedef typename HostGridPartType::template Codim< codimension >::GeometryType HostGeometryType;
    };



    // IdGeometry
    // ----------

    template< int mydim, int cdim, class GridFamily >
    class IdGeometry
    : public IdBasicGeometry< IdGeometryTraits< mydim, GridFamily > >
    {
      typedef IdBasicGeometry< IdGeometryTraits< mydim, GridFamily > > Base;

    public:
      typedef typename Base::HostGeometryType HostGeometryType;

      IdGeometry ()
      {}

      IdGeometry ( const HostGeometryType &hostGeometry )
      : Base( hostGeometry )
      {}
    };



    // IdLocalGeometryTraits
    // ---------------------

    template< int mydim, class GridFamily >
    struct IdLocalGeometryTraits
    {
      typedef typename std::remove_const< GridFamily >::type::Traits::HostGridPartType HostGridPartType;

      static const int dimension = HostGridPartType::dimension;
      static const int mydimension = mydim;
      static const int codimension = dimension - mydimension;

      typedef typename HostGridPartType::template Codim< codimension >::LocalGeometryType HostGeometryType;
    };



    // IdLocalGeometry
    // -------------------

    template< int mydim, int cdim, class GridFamily >
    class IdLocalGeometry
    : public IdBasicGeometry< IdLocalGeometryTraits< mydim, GridFamily > >
    {
      typedef IdBasicGeometry< IdLocalGeometryTraits< mydim, GridFamily > > Base;

    public:
      typedef typename Base::HostGeometryType HostGeometryType;

      IdLocalGeometry ()
      {}

      IdLocalGeometry ( const HostGeometryType &hostGeometry )
      : Base( hostGeometry )
      {}
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_GRIDPART_IDGRIDPART_GEOMETRY_HH
