/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_GRIDPART_IDGRIDPART_INDEXSET_HH
#define DUNE_FEM_GRIDPART_IDGRIDPART_INDEXSET_HH

#include <type_traits>
#include <vector>

#include <dune/geometry/type.hh>

#include <dune/fem/gridpart/common/indexset.hh>
#include <dune/fem/gridpart/common/persistentindexset.hh>

#include <dune/fem/io/streams/streams.hh>

namespace Dune
{

  namespace Fem
  {

    // Internal forward declaration
    // ----------------------------

    template< class GridFamily >
    class IdIndexSet;



    namespace __IdIndexSet
    {

      // IndexSet
      // --------

      template< class GridFamily >
      class IndexSet
      {
      protected:
        typedef typename std::remove_const< GridFamily >::type::Traits Traits;

      public:
        typedef typename Traits::HostGridPartType::IndexSetType HostIndexSetType;

        static const int dimension = HostIndexSetType::dimension;

        template< int codim >
        struct Codim
        {
          typedef typename Traits::template Codim< codim >::Entity Entity;
        };

        typedef typename HostIndexSetType::IndexType IndexType;

        typedef typename HostIndexSetType::Types Types;

        explicit IndexSet ( const HostIndexSetType &hostIndexSet )
          : hostIndexSet_( hostIndexSet )
        {}

        Types types ( int codim ) const
        {
          return hostIndexSet().types( codim );
        }

        const std::vector< GeometryType > &geomTypes ( int codim ) const
        {
          return hostIndexSet().geomTypes( codim );
        }

        template< class Entity >
        bool contains ( const Entity &entity ) const
        {
          return hostIndexSet().contains( entity.impl().hostEntity() );
        }

        IndexType size ( GeometryType type ) const
        {
          return hostIndexSet().size( type );
        }

        IndexType size ( int codim ) const
        {
          return hostIndexSet().size( codim );
        }

        template< class Entity >
        IndexType index ( const Entity &entity ) const
        {
          return index< Entity::codimension >( entity );
        }

        template< int codim >
        IndexType index ( const typename Codim< codim >::Entity &entity ) const
        {
          return hostIndexSet().template index< codim >( entity.impl().hostEntity() );
        }

        template< class Entity >
        IndexType subIndex ( const Entity &entity, int i, unsigned int cd ) const
        {
          return subIndex< Entity::codimension >( entity, i, cd );
        }

        template< int codim >
        IndexType subIndex ( const typename Codim< codim >::Entity &entity, int i, unsigned int cd ) const
        {
          return hostIndexSet().template subIndex< codim >( entity.impl().hostEntity(), i, cd );
        }

        const HostIndexSetType &hostIndexSet () const
        {
          return hostIndexSet_;
        }

      private:
        const HostIndexSetType &hostIndexSet_;
      };



      // ConsecutiveIndexSet
      // -------------------

      template< class GridFamily >
      class ConsecutiveIndexSet
        : public IndexSet< GridFamily >
      {
        typedef IndexSet< GridFamily > BaseType;

      public:
        typedef typename BaseType::HostIndexSetType HostIndexSetType;

        using BaseType::hostIndexSet;

        explicit ConsecutiveIndexSet ( const HostIndexSetType &hostIndexSet )
          : BaseType ( hostIndexSet )
        {}

        bool consecutive () const { return hostIndexSet().consecutive(); }

        void resize () { hostIndexSet().resize(); }

        bool compress () { return hostIndexSet().compress(); }

        void insertEntity ( const typename BaseType::template Codim< 0 >::Entity &entity )
        {
          hostIndexSet().insertEntity( entity.impl().hostEntity() );
        }

        void removeEntity ( const typename BaseType::template Codim< 0 >::Entity &entity )
        {
          hostIndexSet().removeEntity( entity.impl().hostEntity() );
        }

        void backup () const { hostIndexSet().backup(); }

        void restore () { hostIndexSet().restore(); }

        template< class T >
        void write ( OutStreamInterface< T > &stream )
        {
          hostIndexSet().write( stream );
        }

        template< class T >
        void read ( InStreamInterface< T > &stream )
        {
          hostIndexSet().read( stream );
        }

      protected:
        HostIndexSetType &hostIndexSet ()
        {
          return const_cast< HostIndexSetType & >( BaseType::hostIndexSet() );
        }
      };



      // AdaptiveIndexSet
      // ----------------

      template< class GridFamily >
      class AdaptiveIndexSet
        : public ConsecutiveIndexSet< GridFamily >
      {
        typedef ConsecutiveIndexSet< GridFamily > BaseType;

      public:
        explicit AdaptiveIndexSet ( const typename BaseType::HostIndexSetType &hostIndexSet )
          : BaseType ( hostIndexSet )
        {}

        int numberOfHoles ( GeometryType type ) const
        {
          return this->hostIndexSet().numberOfHoles( type );
        }
        
        int numberOfHoles ( int codim ) const
        {
          return this->hostIndexSet().numberOfHoles( codim );
        }

        int oldIndex ( int hole, GeometryType type ) const
        {
          return this->hostIndexSet().oldIndex( hole, type );
        }
        
        int oldIndex ( int hole, int codim ) const
        {
          return this->hostIndexSet().oldIndex( hole, codim );
        }

        int newIndex ( int hole, GeometryType type ) const
        {
          return this->hostIndexSet().newIndex( hole, type );
        }

        int newIndex ( int hole, int codim ) const
        {
          return this->hostIndexSet().newIndex( hole, codim );
        }
      };



      // Implementation
      // --------------

      template< class GridFamily,
                class HostIndexSet = typename std::remove_const< GridFamily >::type::Traits::HostGridPartType::IndexSetType,
                bool consecutive = Capabilities::isConsecutiveIndexSet< HostIndexSet >::v,
                bool adaptive = Capabilities::isAdaptiveIndexSet< HostIndexSet >::v >
      struct Implementation
      {
        typedef typename std::conditional< adaptive,
            AdaptiveIndexSet< GridFamily >,
            typename std::conditional< consecutive,
                ConsecutiveIndexSet< GridFamily >,
                IndexSet< GridFamily >
              >::type
          >::type Type;
      };

    } // namespace __IdIndexSet



    // IdIndexSet
    // ----------

    template< class GridFamily >
    class IdIndexSet
      : public __IdIndexSet::Implementation< GridFamily >::Type
    {
      typedef typename __IdIndexSet::Implementation< GridFamily >::Type BaseType;

      friend class Capabilities::isPersistentIndexSet< IdIndexSet< GridFamily > >;

    public:
      explicit IdIndexSet ( const typename BaseType::HostIndexSetType &hostIndexSet )
        : BaseType ( hostIndexSet )
      {}
    };



    namespace Capabilities
    {

      template< class GridFamily >
      struct isConsecutiveIndexSet< IdIndexSet< GridFamily > >
        : public isConsecutiveIndexSet< typename IdIndexSet< GridFamily >::HostIndexSetType >
      {};

      template< class GridFamily >
      struct isAdaptiveIndexSet< IdIndexSet< GridFamily > >
        : public isAdaptiveIndexSet< typename IdIndexSet< GridFamily >::HostIndexSetType >
      {};

      template< class GridFamily >
      struct isPersistentIndexSet< IdIndexSet< GridFamily > >
      {
      private:
        typedef IdIndexSet< GridFamily > IndexSetType;
        typedef typename IndexSetType::HostIndexSetType HostIndexSetType;

      public:
        static const bool v = isPersistentIndexSet< HostIndexSetType >::v;

        static constexpr PersistentIndexSetInterface *map ( IndexSetType &indexSet ) noexcept
        {
          return isPersistentIndexSet< HostIndexSetType >::map( indexSet.hostIndexSet() );
        }
      };

    } // namespace Capabilities

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_GRIDPART_IDGRIDPART_INDEXSET_HH
