/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_GRIDPART_TEST_CHECKSEED_HH
#define DUNE_FEM_GRIDPART_TEST_CHECKSEED_HH

//- C++ includes
#include <limits>
#include <type_traits>

//- dune-common includes
#include <dune/common/forloop.hh>
#include <dune/common/exceptions.hh>

//- dune-fem includes
#include <dune/fem/gridpart/common/capabilities.hh>


namespace Dune
{

  namespace Fem
  {

    /** \brief Iterate over all entities of all codimensions available in the grid part,
     *         and perform a number of tests on each entity.
     */

    template< class GridPartType >
    class CheckEntitySeed
    {
      template< class GridPart, bool >
      struct If
      {
        template< int codim >
        static void check ( const GridPart & )
        { }
      };

      template< class GridPart >
      struct If< GridPart, true >
      {
        template< const int codim >
        static void check ( const GridPart &gridPart )
        {
          typedef typename GridPart::template Codim< codim >::IteratorType IteratorType;
          typedef typename GridPart::template Codim< codim >::EntityType EntityType;
          typedef typename GridPart::template Codim< codim >::EntitySeedType EntitySeedType;
          static_assert( std::is_same< EntitySeedType, typename EntityType::EntitySeed >::value, "Types of EntitySeed do not coincide" );

          const IteratorType end = gridPart.template end< codim >();
          for( IteratorType it = gridPart.template begin< codim >(); it != end; ++it )
          {
            EntityType entity = *it;
            EntitySeedType seed = entity.seed();
            if( gridPart.entity( seed ) != entity )
              DUNE_THROW( Dune::Exception, "Could not recover entity from seed" );
          }
        }
      };

      template< int codim >
      struct CheckCodim
      {
        template< class GridPart >
        static void apply ( const GridPart &gridPart )
        {
          // check whether grid part has entity of given codimension
          const bool hasEntity = Dune::Fem::GridPartCapabilities::hasEntity< GridPart, codim >::v;
          If< GridPart, hasEntity >::template check< codim >( gridPart );
        }
      };

    public:
      static const int dimension = GridPartType::dimension;

      static void check ( const GridPartType &gridPart )
      {
        // generic for loop over all codimensions
        Dune::ForLoop< CheckCodim, 0, dimension >::apply( gridPart );
      }
    };

  } // end namespace Fem

} // end namespace Dune

#endif // #ifndef DUNE_FEM_GRIDPART_TEST_CHECKSEED_HH
