/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
// (c) Robert Kloefkorn
#ifndef DUNE_FEM_INPUTOUPUTLOCK_HH
#define DUNE_FEM_INPUTOUPUTLOCK_HH

//- system includes
#include <cstdio>
#include <cstdlib>

#include <iostream>
#include <fstream>
#include <string>

namespace Dune
{

  namespace Fem
  {

    //! creates and removes lock file during backup process
    class FileIOLock
    {
      std::string filename_;
      FileIOLock ( const FileIOLock & );
    public :
      //! creates lock file
      FileIOLock(const std::string& fn);
      //! removes lock file
     ~FileIOLock() ;

      //! suffix that is appended to lock files
      static const char * suffix() { return "lock"; }
    };

    //! check if lock file exists and aborts if true
    class FileIOCheckError
    {
      FileIOCheckError( const FileIOCheckError & );
    public :
      //! creates lock file
      FileIOCheckError(const std::string& fn)
      {
        std::string lockfile(fn);
        lockfile += ".";
        lockfile += FileIOLock::suffix();

        std::ifstream file ( lockfile.c_str () );
        if( file.is_open() )
        {
          std::cerr << "ERROR: data set `"<<fn<<"' not complete, lock file exists! " << std::endl;
          abort();
        }
      }
    };

    ////////////////////////////////////////////////////////////////
    //
    //  INLINE
    //
    ////////////////////////////////////////////////////////////////

    // create lock file
    inline FileIOLock :: FileIOLock (const std::string& fn)
      : filename_(fn)
    {
      if( filename_ == "" )
      {
        filename_ = suffix();
      }
      else   {
        filename_ += ".";
        filename_ += suffix();
      }

      std::ofstream file ( filename_.c_str() );
      if( !file )
      {
        std::cerr << "WARNING: Couldn't open lock file `"<<filename_<<"' in: ";
        std::cerr << __FILE__<< " line: "<< __LINE__ << std::endl;
      }
      else
      {
        file.close();
      }
      return ;
    }

    // remove lock file
    inline FileIOLock :: ~FileIOLock ()
    {
      if (filename_ != "")
      {
        int test = remove (filename_.c_str()) ;
        if (test != 0)
        {
          std::cerr << "WARNING: Couldn't remove lock file `"<<filename_<<"' in: ";
          std::cerr <<__FILE__<<" line: " <<__LINE__ << std::endl ;
        }
      }
      return ;
    }

  } // namespace Fem

} // namespace Dune
#endif // #ifndef DUNE_FEM_INPUTOUPUTLOCK_HH
