/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_IO_PARAMETER_READER_HH
#define DUNE_FEM_IO_PARAMETER_READER_HH

#include <cassert>

#include <functional>
#include <iostream>
#include <string>
#include <utility>

#include <dune/fem/io/parameter/exceptions.hh>
#include <dune/fem/io/parameter/parser.hh>

namespace Dune
{

  namespace Fem
  {

    // BasicParameterReader
    // --------------------

    template< class Parameter >
    struct BasicParameterReader
    {
      explicit BasicParameterReader ( Parameter parameter = Parameter() )
        : parameter_( std::move( parameter ) )
      {}

      /**
       * \brief check, whether a parameter is defined
       *
       * \param[in]   key    name of the parameter to check
       *
       * \returns \b true, if the parameter is found, \b false otherwise
       */
      bool exists ( const std::string &key ) const { return static_cast< bool >( parameter_( key, nullptr ) ); }

      /**
       * \brief get mandatory parameter
       *
       * \note This method throws an exception, if the parameter cannot be
       *       found.
       *
       * \param[in]   key    name of the parameter to get
       * \param[out]  value  value of the parameter
       */
      template< class T >
      void get ( const std::string &key, T &value ) const
      {
        const std::string *string = parameter_( key, nullptr );
        if( !string )
          DUNE_THROW( ParameterNotFound, "Parameter '" << key << "' not found." );
        if( !ParameterParser< T >::parse( *string, value ) )
          DUNE_THROW( ParameterInvalid, "Parameter '" << key << "' invalid." );
      }

      /**
       * \brief get optional parameter
       *
       * \note This method returns a default value, if the parameter cannot be
       *       found.
       *
       * \param[in]   key           name of the parameter to get
       * \param[in]   defaultValue  default value for this parameter
       * \param[out]  value         value of the parameter
       */
      template< class T >
      void get ( const std::string &key, const T &defaultValue, T &value ) const
      {
        const std::string defaultString = ParameterParser< T >::toString( defaultValue );
        const std::string *string = parameter_( key, &defaultString );
        assert( string );
        if( !ParameterParser< T >::parse( *string, value ) )
          DUNE_THROW( ParameterInvalid, "Parameter '" << key << "' invalid." );
      }

      /**
       * \brief get optional parameter (special case for string)
       *
       * \note This method returns a default value, if the parameter cannot be
       *       found.
       *
       * \param[in]   key           name of the parameter to get
       * \param[in]   defaultValue  default value for this parameter
       * \param[out]  value         value of the parameter
       */
      void get ( const std::string &key, const char* defaultValue, std::string &value ) const
      {
        const std::string defaultString( defaultValue );
        const std::string *string = parameter_( key, &defaultString );
        assert( string );
        if( !ParameterParser< std::string >::parse( *string, value ) )
          DUNE_THROW( ParameterInvalid, "Parameter '" << key << "' invalid." );
      }

      /**
       * \brief get mandatory parameter
       *
       * \note This method throws an exception, if the parameter cannot be
       *       found.
       *
       * \param[in]   key        name of the parameter to get
       * \param[in]   validator  validator for the parameter value
       * \param[out]  value      value of the parameter
       */
      template< class T, class Validator >
      void getValid ( const std::string &key, const Validator &validator, T &value ) const
      {
        const std::string *string = parameter_( key, nullptr );
        if( !string )
          DUNE_THROW( ParameterNotFound, "Parameter '" << key << "' not found." );
        if( !ParameterParser< T >::parse( *string, value ) || !validator( value ) )
          DUNE_THROW( ParameterInvalid, "Parameter '" << key << "' invalid." );
      }

      /**
       * \brief get optional parameter
       *
       * \note This method returns a default value, if the parameter cannot be
       *       found.
       *
       * \param[in]   key           name of the parameter to get
       * \param[in]   defaultValue  default value for this parameter
       * \param[in]   validator     validator for the parameter value
       * \param[out]  value         value of the parameter
       */
      template< class T, class Validator >
      void getValid ( const std::string &key, const T &defaultValue, const Validator &validator, T &value ) const
      {
        const std::string defaultString = ParameterParser< T >::toString( defaultValue );
        const std::string *string = parameter_( key, &defaultString );
        assert( string );
        if( !ParameterParser< T >::parse( *string, value ) || !validator( value ) )
          DUNE_THROW( ParameterInvalid, "Parameter '" << key << "' invalid." );
      }

      /**
       * \brief get mandatory parameter
       *
       * \note This method throws an exception, if the parameter cannot be
       *       found.
       *
       * \param[in]   key    name of the parameter to get
       *
       * \returns value of the parameter
       */
      template< class T >
      T getValue ( const std::string &key ) const
      {
        T value;
        get( key, value );
        return value;
      }

      /**
       * \brief get optional parameter
       *
       * \note This method returns a default value, if the parameter cannot be
       *       found.
       *
       *  \param[in]   key           name of the parameter to get
       *  \param[in]   defaultValue  default value for this parameter
       *
       *  \returns value of the parameter
       */
      template< class T >
      T getValue ( const std::string &key, const T &defaultValue ) const
      {
        T value = defaultValue;
        get( key, defaultValue, value );
        return value;
      }

      /**
       * \brief get optional parameter
       *
       * \note This method returns a default value, if the parameter cannot be
       *       found.
       *
       *  \param[in]   key           name of the parameter to get
       *  \param[in]   validator     validator for the parameter value
       *
       *  \returns value of the parameter
       */
      template< class T, class Validator >
      T getValidValue ( const std::string &key, const Validator &validator ) const
      {
        T value;
        getValid( key, validator, value );
        return value;
      }

      /**
       * \brief get optional parameter
       *
       * \note This method returns a default value, if the parameter cannot be
       *       found.
       *
       * \param[in]   key           name of the parameter to get
       * \param[in]   defaultValue  default value for this parameter
       * \param[in]   validator     validator for the parameter value
       *
       * \returns value of the parameter
       */
      template< class T, class Validator >
      T getValidValue ( const std::string &key, const T &defaultValue, const Validator &validator ) const
      {
        T value;
        getValid( key, defaultValue, validator, value );
        return value;
      }

      template< int n >
      int getEnum ( const std::string &key, const std::string (&values)[ n ] ) const
      {
        const std::string *string = parameter_( key, nullptr );
        if( !string )
          DUNE_THROW( ParameterNotFound, "Parameter '" << key << "' not found." );
        return getEnumeration( key, *string, values );
      }

      template< int n >
      int getEnum ( const std::string &key, const std::string (&values)[ n ], int defaultValue ) const
      {
        const std::string *string = parameter_( key, &values[ defaultValue ] );
        return getEnumeration( key, *string, values );
      }

    private:
      template< int n >
      static int getEnumeration ( const std::string &key, const std::string& value, const std::string (&values)[ n ] )
      {
        for( int i = 0; i < n; ++i )
        {
          if( value == values[ i ] )
            return i;
        }

        int j = -1;
        if( !ParameterParser< int >::parse( value, j ) )
          j = -1;
        if( (j < 0) || (j >= n) )
        {
          std::cerr << std::endl << "Parameter '" << key << "' invalid." << std::endl;
          std::cerr << "Valid values are: ";
          for( int i = 0; i < n; ++i )
            std::cerr << values[ i ] << (i < n-1 ? ", " : "");
          std::cerr << std::endl << std::endl;
          DUNE_THROW( ParameterInvalid, "Parameter '" << key << "' invalid." );
        }
        return j;
      }

    protected:
      Parameter parameter_;
    };



    // ParameterReader
    // ---------------

    typedef BasicParameterReader< std::function< const std::string *( const std::string &, const std::string * ) > > ParameterReader;

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_IO_PARAMETER_READER_HH
