/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
//**************************************************************
//  (C) written and directecd by Robert Kloefkorn
//**************************************************************
#ifndef DATADISP_READIOTUPLE_CC
#define DATADISP_READIOTUPLE_CC

// needed in readiotparams.cc
#define USE_GRAPE_DISPLAY

//- system includes
#include <stack>

//- include grape io stuff
#include <dune/fem/io/file/asciiparser.hh>
#include <dune/fem/io/file/iotuple.hh>
#include <dune/fem/io/file/iointerface.hh>

//! type of used grid
typedef Dune::GridSelector::GridType GR_GridType;

//! type of GrapeDisplay
typedef GrapeDataDisplay<GR_GridType> GrapeDispType;

//! type of discrete function tuple
typedef GR_InputType GR_DiscFuncType;

static std::stack<GR_GridType *> gridStack;
static std::stack<GrapeDispType *> dispStack;

template <class T>
inline void deleteObjects(std::stack<T *> & stack);

inline void dataDispErrorExit(std::string msg)
{
  std::cerr << msg << std::endl;
  std::cerr.flush();
  exit(EXIT_FAILURE);
}

inline GrapeDispType * readTupleData(const char * path, const char * filename,
            double & time , int n,
            int timestep, int myRank, int mySize,
            INFO* info)
{
  // check whether we use a fixed mesh
  const bool fixedMesh = (info->fix_mesh == 1) ? true : false;

  // the grid is new if not a fixed mesh or the stack is empty
  const bool newGrid = ( ! fixedMesh || gridStack.empty() );

  // init grid
  GR_GridType * grid = ( ! newGrid ) ? gridStack.top() : 0;

  assert(filename);
  std::string fn (filename);
  DATAINFO * dinf = info[n].datinf;

  Fem::IOTuple<GR_DiscFuncType>::ReturnType* tup =
    Fem::IOTuple<GR_DiscFuncType>::input(grid,time,myRank,mySize,path,fn);
  std::cout << "Finished reading grid" << std::endl;

  // push all new grids to grid stack
  if( newGrid ) gridStack.push(grid);

  GrapeDispType * disp = new GrapeDispType ( *grid, myRank );
  dispStack.push(disp);

  // discrete functions of non-valid data are removed
  Fem::IOTuple<GR_DiscFuncType>::addToDisplayOrRemove(*disp,dinf,time,*tup);

  //Element<0> :: get( *tup );
  //    *(Element<0> :: get( *tup ))
  // do some post processing
  postProcessing(*disp,*grid,time,timestep, *tup);
  return disp;
}

// setup the hole data tree for grape
inline INFO * readData(INFO * info , const char * path, int i_start, int i_end,
    int i_delta, int n, double timestep, int numProcs)
{
  double t_start = 1e308;
  double t_end = -t_start, t_act = 0.0;
  typedef CombinedGrapeDisplay < GrapeDispType > CombinedDisplayType;

  CombinedGrapeDisplay < GrapeDispType > * comdisp = new CombinedDisplayType ();

  int  ntime, n_step = 0;

  for (ntime = i_start; ntime <= i_end; ntime += i_delta)
  {
    printf("timestep = %d | last timestep = %d | stepsize = %d\n", ntime, i_end, i_delta);
    {
      int anzProcs = numProcs;

      for(int proc=0; proc<anzProcs; ++proc)
      {
        assert(path || numProcs <= 1);
        GrapeDispType *newdisp = 0;

        int anz = n; // (n > 0) ? n : 1;
        for(int i=0; i<anz; ++i)
        {
          // use standard procedure to create path name
          std::string newpath =
            Fem::IOInterface::createRecoverPath(path,proc,info[i].name,ntime);

          newdisp = readTupleData(newpath.c_str(), info[i].name,
                                  t_act , i , ntime, proc, anzProcs, info);

          if( comdisp )
          {
            assert(newdisp != 0);
            assert( comdisp );
            comdisp->addDisplay( *newdisp );
          }
        }
        assert(newdisp);
        newdisp->addMyMeshToTimeScene(info[0].tsc,t_act,proc);
      }

      if( comdisp )
      {
        // this is the combine object, which is put to the last time scene
        comdisp->addMyMeshToGlobalTimeScene(t_act,0);
      }
    }
    printf("actual time: %f (timestep size: %e)\n\n",t_act,timestep);


    if (ntime == i_start) t_start = t_end = t_act;
    t_start = std::min(t_start, t_act);
    t_end = std::max(t_end, t_act);

    if (timestep > 0) t_act += timestep*i_delta;
    n_step++;
  }
  return info;
}

template <class T>
inline void deleteObjects(std::stack<T *> & stack)
{
  while(! stack.empty() )
  {
    T * obj = stack.top();
    stack.pop();
    delete obj;
  }
  return;
}

inline void deleteAllObjects()
{
  deleteObjects(dispStack);
  deleteObjects(gridStack);
  return ;
}
#endif
