/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#include <config.h>

#include <dune/fem/misc/mpimanager.hh>
#include <dune/fem/misc/femtimer.hh>

namespace Dune
{

  namespace Fem
  {

    Timer< true >::Timer ()
    : timesS_(),
      timers_(),
      output_(),
      stepCount_(0),
      changed_(true)
    {
      push_time();
    }


    unsigned int Timer< true >::add ( const std::string &name, int nr )
    {
      unsigned int id;
      const unsigned int numTimers = timers_.size();
      for( id = 0; id < numTimers; ++id )
      {
        if( timers_[ id ].times.empty() )
          break;
      }

      if( id == numTimers )
        timers_.push_back( TimerInfo( name, nr ) );
      else
        timers_[ id ] = TimerInfo( name, nr );

      reset_timer( id );
      changed_ = true;
      return id;
    }


    void Timer< true >::remove ( unsigned int id )
    {
      timers_[ id ] = TimerInfo( "", 0 );
      changed_ = true;
    }


    void Timer< true >::remove ()
    {
      timers_.clear();
      changed_ = true;
    }


    void Timer< true >::print_timer ( std::ostream &out, int id )
    {
      const TimerInfo &info = timers_[ id ];
      const unsigned int numTimers = timers_.size();
      if( numTimers > 1 )
        out << "(" << info.name << ":";

      out << info.times[ 0 ];
      for( unsigned int i = 1; i < info.times.size(); ++i )
        out << "," << info.times[ i ] / info.times[ 0 ];

      if( numTimers > 1 )
        out << ") ";
    }


    void Timer< true >::print_timer ( std::ostream &out, const std::string &msg )
    {
      out << msg << " : ";
      for( unsigned int i = 0; i < timers_.size(); ++i )
        print_timer( out, i );
      out << std::endl;
    }

    void Timer< true >::printToFile ()
    {
      for( unsigned int i = 0; i< timers_.size(); ++i )
      {
        const TimerInfo &info = timers_[ i ];
        if( info.times.empty() )
          continue;

        output_ << std::setw( 6 ) << inMS( info.times[ 0 ] ) << "ms";
        const unsigned int numTimes = info.times.size();
        if( numTimes > 1 )
        {
          output_ << " ( ";
          for( unsigned int nr = 1; nr < numTimes; ++nr )
            output_ << std::setw( 3 ) << inProz( info.times[ nr ], info.times[ 0 ] ) << "% ";
          output_ << ") ";
        }
      }
      output_ << std::endl;
    }


    void Timer< true >::printToFile ( const std::string &fileName, int step )
    {
      if (!output_.is_open()) {
        output_.open(fileName.c_str());
        if( !output_ )
          DUNE_THROW( IOError, "FemTimer: Unable to open '" << fileName << "' for writing." );
        changed_=true;
      }
      if( changed_ )
      {
        for( unsigned int i = 0; i < timers_.size(); ++i )
        {
          const TimerInfo &info = timers_[ i ];
          if( !info.times.empty() )
            output_ << std::setw( 12 + (info.times.size()-1)*5 ) << info.name;
        }
        output_ << std::endl;
        stepCount_ = 0;
        changed_ = false;
      }
      if( stepCount_ % step == 0 )
        printToFile();
      ++stepCount_;
    }


    void Timer< true >::printToFile ( const TimeProviderBase &tp, const std::string &fileName, int step )
    {
      if (!output_.is_open()) {
        output_.open(fileName.c_str());
        if( !output_ )
          DUNE_THROW( IOError, "FemTimer: Unable to open '" << fileName << "' for writing." );
        changed_=true;
      }
      if( changed_ )
      {
        output_ << std::endl << std::endl;
        output_ << std::setw( 12 ) << "Time" << " ";
        output_ << std::setw( 12 ) << "dt" << " ";
        for( unsigned int i = 0; i < timers_.size(); ++i )
        {
          const TimerInfo &info = timers_[ i ];
          if( !info.times.empty() )
            output_ << std::setw( 12+(info.times.size()-1)*5 ) << info.name;
        }
        output_ << std::endl;
        stepCount_ = 0;
        changed_ = false;
      }
      if( stepCount_ % step == 0 )
      {
        output_ << std::setw( 10 ) << std::scientific << tp.time() << " ";
        output_ << std::setw( 10 ) << std::scientific << tp.deltaT() << " ";
        printToFile();
      }
      ++stepCount_;
    }

  } // namespace Fem

} // namespace Dune
