/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_FIELDMATRIXCONVERTER_HH
#define DUNE_FEM_FIELDMATRIXCONVERTER_HH

#include <cassert>

#include <dune/common/fmatrix.hh>

namespace Dune
{

  namespace Fem
  {

    // Internal Forward Declarations
    // -----------------------------

    template< class VectorType, class ConvertToType >
    class FieldMatrixConverter;

    template< class K, int m >
    class FieldMatrixConverterRow;

  }


  template< class K, int n, int m >
  struct DenseMatVecTraits< Fem::FieldMatrixConverter< FieldVector< K, n *m >, FieldMatrix< K, n, m > > >
  {
    typedef Fem::FieldMatrixConverter< FieldVector< K, n *m >, FieldMatrix< K, n, m > > derived_type;
    typedef DenseMatVecTraits< FieldMatrix< K, n, m > > Traits;
    typedef typename Traits::container_type container_type;
    typedef typename Traits::value_type value_type;
    typedef typename Traits::size_type size_type;
    typedef typename Traits::row_type row_type;

    typedef Fem::FieldMatrixConverterRow< K, m > row_reference;
    typedef Fem::FieldMatrixConverterRow< K, m > const_row_reference;
  };

  template< class K, int m >
  struct DenseMatVecTraits< Fem::FieldMatrixConverterRow< K, m > >
  {
    typedef Fem::FieldMatrixConverterRow< K, m > derived_type;
    typedef K *container_type;
    typedef K value_type;
    typedef size_t size_type;
  };


  namespace Fem
  {

    // derive from dense vector to inherit functionality
    template< class K, int m >
    class FieldMatrixConverterRow
      : public Dune::DenseVector< FieldMatrixConverterRow< K, m > >
    {
      typedef FieldMatrixConverterRow< K, m > This;
      typedef Dune::DenseVector< FieldMatrixConverterRow< K, m > > Base;

    public:
      typedef typename Base::size_type size_type;

      using Base::operator=;
      using Base::operator*;

      FieldMatrixConverterRow ( K *ptr )
        : ptr_( ptr )
      {
        assert( ptr_ );
      }

      FieldMatrixConverterRow ( const This &other )
        : ptr_( other.ptr_ )
      {}

      FieldMatrixConverterRow &operator= ( const This &other )
      {
        for( size_t i = 0; i < vec_size(); ++i )
          vec_access( i ) = other[ i ];
        return *this;
      }

      template< class Impl >
      FieldMatrixConverterRow &operator= ( const DenseVector< Impl > &other )
      {
        assert( other.size() == vec_size() );
        for( size_t i = 0; i < vec_size(); ++i )
          vec_access( i ) = other[ i ];
        return *this;
      }

      template< class V >
      K operator* ( const DenseVector< V > &x ) const
      {
        assert( vec_size() == x.size() );
        K result( 0 );
        for( size_type i = 0; i < vec_size(); ++i )
          result += vec_access( i ) * x[ i ];
        return result;
      }

      // make this thing a vector
      size_t vec_size () const { return m; }
      K &vec_access ( size_t i ) { assert( i < vec_size() ); return ptr_[ i ]; }
      const K &vec_access ( size_t i ) const { assert( i < vec_size() ); return ptr_[ i ]; }

    private:
      K *ptr_;
    };




    //! convert a FieldVector with length n * m to a FieldMatrix with n rows and m cols
    template< typename K, int n, int m >
    class FieldMatrixConverter< FieldVector< K, n *m >, FieldMatrix< K, n, m > >
      : public Dune::DenseMatrix< FieldMatrixConverter< FieldVector< K, n *m >, FieldMatrix< K, n, m > > >
    {
      typedef Dune::DenseMatrix< FieldMatrixConverter< FieldVector< K, n *m >, FieldMatrix< K, n, m > > > Base;

    public:
      //! internal storage of matrix
      typedef FieldVector< K, n *m > InteralVectorType;

      //! type of class return upon operator [] which behaves like a reference
      //typedef FieldMatrixConverterRow< K , m> RowType;

      typedef typename Base::row_type row_type;
      typedef typename Base::row_reference row_reference;
      typedef typename Base::const_row_reference const_row_reference;

      //! export the type representing the field
      typedef K field_type;

      //! export the type representing the components
      typedef K block_type;

      //! The type used for the index access and size operations.
      typedef std::size_t size_type;

      enum {
        //! The number of rows.
        rows = n,
        //! The number of columns.
        cols = m,
        //! The total dimension
        dimension = n * m
      };

      FieldMatrixConverter ( InteralVectorType &v )
        : vec_( &v )
#ifndef NDEBUG
          , mutableVec_( true )
#endif
      {}

      FieldMatrixConverter ( const InteralVectorType &v )
        : vec_( const_cast< InteralVectorType * >( &v ) )
#ifndef NDEBUG
          , mutableVec_( false )
#endif
      {}

      FieldMatrixConverter ( const FieldMatrixConverter &other )
        : vec_( other.vec_ )
#ifndef NDEBUG
          , mutableVec_( other.mutableVec_ )
#endif
      {}

      FieldMatrixConverter &operator= ( const FieldMatrixConverter &other )
      {
        assert( mutableVec_ );
        vec_ = other.vec_;
    #ifndef NDEBUG
        mutableVec_ = other.mutableVec_;
    #endif
        return *this;
      }

      FieldMatrixConverter &operator= ( const FieldMatrix< K, n, m > &matrix )
      {
        assert( mutableVec_ );
        for( size_t i = 0; i < rows; ++i )
          for( size_t j = 0; j < cols; ++j )
            (*vec_)[ i * cols + j ] = matrix[ i ][ j ];

        return *this;
      }
      FieldMatrixConverter &operator+= ( const FieldMatrix< K, n, m > &matrix )
      {
        assert( mutableVec_ );
        for( size_t i = 0; i < rows; ++i )
          for( size_t j = 0; j < cols; ++j )
            (*vec_)[ i * cols + j ] += matrix[ i ][ j ];

        return *this;
      }

      /** \brief Sends the matrix to an output stream */
      friend std::ostream &operator<< ( std::ostream &s,
                                        const FieldMatrixConverter &a )
      {
        for( size_type i = 0; i < n; ++i )
        {
          for( size_type j = 0; j < m; ++j )
            s << a[ i ][ j ] << " ";
          s << std::endl;
        }
        return s;
      }

      // make this thing a matrix
      size_type mat_rows () const { return rows; }
      size_type mat_cols () const { return cols; }

      row_reference mat_access ( size_type i )
      {
        assert( i < rows );
        return row_reference( (&(*vec_)[ i * cols ]) );
      }

      const_row_reference mat_access ( size_type i ) const
      {
        assert( i < rows );
        return const_row_reference( (&(*vec_)[ i * cols ]) );
      }

    protected:
      InteralVectorType *vec_;
    #ifndef NDEBUG
      bool mutableVec_;
    #endif
    };

  } // namespace Fem

  // method that implements the operator= of a FieldMatrix taking a FieldMatrixConverter
  template< class K, int n, int m >
  inline void istl_assign_to_fmatrix ( FieldMatrix< K, n, m > &A,
                                       const Fem::FieldMatrixConverter< FieldVector< K, n *m >, FieldMatrix< K, n, m > > &B )
  {
    for( size_t i = 0; i < n; ++i )
      for( size_t j = 0; j < m; ++j )
        A[ i ][ j ] = B[ i ][ j ];

  }
  template< class K, int n, int m >
  inline void istl_assign_to_fmatrix ( DenseMatrix< FieldMatrix< K, n, m > > &A,
                                       const Fem::FieldMatrixConverter< FieldVector< K, n *m >, FieldMatrix< K, n, m > > &B )
  {
    for( size_t i = 0; i < n; ++i )
      for( size_t j = 0; j < m; ++j )
        A[ i ][ j ] = B[ i ][ j ];

  }

} // namespace Dune

#endif // #ifndef DUNE_FEM_FIELDMATRIXCONVERTER_HH
