/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_MISC_FUNCTOR_HH
#define DUNE_FEM_MISC_FUNCTOR_HH

#include <cstddef>

namespace Dune
{

  namespace Fem
  {

    // DefaultAssign
    // -------------

    struct DefaultAssign
    {
      template< class T, class U >
      void operator() ( const T &a, U &b ) const
      {
        b = a;
      }
    };



    // AssignFunctor
    // -------------

    template< class Array, class Assign = DefaultAssign >
    struct AssignFunctor
    {
      explicit AssignFunctor ( Array &array, const Assign &assign = Assign() )
      : array_( array ),
        assign_( assign )
      {}

      template< class Value >
      void operator() ( const std::size_t local, const Value &value )
      {
        assign_( value, array_[ local ] );
      }

    private:
      Array &array_;
      Assign assign_;
    };

    template< class T, class Assign >
    struct AssignFunctor< T *, Assign >
    {
      explicit AssignFunctor ( T *array, const Assign &assign = Assign() )
      : array_( array ),
        assign_( assign )
      {}

      template< class Value >
      void operator() ( const std::size_t local, const Value &value )
      {
        assign_( value, array_[ local ] );
      }

    private:
      T *array_;
      Assign assign_;
    };



    // AssignSingleFunctor
    // -------------------

    template< class Value >
    struct AssignSingleFunctor
    {
      explicit AssignSingleFunctor ( const std::size_t i, Value &value )
      : localFixed_( i ),
        value_( value )
      {}

      void operator() ( const std::size_t local, const Value &globalKey )
      {
        if( local == localFixed_ )
          value_ = globalKey;
      }

    private:
      std::size_t localFixed_;
      Value &value_;
    };

    template <class Mapper2, class Entity2, class Functor>
    struct MatrixFunctor
    {
      explicit MatrixFunctor( const Mapper2 &mapper2, const Entity2 &entity2, Functor functor )
      : mapper2_(mapper2),
        entity2_(entity2),
        functor_(functor)
      {}
      void operator() ( const std::size_t local1, const typename Functor::GlobalKey &globalKey1 )
      {
        functor_.set(local1,globalKey1);
        mapper2_.mapEach(entity2_, functor_);
      }
      private:
      const Mapper2 &mapper2_;
      const Entity2 &entity2_;
      Functor functor_;
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_MISC_FUNCTOR_HH
