/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_OMPMANAGER_HH
#define DUNE_FEM_OMPMANAGER_HH

#include <cassert>
#include <cstdlib>

#if USE_PTHREADS && HAVE_PTHREAD == 0
#warning "pthreads were not found!"
#undef USE_PTHREADS
#endif

#if defined _OPENMP || USE_PTHREADS == 1
#ifndef USE_SMP_PARALLEL
#define USE_SMP_PARALLEL
#endif
#endif

#if USE_PTHREADS
#include <pthread.h>
#endif

#ifdef _OPENMP
#include <omp.h>
#endif


namespace Dune
{

  namespace Fem
  {

#ifdef _OPENMP
    struct ThreadManager
    {
      /** return maximal number of threads possbile in the current run
          \note can be set by the OMP_NUM_THREADS environment variable
                from outside of the code */
      static inline int maxThreads()
      {
        return omp_get_max_threads();
      }

      //! return number of current threads
      static inline int currentThreads()
      {
        return omp_get_num_threads();
      }

      //! return thread number
      static inline int thread()
      {
        return omp_get_thread_num();
      }

      //! return true if the current thread is the master thread (i.e. thread 0)
      static inline bool isMaster()
      {
        return thread() == 0 ;
      }

      //! set maximal number of threads available during run
      static inline void setMaxNumberThreads( const int numThreads )
      {
        omp_set_num_threads( numThreads );
      }

      //! returns true if program is operating on one thread currently
      static inline bool singleThreadMode()
      {
        return currentThreads() == 1 ;
      }
    }; // end class ThreadManager
#elif USE_PTHREADS
#warning "ThreadManager: using pthreads"

    struct ThreadManager
    {
    private:
      struct Manager
      {
        static inline Manager& instance()
        {
          static thread_local Manager mg ;
          return mg ;
        }

        inline void initThread( const int maxThreads, const int threadNum )
        {
          // thread number 0 is reserved for the master thread
          maxThreads_ = maxThreads;
          threadNum_  = threadNum ;
        }

        inline void singleThreadMode()
        {
          activeThreads_ = 1;
        }

        inline void multiThreadMode(const int nThreads )
        {
          activeThreads_ = nThreads;
        }

        inline int maxThreads() const { return maxThreads_; }
        inline int currentThreads() const { return activeThreads_; }
        inline int thread()
        {
          assert( threadNum_ >= 0 );
          return threadNum_;
        }

      private:
        int threadNum_;
        int maxThreads_;
        int activeThreads_;

        Manager()
          : threadNum_( -1 ), maxThreads_( 1 ), activeThreads_( 1 )
        {}
      };

      static inline Manager& manager()
      {
        return Manager :: instance();
      }

    public:
      ///////////////////////////////////////////////////////
      //  begin of pthread specific interface
      ///////////////////////////////////////////////////////
      //! initialize single thread mode
      static inline void initSingleThreadMode()
      {
        manager().singleThreadMode();
      }

      //! initialize multi thread mode
      static inline void initMultiThreadMode( const int nThreads )
      {
        manager().multiThreadMode( nThreads );
      }

      //! set max number of threads and thread number for this thread
      static inline void initThread( const int maxThreads, const int threadNum )
      {
        manager().initThread( maxThreads, threadNum );
      }

      ///////////////////////////////////////////////////////
      //  INTERFACE
      ///////////////////////////////////////////////////////
      //! return maximal number of threads possbile in the current run
      static inline int maxThreads()
      {
        return manager().maxThreads();
      }

      //! return number of current threads
      static inline int currentThreads()
      {
        return manager().currentThreads();
      }

      //! return thread number
      static inline int thread()
      {
        return manager().thread();
      }

      //! set maximal number of threads available during run
      static inline void setMaxNumberThreads( const int numThreads )
      {
        // this call also initiates the master thread
        manager().initThread( numThreads, 0 );
      }

      //! return true if the current thread is the master thread (i.e. thread 0)
      static inline bool isMaster()
      {
        return thread() == 0;
      }

      //! returns true if program is operating on one thread currently
      static inline bool singleThreadMode()
      {
        return currentThreads() == 1 ;
      }
    }; // end class ThreadManager (pthreads)

#else
    /** \class ThreadManager
     *  \ingroup Utility
     *  \brief The ThreadManager wrapps basic shared memory functionality
     *         provided by OpenMP or pthreads such as thread id, number of threads, etc.
     *
     *  \note All methods are static members.
     */
    struct ThreadManager
    {
      //! \brief return maximal number of threads possbile in the current run
      static inline int maxThreads() { return 1;  }

      //! \brief return number of current threads
      static inline int currentThreads() { return 1; }

      //! \brief return thread number
      static inline int thread() {  return 0; }

      //! \brief return true if the current thread is the master thread (i.e. thread 0)
      static inline bool isMaster() { return true ; }

      //! \brief set maximal number of threads available during run
      static inline void setMaxNumberThreads( const int numThreads ) { }

      //! \brief returns true if program is operating on one thread currently
      static inline bool singleThreadMode() { return true ; }
    }; // end class ThreadManager
#endif

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_OMPMANAGER_HH
