/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_OPERATOR_PROJECTION_LOCAL_RIESZ_ORTHONORMAL_HH
#define DUNE_FEM_OPERATOR_PROJECTION_LOCAL_RIESZ_ORTHONORMAL_HH

#include <cassert>
#include <cstddef>

#include <utility>

#include <dune/geometry/referenceelements.hh>

#include "localrieszprojection.hh"

namespace Dune
{

  namespace Fem
  {

    // OrthonormalLocalRieszProjection
    // -------------------------------

    template< class BasisFunctionSet >
    class OrthonormalLocalRieszProjection
    : public LocalRieszProjection< BasisFunctionSet, OrthonormalLocalRieszProjection< BasisFunctionSet > >
    {
      typedef OrthonormalLocalRieszProjection< BasisFunctionSet > ThisType;
      typedef LocalRieszProjection< BasisFunctionSet, OrthonormalLocalRieszProjection< BasisFunctionSet > > BaseType;

    public:
      /** \copydoc Dune::Fem::LocalRieszProjection::BasisFunctionSetType */
      typedef typename BaseType::BasisFunctionSetType BasisFunctionSetType;

    private:
      typedef typename BasisFunctionSet::FunctionSpaceType::RangeFieldType RangeFieldType;

    public:
      /** \name Construction
       *  \{
       */

      explicit OrthonormalLocalRieszProjection ( const BasisFunctionSetType &basisFunctionSet )
        : basisFunctionSet_( std::forward< BasisFunctionSetType >( basisFunctionSet ) ),
          factor_( ratio( basisFunctionSet.entity().geometry() ) )
      {}

      explicit OrthonormalLocalRieszProjection ( BasisFunctionSetType &&basisFunctionSet )
        : basisFunctionSet_( std::forward< BasisFunctionSetType >( basisFunctionSet ) ),
          factor_( ratio( basisFunctionSet.entity().geometry() ) )
      {}

      /** \} */

      /** \name Copying and assignment
       *  \{
       */

      OrthonormalLocalRieszProjection ( const ThisType & ) = default;

      OrthonormalLocalRieszProjection ( ThisType &&other )
        : basisFunctionSet_( std::move( other.basisFunctionSet_ ) ),
          factor_( other.factor_ )
      {}

      ThisType &operator= ( const ThisType & ) = default;

      ThisType &operator= ( ThisType &&other )
      {
        basisFunctionSet_ = std::move( other.basisFunctionSet_ );
        factor_( other.factor_ );
        return *this;
      }

      /** \} */

      /** \name Public member methods
       *  \{
       */

      /** \copydoc Dune::Fem::LocalRieszProjection::basisFunctionSet */
      BasisFunctionSetType basisFunctionSet () const
      {
        return basisFunctionSet_;
      }

      /** \copydoc Dune::Fem::LocalRieszProjection::apply */
      template< class F, class LocalDofVector >
      void apply ( const F &f, LocalDofVector &dofs ) const
      {
        assert( f.size() == dofs.size() );
        const std::size_t size = dofs.size();
        for( std::size_t i = 0u; i < size; ++i )
          dofs[ i ] = factor_*f[ i ];
      }

      /** \} */

    private:
      template< class Geometry >
      static RangeFieldType ratio ( const Geometry &geometry )
      {
        assert( geometry.affine() );
        const auto &referenceElement
          = Dune::ReferenceElements< typename Geometry::ctype, Geometry::mydimension >::general( geometry.type() );
        return referenceElement.volume()/geometry.volume();
      }

      BasisFunctionSetType basisFunctionSet_;
      RangeFieldType factor_;
    };

  } // namespace Fem

} // namepsace Dune

#endif // #ifndef DUNE_FEM_OPERATOR_PROJECTION_LOCAL_RIESZ_ORTHONORMAL_HH
