/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_PASS_INSERTOPERATOR_HH
#define DUNE_FEM_PASS_INSERTOPERATOR_HH

#include <string>
#include <tuple>
#include <type_traits>

#include <dune/common/tupleutility.hh>

#include <dune/fem/operator/common/operator.hh>
#include <dune/fem/pass/common/pass.hh>

namespace Dune
{

  namespace Fem
  {

#ifndef DOXYGEN

    namespace __InsertOperatorPass
    {

      // isOperator
      // ----------

      template< class DomainFunction, class RangeFunction >
      std::true_type __isOperator ( const Dune::Fem::Operator< DomainFunction, RangeFunction > & );

      std::false_type __isOperator ( ... );

      template< class Operator >
      using isOperator = decltype( __isOperator( std::declval< Operator >() ) );



      template< class Operator >
      struct DiscreteModel
      {
        struct Traits
        {
          typedef typename Operator::RangeFunctionType DestinationType;
        };
      };

    } // namespace __InsertOperatorPass

#endif // #ifndef DOXYGEN



    // InsertOperatorPass
    // ------------------

    /** \brief include a Dune::Fem::Operator into a pass
     *
     *  \note This pass stores a reference to the operator passed to the
     *  constructor. The operator will always be called on the result of
     *  PreviousPass, so the position of this pass in your pass tree is of
     *  importance!
     *
     *  \tparam  Operator  a Dune::Fem::Operator
     *  \tparam  PreviousPass  type of previous pass
     *  \tparam  id  unique id for this pass
     */
    template< class Operator, class PreviousPass, int id >
    class InsertOperatorPass final
      : public Dune::Fem::Pass< __InsertOperatorPass::DiscreteModel< Operator >, PreviousPass, id >
    {
      typedef Dune::Fem::Pass< __InsertOperatorPass::DiscreteModel< Operator >, PreviousPass, id > BaseType;

      static_assert( __InsertOperatorPass::isOperator< Operator >::value, "Template parameter Operator is not derived from Dune::Fem::Operator" );

    public:
      /** \brief export operator type */
      typedef Operator OperatorType;

      /** \copydoc Dune::Fem::Pass::PreviousPassType */
      typedef typename BaseType::PreviousPassType PreviousPassType;

      /** \copydoc Dune::Fem::Pass::TotalArgumentType */
      typedef typename BaseType::TotalArgumentType TotalArgumentType;
      /** \copydoc Dune::Fem::Pass::DestinationType */
      typedef typename BaseType::DestinationType DestinationType;

      /** \copydoc Dune::Fem::Pass::passNumber */
      using BaseType::passNumber;

    protected:
      using BaseType::deleteHandler_;
      using BaseType::destination_;

    public:
      /** \name Construction
       *  \{
       */

      /** \brief constructor
       *
       *  \param[in]  space  range discrete function space
       *  \param[in]  op  an operator instance
       *  \param[in]  pass  previous pass
       */
      InsertOperatorPass ( const typename OperatorType::RangeFunctionType::DiscreteFunctionSpaceType &space,
                           const OperatorType &op, PreviousPassType &pass )
        : BaseType( pass ),
          space_( space ),
          operator_( op )
      {}

      /** \} */

#ifndef DOXYGEN

      ~InsertOperatorPass () {}

#endif // #ifndef DOXYGEN

      /** \name Public member methods
       *  \{
       */

      /** \copydoc Dune::Fem::Pass::allocateLocalMemory */
      void allocateLocalMemory ()
      {
        if( destination_ )
          return;
        destination_ = new DestinationType( "stabilization_pass_" + std::to_string( passNumber() ), space() );
        deleteHandler_ = &(BaseType::DeleteHandlerType::instance());
      }

      /** \copydoc Dune::Fem::Pass::printTexInfo */
      void printTexInfo ( std::ostream & ) const
      {
        DUNE_THROW( NotImplemented, "Method printTexInfo() not implemented" );
      }

      /** \copydoc Dune::Fem::Pass::compute */
      void compute ( const TotalArgumentType &arg, DestinationType &dest ) const
      {
        apply( arg, dest );
      }

      /** \} */

      /** \name Non-interface methods
       *  \{
       */

      /** \brief return discrete function space */
      const typename OperatorType::RangeFunctionType::DiscreteFunctionSpaceType &space () const
      {
        return space_;
      }

      /** \} */

    protected:
      void apply ( const TotalArgumentType &u, DestinationType &w ) const
      {
        // get argument
        typedef std::integral_constant< int, PreviousPassType::passId > Key;
        const std::size_t position = Dune::FirstTypeIndex< typename PreviousPassType::PassIds, Key >::type::value;
        const typename OperatorType::DomainFunctionType &v = *std::get< position >( u );

        // apply operator
        operator_( v, w );
      }

    private:
      const typename OperatorType::RangeFunctionType::DiscreteFunctionSpaceType &space_;
      const OperatorType &operator_;
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_PASS_INSERTOPERATOR_HH
