/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_TWISTUTILITY_HH
#define DUNE_FEM_TWISTUTILITY_HH

#include <cassert>

#include <dune/common/version.hh>
#include <dune/geometry/type.hh>
#include <dune/geometry/genericgeometry/topologytypes.hh>

#include <dune/grid/utility/hostgridaccess.hh>

// this also includes the forward declarations
#include <dune/fem/misc/capabilities.hh>
#include <dune/fem/misc/compatibility.hh>

namespace Dune
{

  namespace Fem
  {

    /** \brief TwistFreeTwistUtility provides the default implementation for twistfree grid
         such as Cartesian grids.
    */
    template< class Grid >
    struct TwistFreeTwistUtility
    {
      typedef Grid GridType;

      //! \brief return 0 for inner face
      template< class Intersection >
      static int twistInSelf ( const GridType &, const Intersection & )
      {
        return 0;
      }

      //! \brief return 0 for outer face
      template< class Intersection >
      static int twistInNeighbor ( const GridType &, const Intersection & )
      {
        return 0;
      }

      /** \brief return geometry type of inside or outside entity */
      template< class Intersection >
      static GeometryType elementGeometry ( const Intersection &intersection, const bool inside )
      {
        typedef Dune::Capabilities::hasSingleGeometryType< GridType > hasSingleGeometryType;
        if( hasSingleGeometryType::v )
          return GeometryType( hasSingleGeometryType::topologyId, GridType::dimension );
        else
          return inside ? make_entity( intersection.inside() ).type() : make_entity( intersection.outside() ).type();
      }
    };



    /** \brief Utility to get twist from IntersectionIterator,
        if provided by grid (i.e. AlbertaGrid, ALUGrid)
        otherwise return default values (correct for YASP/SGRID).

        The twist (t) of a face is defined in the following way:
        - sign(t) gives information on the relationship between the
          orientation of the intersection geometry and the geometry
          of the corresponding codim 1 entity of the inside/outside
          entity:
          - sign(t)>=0: same orientation
          - sign(t)<0:  opposite orientation

        - The value of the twist gives information on the local numbering
          of the corners of the corresponding geometries. This value
          is only correctly defined for conforming grids, i.e.,
          the intersection is identical to an codim 1 entity of inside/outside.
          In this case we have the following definition:
          - sign(t)>=0: corner[0] of inside/outside face is equal to
                        corner[t] of intersection.
          - sign(t)<0:  corner[0] of inside/outside face is equal to
                        corner[t'] of intersection with t' = abs(t)+1.

        \note This class needs to be explicitly specialized for each grid!
    */
    template< class Grid >
    struct TwistUtility;
    //{
    //  static_assert( false, "TwistUtility not specialized, please specialize TwistUtility!");
    //};

    // Specialization for YaspGrid
    // ------------------------------

    template< int dimw, class CoordCont >
    struct TwistUtility< YaspGrid< dimw, CoordCont > >
      : public TwistFreeTwistUtility< YaspGrid< dimw, CoordCont > >
    {
    };


    // Specialization for SGrid
    // ------------------------------

    template< int dim, int dimworld, class ctype >
    struct TwistUtility< SGrid<dim, dimworld, ctype> >
      : public TwistFreeTwistUtility< SGrid<dim, dimworld, ctype> >
    {
    };


    // Specialization for OneDGrid
    // ------------------------------

    template <>
    struct TwistUtility< OneDGrid >
      : public TwistFreeTwistUtility< OneDGrid >
    {
    };


    // Specialization for AlbertaGrid
    // ------------------------------

#if HAVE_ALBERTA
    /** \brief Specialization of TwistUtility for AlbertaGrid.
    */
    template< int dim, int dimW >
    struct TwistUtility< AlbertaGrid< dim, dimW > >
    {
      typedef AlbertaGrid<dim, dimW> GridType;
      typedef typename GridType::Traits::LeafIntersectionIterator  LeafIntersectionIterator;
      typedef typename LeafIntersectionIterator::Intersection  LeafIntersection;
      typedef typename GridType::Traits::LevelIntersectionIterator LevelIntersectionIterator;
      typedef typename LevelIntersectionIterator::Intersection LevelIntersection;

      static const int dimension = GridType::dimension;

    public:
      //! \brief return twist for inner face
      static int twistInSelf ( const GridType &grid, const LeafIntersection &it )
      {
        return grid.getTwistInInside( it );
      }

      //! \brief return twist for outer face
      static int twistInNeighbor ( const GridType &grid, const LeafIntersection &it )
      {
        return grid.getTwistInOutside( it );
      }

      /** \brief return element geometry type of inside or outside entity
      */
      template <class Intersection>
      static inline GeometryType
      elementGeometry(const Intersection& intersection,
                      const bool inside)
      {
        return GeometryType( GenericGeometry::SimplexTopology< dimension >::type::id,
                             dimension );
      }
    };
#endif



    // Specialization for ALUGrid
    // --------------------------

#if HAVE_ALUGRID
    /** \brief Specialization of TwistUtility for ALUGridSimplex.
    */
    template< int dim, int dimw >
    struct TwistUtility< ALUSimplexGrid< dim, dimw > >
    {
      typedef ALUSimplexGrid< dim, dimw > GridType;

      typedef typename GridType::Traits::LeafIntersectionIterator
        LeafIntersectionIterator;
      typedef typename LeafIntersectionIterator::Intersection LeafIntersection;

      typedef typename GridType::Traits::LevelIntersectionIterator
        LevelIntersectionIterator;
      typedef typename LevelIntersectionIterator::Intersection LevelIntersection;

      static const int dimension = GridType::dimension;

    public:
      //! \brief return twist for inner face
      template <class Intersection>
      static inline int twistInSelf(const GridType & grid,
                                    const Intersection& intersection)
      {
        return grid.getRealIntersection( intersection ).twistInInside();
      }

      //! \brief return twist for outer face
      template <class Intersection>
      static inline int twistInNeighbor(const GridType & grid,
                                        const Intersection& intersection)
      {
        return grid.getRealIntersection( intersection ).twistInOutside();
      }

      /** \brief return element geometry type of inside or outside entity
      */
      template <class Intersection>
      static inline GeometryType
      elementGeometry(const Intersection& intersection,
                      const bool inside)
      {
        return GeometryType( GenericGeometry::SimplexTopology< dim >::type::id,
                             dim );
      }
    private:
      TwistUtility(const TwistUtility&);
      TwistUtility& operator=(const TwistUtility&);
    };

    /** \brief Specialization of TwistUtility for ALUSimplexGrid.
    */
    template< int dim, int dimw >
    struct TwistUtility< ALUCubeGrid< dim, dimw > >
    {
      typedef ALUCubeGrid< dim, dimw > GridType;
      typedef typename GridType::Traits::LeafIntersectionIterator LeafIntersectionIterator;
      typedef typename LeafIntersectionIterator::Intersection LeafIntersection;
      typedef typename GridType::Traits::LevelIntersectionIterator LevelIntersectionIterator;
      typedef typename LevelIntersectionIterator::Intersection LevelIntersection;

    public:
      //! \brief return twist for inner face
      template <class Intersection>
      static inline int twistInSelf(const GridType & grid,
                                    const Intersection& intersection)
      {
        return grid.getRealIntersection( intersection ).twistInInside();
      }

      //! \brief return twist for outer face
      template <class Intersection>
      static inline int twistInNeighbor(const GridType & grid,
                                        const Intersection& intersection)
      {
        return grid.getRealIntersection( intersection ).twistInOutside();
      }

      /** \brief return element geometry type of inside or outside entity
      */
      template <class Intersection>
      static inline GeometryType
      elementGeometry(const Intersection& intersection,
                      const bool inside)
      {
        return GeometryType( GenericGeometry::CubeTopology< dim >::type::id,
                             dim );
      }

    private:
      TwistUtility(const TwistUtility&);
      TwistUtility& operator=(const TwistUtility&);
    };

    /** \brief Specialization of TwistUtility for ALUConformGrid.
    */
    template< int dim, int dimw >
    struct TwistUtility< ALUConformGrid< dim, dimw > >
    {
      typedef ALUConformGrid< dim, dimw > GridType;
      typedef typename GridType::Traits::LeafIntersectionIterator LeafIntersectionIterator;
      typedef typename LeafIntersectionIterator::Intersection LeafIntersection;
      typedef typename GridType::Traits::LevelIntersectionIterator LevelIntersectionIterator;
      typedef typename LevelIntersectionIterator::Intersection LevelIntersection;

    public:
      //! \brief return twist for inner face
      static inline int twistInSelf(const GridType & grid, const LeafIntersection& intersection)
      {
        return grid.getRealIntersection( intersection ).twistInInside();
      }

      //! \brief return twist for outer face
      static inline int twistInNeighbor(const GridType &grid, const LeafIntersection& intersection )
      {
        return grid.getRealIntersection( intersection ).twistInOutside();
      }

      /** \brief return element geometry type of inside or outside entity
      */
      template <class Intersection>
      static inline GeometryType
      elementGeometry(const Intersection& intersection,
                      const bool inside)
      {
        return GeometryType( GenericGeometry::SimplexTopology< dim >::type::id,
                             dim );
      }

    private:
      TwistUtility(const TwistUtility&);
      TwistUtility& operator=(const TwistUtility&);
    };
#endif // #if HAVE_ALUGRID

#if HAVE_ALUGRID || HAVE_DUNE_ALUGRID
    /** \brief Specialization of TwistUtility for ALUGrid.
    */
    template< int dim, int dimw, ALUGridElementType elType, ALUGridRefinementType refineType, class Comm >
    struct TwistUtility< ALUGrid< dim, dimw, elType, refineType, Comm > >
    {
      typedef ALUGrid< dim, dimw, elType, refineType, Comm > GridType;
      typedef typename GridType::Traits::LeafIntersectionIterator LeafIntersectionIterator;
      typedef typename LeafIntersectionIterator::Intersection LeafIntersection;
      typedef typename GridType::Traits::LevelIntersectionIterator LevelIntersectionIterator;
      typedef typename LevelIntersectionIterator::Intersection LevelIntersection;

    public:
      //! \brief return twist for inner face
      static inline int twistInSelf(const GridType & grid, const LeafIntersection& intersection)
      {
        return (dim == 2) ? std::abs(grid.getRealIntersection( intersection ).twistInInside()) : grid.getRealIntersection( intersection ).twistInInside();
      }

      //! \brief return twist for outer face
      static inline int twistInNeighbor(const GridType &grid, const LeafIntersection& intersection )
      {
        return (dim == 2) ? std::abs(grid.getRealIntersection( intersection ).twistInOutside()) : grid.getRealIntersection( intersection ).twistInOutside();
      }

      /** \brief return element geometry type of inside or outside entity
      */
      template <class Intersection>
      static inline GeometryType
      elementGeometry(const Intersection& intersection,
                      const bool inside)
      {
        return GeometryType( Dune::Capabilities::hasSingleGeometryType< GridType > :: topologyId,
                             dim );
      }

    private:
      TwistUtility(const TwistUtility&);
      TwistUtility& operator=(const TwistUtility&);
    };
#endif // #if HAVE_ALUGRID || HAVE_DUNE_ALUGRID


    // Specialization for UGGrid
    // -------------------------

#if HAVE_UG
    template< int dim >
    struct TwistUtility< UGGrid< dim > >
    {
      typedef UGGrid< dim > GridType;

      typedef typename GridType::Traits::LeafIntersectionIterator LeafIntersectionIterator;
      typedef typename LeafIntersectionIterator::Intersection LeafIntersection;
      typedef typename GridType::Traits::LevelIntersectionIterator LevelIntersectionIterator;
      typedef typename LevelIntersectionIterator::Intersection LevelIntersection;

      static int twistInSelf ( const GridType &grid, const LeafIntersection &it );
      static int twistInSelf ( const GridType &grid, const LevelIntersection &it );

      static int twistInNeighbor ( const GridType &grid, const LeafIntersection &it );
      static int twistInNeighbor ( const GridType &grid, const LevelIntersection &it );

      template< class Intersection >
      static GeometryType
      elementGeometry ( const Intersection &intersection, const bool inside )
      {
        return (inside ? intersection.inside().type() : intersection.outside().type());
      }
    };
#endif // #ifdef ENABLE_UG



    // Specialization for GeometryGrid
    // -------------------------------

    template< class HostGrid, class CoordFunction, class Allocator >
    struct TwistUtility< GeometryGrid< HostGrid, CoordFunction, Allocator > >
    {
      typedef GeometryGrid< HostGrid, CoordFunction, Allocator > GridType;
      typedef typename GridType::Traits::LeafIntersectionIterator LeafIntersectionIterator;
      typedef typename LeafIntersectionIterator::Intersection LeafIntersection;
      typedef typename GridType::Traits::LevelIntersectionIterator LevelIntersectionIterator;
      typedef typename LevelIntersectionIterator::Intersection LevelIntersection;

    private:
      typedef TwistUtility< HostGrid > HostTwistUtility;
      typedef Dune::HostGridAccess< GridType > HostGridAccess;

    public:
      //! \brief return twist for inner face
      template< class Intersection >
      static int twistInSelf ( const GridType &grid, const Intersection &intersection )
      {
        return HostTwistUtility::twistInSelf( grid.hostGrid(), HostGridAccess::hostIntersection( intersection ) );
      }

      //! \brief return twist for outer face
      template< class Intersection >
      static int twistInNeighbor ( const GridType &grid, const Intersection &intersection )
      {
        return HostTwistUtility::twistInNeighbor( grid.hostGrid(), HostGridAccess::hostIntersection( intersection ) );
      }

      /** \brief return element geometry type of inside or outside entity */
      template< class Intersection >
      static GeometryType elementGeometry ( const Intersection &intersection, bool inside )
      {
        return HostTwistUtility::elementGeometry( HostGridAccess::hostIntersection( intersection ), inside );
      }
    };

  }  // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_TWISTUTILITY_HH
