/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_CORNERPOINTSET_HH
#define DUNE_FEM_CORNERPOINTSET_HH

#include <dune/geometry/genericgeometry/referencedomain.hh>

#include <dune/fem/quadrature/cachingpointlist.hh>

namespace Dune
{

  namespace Fem
  {

    // Internal Forward Declarations
    // -----------------------------

    template< class GridPart >
    class CornerPointSet;

    template< class ct, class Topology >
    class CornerPointList;



    // CornerPointSetTraits
    // --------------------

    template< class GridPart >
    class CornerPointSetTraits
    {
      template< class ct, int dim >
      struct PointListTraits;

    public:
      typedef IntegrationPointList< typename GridPart::ctype, GridPart::dimension, PointListTraits >
        IntegrationPointListType;
      typedef typename IntegrationPointListType::CoordinateType CoordinateType;
    };



    // CornerPointSetTraits::PointListTraits
    // -------------------------------------

    template< class GridPart >
    template< class ct, int dim >
    class CornerPointSetTraits< GridPart >::PointListTraits
    {
      static const int pdim = (dim > 0 ? dim : 1);

    public:
      typedef IntegrationPointListImp< ct, dim > IntegrationPointListType;

      typedef CornerPointList< ct, typename GenericGeometry::SimplexTopology< dim >::type > SimplexQuadratureType;
      typedef CornerPointList< ct, typename GenericGeometry::CubeTopology< dim >::type > CubeQuadratureType;
      typedef CornerPointList< ct, typename GenericGeometry::PrismTopology< pdim >::type > PrismQuadratureType;
      typedef CornerPointList< ct, typename GenericGeometry::PyramidTopology< pdim >::type > PyramidQuadratureType;

      typedef SimplexQuadratureType PointQuadratureType;
      typedef SimplexQuadratureType LineQuadratureType;
    };



    // CornerPointSet
    // --------------

    template< class GridPart >
    class CornerPointSet
    : public CachingPointList< GridPart, 0, CornerPointSetTraits< GridPart > >
    {
      typedef CachingPointList< GridPart, 0, CornerPointSetTraits< GridPart > > BaseType;

    public:
      CornerPointSet ( const GeometryType &type )
      : BaseType( type, 1 )
      {}

      CornerPointSet ( const typename GridPart::template Codim< 0 >::EntityType &entity )
      : BaseType( entity.type(), 1 )
      {}
    };



    // CornerPointList
    // ---------------

    template< class ct, class Topology >
    class CornerPointList
    : public IntegrationPointListImp< ct, Topology::dimension >
    {
      typedef IntegrationPointListImp< ct, Topology::dimension > BaseType;

      typedef GenericGeometry::ReferenceDomain< Topology > ReferenceDomain;

    public:
      typedef typename BaseType::CoordinateType CoordinateType;

      explicit CornerPointList ( const size_t id );
      CornerPointList ( const GeometryType &type, const int order, const size_t id );

      int order () const { return 1; }

      static unsigned int maxOrder () { return 1; }

      GeometryType geometryType () const { return GeometryType( Topology() ); }

    protected:
      using BaseType::addIntegrationPoint;

    private:
      void initialize ();
    };



    // Implementation of CornerPointList
    // ---------------------------------

    template< class ct, class Topology >
    inline CornerPointList< ct, Topology >::CornerPointList ( const size_t id )
    : BaseType( id )
    {
      initialize();
    }


    template< class ct, class Topology >
    inline CornerPointList< ct, Topology >
      ::CornerPointList ( const GeometryType &type, const int order, const size_t id )
    : BaseType( id )
    {
      initialize();
    }


    template< class ct, class Topology >
    inline void CornerPointList< ct, Topology >::initialize ()
    {
      for( unsigned int i = 0; i < ReferenceDomain::numCorners; ++i )
      {
        CoordinateType pt;
        ReferenceDomain::corner( i, pt );
        addIntegrationPoint( pt );
      }
    }

  } //namespace Fem

} //namespace Dune

#endif // #ifndef DUNE_FEM_CORNERPOINTSET_HH
