/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_DEFAULTQUADRATURES_HH
#define DUNE_FEM_DEFAULTQUADRATURES_HH

//#include <vector>
#include <cassert>

#include <dune/common/fvector.hh>
#include <dune/common/exceptions.hh>
#include <dune/grid/common/grid.hh>

#include <dune/fem/storage/array.hh>
#include <dune/fem/quadrature/idprovider.hh>

// don't use quadratures from dune-grid
//#define USE_DUNE_QUADRATURES

// include quadrature points
#ifdef USE_DUNE_QUADRATURES
#warning "Don't use DUNE Quadratures!!!"
#include "dunequadratures.hh"
#else
#include "femquadratures.hh"
#endif

namespace Dune
{

  namespace Fem
  {

    // default defines for used quadratures
    template< typename FieldType, int dim >
    struct DefaultQuadratureTraits
    {
#ifdef USE_DUNE_QUADRATURES
      typedef QuadratureRulesFactory< FieldType, dim > CubeQuadratureType;
#else
      typedef CubeQuadrature< FieldType, dim > CubeQuadratureType;
#endif
      typedef QuadratureImp< FieldType, dim > IntegrationPointListType;
    };



    // quadratures for points
    template< typename FieldType >
    struct DefaultQuadratureTraits< FieldType, 0 >
    {
#ifdef USE_DUNE_QUADRATURES
      typedef QuadratureRulesFactory< FieldType, 0 > PointQuadratureType;
#else
      typedef CubeQuadrature< FieldType, 0 > PointQuadratureType;
#endif
      typedef QuadratureImp< FieldType, 0 > IntegrationPointListType;
    };



    // quadratures for lines
    template< typename FieldType >
    struct DefaultQuadratureTraits< FieldType, 1 >
    {
#ifdef USE_DUNE_QUADRATURES
      typedef QuadratureRulesFactory< FieldType, 1 > LineQuadratureType;
#else
      typedef CubeQuadrature< FieldType, 1 > LineQuadratureType;
#endif
      typedef QuadratureImp< FieldType, 1 > IntegrationPointListType;
    };



    // quadratures for simplex and cubes
    template< typename FieldType >
    struct DefaultQuadratureTraits< FieldType, 2 >
    {
#ifdef USE_DUNE_QUADRATURES
      typedef QuadratureRulesFactory< FieldType, 2 > SimplexQuadratureType;
      typedef QuadratureRulesFactory< FieldType, 2 > CubeQuadratureType;
#else
      typedef CubeQuadrature< FieldType, 2 > CubeQuadratureType;
      typedef SimplexQuadrature< FieldType, 2 > SimplexQuadratureType;
#endif
      typedef QuadratureImp< FieldType, 2 > IntegrationPointListType;
    };



    // quadratures for simplex, cubes, prisms, and pyramids
    template< typename FieldType >
    struct DefaultQuadratureTraits< FieldType , 3 >
    {
#ifdef USE_DUNE_QUADRATURES
      typedef QuadratureRulesFactory< FieldType, 3 > SimplexQuadratureType;
      typedef QuadratureRulesFactory< FieldType, 3 > CubeQuadratureType;

      typedef QuadratureRulesFactory< FieldType, 3 > PrismQuadratureType;
      typedef QuadratureRulesFactory< FieldType, 3 > PyramidQuadratureType;
#else
      typedef CubeQuadrature< FieldType, 3 > CubeQuadratureType;
      typedef SimplexQuadrature< FieldType, 3 > SimplexQuadratureType;

      typedef PrismQuadrature< FieldType > PrismQuadratureType;
      typedef PyramidQuadrature< FieldType > PyramidQuadratureType;
#endif

      typedef QuadratureImp< FieldType, 3 > IntegrationPointListType;
    };

  } // namespace Fem

} // namespace Dune

#undef USE_DUNE_QUADRATURES
#endif // #ifndef DUNE_FEM_DEFAULTQUADRATURES_HH
