/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef CGHS_BLAS_H
#define CGHS_BLAS_H

// ============================================================================
//
//  CG nach Hestenes und Stiefel
//
//  siehe auch:
//  Ashby, Manteuffel, Saylor
//     A taxononmy for conjugate gradient methods
//     SIAM J Numer Anal 27, 1542-1568 (1990)
//
//  oder:
//  Willy D"orfler:
//     Orthogonale Fehlermethoden
//
//                                                 ----------------------------
//                                                 Christian Badura, Mai 1998
//
// Modification for parallel use: Robert Kloefkorn
//
// ============================================================================

#include <utility>

// ============================================================================
#include "cblas.h"
#include "tmpmem.hh"

static OEMTmpMem cghsMem;

// CommunicatorType has to fulfill interface of CollectiveCommunication
// see dune-common/common/collectivecommunication.hh
template< bool usePC,
          class CommunicatorType, 
          class MATRIX , 
          class PC_MATRIX >
inline 
std::pair < int , double > 
cghs_algo( const CommunicatorType & comm, 
      unsigned int N, const MATRIX &A, const PC_MATRIX& C,
      const double *b, double *x, double eps, int maxIter,
      bool detailed ) 
{
  if ( N==0 )
  {
    std::cerr << "WARNING: N = 0 in cghs, file: " << __FILE__ << " line:" << __LINE__ << "\n";
    return std::pair<int,double> (-1,0.0);
  }

  // define type of multiplication 
  typedef Mult<MATRIX,PC_MATRIX,usePC> MultType;

  double* tmp = 0;

#ifdef USE_MEMPROVIDER
  cghsMem.resize( (usePC) ? 4*N : 3*N ); 

  double *p = cghsMem.getMem (N);
  double *r = cghsMem.getMem (N);
  double *g = cghsMem.getMem (N);
  if( usePC ) tmp = cghsMem.getMem(N);
#else 
  double *g = new double[N];
  double *r = new double[N];
  double *p = new double[N];
  if( usePC ) tmp = new double [N];
#endif
  
  for(size_t k =0 ; k<N; ++k) 
  {
    g[k] = 0.0;
    r[k] = 0.0;
    p[k] = 0.0;
  }
  
  int its=0;
  double t, gam;
  double rho, sig, tau;
  
  double err= eps * eps * MultType :: ddot(A,b,b);
  
  // apply first multiplication 
  MultType :: mult_pc(A,C,x,g,tmp);
    
  daxpy(N,-1.,b,1,g,1);
  dscal(N,-1.,g,1);
  dcopy(N,g,1,r,1);

  double gg = MultType :: ddot(A,g,g);

  while ( gg > err && its < maxIter ) 
  {
    // apply multiplication 
    MultType :: mult_pc(A,C,r,p,tmp);

    rho = MultType :: ddot(A,p,p);
    sig = MultType :: ddot(A,r,p);
    tau = MultType :: ddot(A,g,r);
    
    t=tau/sig;
    daxpy(N,t,r,1,x,1);
    
    daxpy(N,-t,p,1,g,1);
    gam=(t*t*rho-tau)/tau;
    dscal(N,gam,r,1);
    daxpy(N,1.,g,1,r,1);
    
    gg = MultType :: ddot(A,g,g);
    
    if ( detailed && (comm.rank() == 0) )
    {
      std::cout<<"cghs "<<its<<"\t"<<gg<< std::endl;
    }
    ++its;
  }

  // if right preconditioning then do back solve 
  MultType :: back_solve(N,C,x,tmp);

  std::pair<int,double> val (its,sqrt(gg));
  
#ifdef USE_MEMPROVIDER
  cghsMem.reset();
#else 
  delete[] g;
  delete[] r;
  delete[] p;
  if( tmp ) delete [] tmp;
#endif
  
  return val;
}

// cghs with preconditioning 
template<class CommunicatorType, class MATRIX, class PC_MATRIX >
inline
std::pair < int , double > 
cghs( const CommunicatorType & comm,
      unsigned int N, const MATRIX &A, const PC_MATRIX &C,
      const double *b, double *x, double eps, int maxIter, 
      bool detailed ) 
{
  return cghs_algo<true> (comm,N,A,C,b,x,eps,maxIter,detailed);
}

// cghs without preconditioning 
template<class CommunicatorType, class MATRIX >
inline
std::pair < int , double > 
cghs( const CommunicatorType & comm,
      unsigned int N, const MATRIX &A,
      const double *b, double *x, double eps, 
      int maxIter, bool detailed ) 
{
  return cghs_algo<false> (comm,N,A,A,b,x,eps,maxIter,detailed);
}

#endif // CGHS_BLAS_H
