/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_SOLVER_RUNGEKUTTA_BUTCHERTABLE_HH
#define DUNE_FEM_SOLVER_RUNGEKUTTA_BUTCHERTABLE_HH

//- dune-common includes
#include <dune/common/dynmatrix.hh>
#include <dune/common/dynvector.hh>

namespace DuneODE
{

  // SimpleButcherTable
  // ------------------

  template< class Field >
  class SimpleButcherTable
  {
    typedef SimpleButcherTable< Field > This;

  public:
    typedef Field FieldType;

    SimpleButcherTable ( int stages, int order, const FieldType *a, const FieldType *b, const FieldType *c )
    : stages_( stages ), order_( order ),
      a_( a ), b_( b ), c_( c )
    {}

    Dune::DynamicMatrix< FieldType > A () const { return makeMatrix( stages_, stages_, a_ ); }
    Dune::DynamicVector< FieldType > b () const { return makeVector( stages_, b_ ); }
    Dune::DynamicVector< FieldType > c () const { return makeVector( stages_, c_ ); }

    int order () const { return order_; }
    int stages () const { return stages_; }

  protected:
    static Dune::DynamicMatrix< FieldType > makeMatrix ( int m, int n, const FieldType *data )
    {
      Dune::DynamicMatrix< FieldType > A( m, n );
      for( int i = 0; i < m; ++i )
        std::copy( data + i*n, data + (i+1)*n, A[ i ].begin() );
      return A;
    }

    static Dune::DynamicVector< FieldType > makeVector ( int n, const FieldType *data )
    {
      Dune::DynamicVector< FieldType > v( n );
      std::copy( data, data + n, v.begin() );
      return v;
    }

    int stages_, order_;
    const FieldType *a_, *b_, *c_;
  };



  // implicit butcher tables
  // -----------------------

  SimpleButcherTable< double > implicit34ButcherTable ();
  SimpleButcherTable< double > implicit3ButcherTable ();
  SimpleButcherTable< double > implicitEulerButcherTable ();
  SimpleButcherTable< double > gauss2ButcherTable ();



  // semiimplicit butcher tables
  // ---------------------------

  SimpleButcherTable< double > semiImplicitEulerButcherTable ( bool expl );
  SimpleButcherTable< double > semiImplicit23ButcherTable ( bool expl );
  SimpleButcherTable< double > semiImplicit33ButcherTable ( bool expl );
  SimpleButcherTable< double > semiImplicitSSP222ButcherTable ( bool expl );
  SimpleButcherTable< double > semiImplicitARK34ButcherTable ( bool expl );
  SimpleButcherTable< double > semiImplicitARK46ButcherTable ( bool expl );
  SimpleButcherTable< double > semiImplicitIERK45ButcherTable ( bool expl );

  // ROW butcher tables
  // -----------------------

  template< class Field >
  class ROWSimpleButcherTable : public SimpleButcherTable<Field>
  {
    typedef ROWSimpleButcherTable< Field > This;
    typedef SimpleButcherTable< Field > Base;

  public:
    typedef Field FieldType;

    ROWSimpleButcherTable ( int stages, int order, const FieldType *a, const FieldType *b, const FieldType *c, const FieldType *a2 )
    : Base(stages,order,a,b,c)
    , a2_(a2)
    {}

    Dune::DynamicMatrix< FieldType > B () const { return Base::makeMatrix( stages_, stages_, a2_ ); }

  private:
    using Base::stages_;
    const FieldType *a2_;
  };
  ROWSimpleButcherTable< double > row2ButcherTable ();
  ROWSimpleButcherTable< double > row3ButcherTable ();
} // namespace DuneODE

#endif // #ifndef DUNE_FEM_SOLVER_RUNGEKUTTA_BUTCHERTABLE_HH
