/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#include <config.h>

#include <chrono>
#include <iostream>
#include <random>

#include <dune/common/exceptions.hh>

#include <dune/fem/misc/mpimanager.hh>
#include <dune/fem/solver/timeprovider.hh>

int main ( int argc, char **argv )
try
{
  // initialize MPI
  Dune::Fem::MPIManager::initialize( argc, argv );

  // start and end time for time loop and maximum time step
  const double start = 0., end = 1., maxStep = 0.2;

  // test fixed step time provider
  Dune::Fem::FixedStepTimeProvider<> fixedStepTimeProvider(start,maxStep);
  for( ; fixedStepTimeProvider.time() < end; fixedStepTimeProvider.next())
  {
    // print some info
    std::cerr << "FixedStepTimeProvider<>: time step = " << fixedStepTimeProvider.timeStep() << ", "
              << "time = " << fixedStepTimeProvider.time() << ", "
              << "dt = " << fixedStepTimeProvider.deltaT() << std::endl;

  }
  // create random number generator
  const unsigned int seed = std::chrono::system_clock::now().time_since_epoch().count();
  std::default_random_engine engine( seed );
  std::uniform_real_distribution< double > distribution( start, end );

  // function object returning a random time step estimate
  auto timeStepEstimate = [&distribution, &engine]()
  {
    return distribution( engine );
  };

  // create time provider and provide initial time step estimate
  Dune::Fem::TimeProvider<> timeProvider( start );
  timeProvider.provideTimeStepEstimate( timeStepEstimate() );

  for( timeProvider.init(); timeProvider.time() < end; timeProvider.next() )
  {
    // provide next estimate
    timeProvider.provideTimeStepEstimate( timeStepEstimate() );

    // invalidate time step if estimate was chosen too large and restart
    if( timeProvider.deltaT() > maxStep )
    {
      timeProvider.invalidateTimeStep();
      continue;
    }

    // print some info
    std::cerr << "TimeProvider<>: time step = " << timeProvider.timeStep() << ", "
              << "time = " << timeProvider.time() << ", "
              << "dt = " << timeProvider.deltaT() << std::endl;
  }

  return 0;
}
catch( Dune::Exception &exception )
{
  std::cerr << exception << std::endl;
  return 1;
}
catch( std::exception &exception )
{
  std::cerr << exception.what() << std::endl;
  return 1;
}
catch( ... )
{
  std::cerr << "Unknown exception thrown" << std::endl;
  return 1;
}
