/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_SPACE_BASISFUNCTIONSET_TRANSFORMATION_HH
#define DUNE_FEM_SPACE_BASISFUNCTIONSET_TRANSFORMATION_HH

#include <dune/common/exceptions.hh>
#include <dune/common/fmatrix.hh>
#include <dune/common/fvector.hh>

#include <dune/fem/common/fmatrixcol.hh>

namespace Dune
{

  namespace Fem
  {

    // jacobianTransformation
    // ----------------------

    template< class GeometryJacobianInverseTransposed, class K, int ROWS >
    void jacobianTransformation ( const GeometryJacobianInverseTransposed &gjit,
                                  const FieldMatrix< K, ROWS, GeometryJacobianInverseTransposed::cols > &a,
                                  FieldMatrix< K, ROWS, GeometryJacobianInverseTransposed::rows > &b )
    {
      for( int r = 0; r < ROWS; ++r )
        gjit.mv( a[ r ], b[ r ] );
    }



    // JacobianTransformation
    // ----------------------

    template< class Geometry >
    struct JacobianTransformation
    {
      typedef typename Geometry::LocalCoordinate LocalCoordinate;
      typedef typename Geometry::JacobianInverseTransposed GeometryJacobianInverseTransposed;

      JacobianTransformation ( const Geometry &geometry, const LocalCoordinate &x )
      : gjit_( geometry.jacobianInverseTransposed( x ) )
      {}

      template< class A, class B >
      void operator() ( const A &a, B &b ) const
      {
        jacobianTransformation( gjit_, a, b );
      }

    private:
      GeometryJacobianInverseTransposed gjit_;
    };



    // hessianTransformation
    // ---------------------

    template< class GeometryJacobianInverseTransposed, class K, int SIZE >
    void hessianTransformation ( const GeometryJacobianInverseTransposed &gjit,
                                 const FieldVector< FieldMatrix< K, GeometryJacobianInverseTransposed::cols, GeometryJacobianInverseTransposed::cols >, SIZE > &a,
                                 FieldVector< FieldMatrix< K, GeometryJacobianInverseTransposed::rows, GeometryJacobianInverseTransposed::rows >, SIZE > &b )
    {
      const int dimLocal = GeometryJacobianInverseTransposed::cols;
      const int dimGlobal = GeometryJacobianInverseTransposed::rows;

      for( int r = 0; r < SIZE; ++r )
      {
        // c = J^{-T} a_r^T
        FieldMatrix< K, dimLocal, dimGlobal > c;
        for( int i = 0; i < dimLocal; ++i )
          gjit.mv( a[ r ][ i ], c[ i ] );

        // b_r = J^{-T} c
        for( int i = 0; i < dimGlobal; ++i )
        {
          FieldMatrixColumn< const FieldMatrix< K, dimLocal, dimGlobal > > ci( c, i );
          FieldMatrixColumn< FieldMatrix< K, dimGlobal, dimGlobal > > bi( b[ r ], i );
          gjit.mv( ci, bi );
        }
#if 0
        b[ r ] = K( 0 );
        for( int k = 0; k < GeometryJacobianInverseTransposed::cols; ++k )
        {
          FieldVector< K, GeometryJacobianInverseTransposed::rows > c;
          gjit.mv( a[ r ][ k ], c );
          for( int j = 0; j < GeometryJacobianInverseTransposed::rows; ++j )
            b[ r ][ j ].axpy( gjit[ j ][ k ], c );
        }
#endif
      }
    }



    // HessianTransformation
    // ---------------------

    template< class Geometry >
    struct HessianTransformation
    {
      typedef typename Geometry::LocalCoordinate LocalCoordinate;
      typedef typename Geometry::JacobianInverseTransposed GeometryJacobianInverseTransposed;

      HessianTransformation ( const Geometry &geometry, const LocalCoordinate &x )
      : gjit_( geometry.jacobianInverseTransposed( x ) )
      {
        if( !geometry.affine() )
          DUNE_THROW( NotImplemented, "HessianTransformation not implemented for non-affine geometries." );
      }

      template< class A, class B >
      void operator() ( const A &a, B &b ) const
      {
        hessianTransformation( gjit_, a, b );
      }

    private:
      GeometryJacobianInverseTransposed gjit_;
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_SPACE_BASISFUNCTIONSET_TRANSFORMATION_HH
