/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_SPACE_COMBINEDSPACE_COMBINEDSPACE_HH
#define DUNE_FEM_SPACE_COMBINEDSPACE_COMBINEDSPACE_HH

#include <map>
#include <type_traits>
#include <vector>

#include <dune/fem/version.hh>

#include <dune/fem/space/combinedspace/powerspace.hh>
#include <dune/fem/space/common/discretefunctionspace.hh>
#include <dune/fem/space/common/dofstorage.hh>
#include <dune/fem/space/common/localrestrictprolong.hh>


namespace Dune
{

  namespace Fem
  {

    /** @addtogroup CombinedSpace
        Class to combine N scalar discrete function spaces.
        Policies PointBased and VariableBased decide, how dof are stored in
        vectors. PointBased stores all local dofs consecutive,
        VectorBased stores all dofs for one component consecutive.
       @{

     */

    template< class DiscreteFunctionSpace, int N, DofStoragePolicy policy >
    class CombinedSpace;

    /** @brief
        Combined Space Function Space
     **/

    template< class DiscreteFunctionSpace, int N >
    class CombinedSpace< DiscreteFunctionSpace, N, PointBased >
      : public DiscreteFunctionSpace::template ToNewDimRange< DiscreteFunctionSpace::dimRange *N >::Type
    {
      typedef CombinedSpace< DiscreteFunctionSpace, N, PointBased > ThisType;
      typedef typename DiscreteFunctionSpace::template ToNewDimRange< DiscreteFunctionSpace::dimRange *N >::Type BaseType;

      static const DofStoragePolicy policy = PointBased;

    public:
      typedef typename BaseType::GridPartType GridPartType;
      typedef DiscreteFunctionSpace ContainedDiscreteFunctionSpaceType;

      CombinedSpace ( GridPartType &gridPart,
                      const InterfaceType commInterface = InteriorBorder_All_Interface,
                      const CommunicationDirection commDirection = ForwardCommunication )
        : BaseType( gridPart, commInterface, commDirection ),
          containedSpace_( gridPart, commInterface, commDirection )
      {}

      // prohibit copy constructor and copy assignment opertor
      CombinedSpace ( const ThisType & ) = delete;
      ThisType &operator= ( const ThisType & ) = delete;

      //- Additional methods
      //! number of components
      int numComponents () const { return N; }

      //! policy of this space
      DofStoragePolicy myPolicy () const
      {
        return policy;
      }

      //! contained space
      const ContainedDiscreteFunctionSpaceType &containedSpace () const
      {
        return containedSpace_;
      }

    private:
      ContainedDiscreteFunctionSpaceType containedSpace_;
    };


    template< class DiscreteFunctionSpace, int N >
    class CombinedSpace< DiscreteFunctionSpace, N, VariableBased >
      : public PowerDiscreteFunctionSpace< DiscreteFunctionSpace, N >
    {
      typedef CombinedSpace< DiscreteFunctionSpace, N, VariableBased > ThisType;
      typedef PowerDiscreteFunctionSpace< DiscreteFunctionSpace, N > BaseType;

      static const DofStoragePolicy policy = VariableBased;

    public:
      typedef typename BaseType::GridPartType GridPartType;
      typedef typename BaseType::ContainedDiscreteFunctionSpaceType ContainedDiscreteFunctionSpaceType;

      CombinedSpace ( GridPartType &gridPart,
                      const InterfaceType commInterface = InteriorBorder_All_Interface,
                      const CommunicationDirection commDirection = ForwardCommunication )
        : BaseType( gridPart, commInterface, commDirection )
      {}

      // prohibit copy constructor and copy assignment opertor
      CombinedSpace ( const ThisType & ) = delete;
      ThisType &operator= ( const ThisType & ) = delete;

      //- Additional methods
      //! number of components
      int numComponents () const { return N; }

      //! policy of this space
      DofStoragePolicy myPolicy () const
      {
        return policy;
      }

      //! contained space
      const ContainedDiscreteFunctionSpaceType &containedSpace () const
      {
        return BaseType::containedSpace();
      }
    };



    // specialization of DifferentDiscreteFunctionSpace for this CombinedSapce
    template< class ContainedSpace, int N, DofStoragePolicy policy, class NewFunctionSpace >
    struct DifferentDiscreteFunctionSpace< CombinedSpace< ContainedSpace, N, policy >, NewFunctionSpace >
    {
      typedef CombinedSpace< ContainedSpace, NewFunctionSpace::dimRange, policy > Type;
    };


    // DefaultLocalRestrictProlong ( specialization for CombinedSpace< DiscreteFunctionSpace, N, VariableBased > )
    template< class DiscreteFunctionSpace, int N >
    class DefaultLocalRestrictProlong< CombinedSpace< DiscreteFunctionSpace, N, PointBased > >
      : public DefaultLocalRestrictProlong< typename DiscreteFunctionSpace::template ToNewDimRange< DiscreteFunctionSpace::dimRange *N >::Type >
    {
      typedef DefaultLocalRestrictProlong< CombinedSpace< DiscreteFunctionSpace, N, PointBased > > ThisType;
      typedef DefaultLocalRestrictProlong< typename DiscreteFunctionSpace::template ToNewDimRange< DiscreteFunctionSpace::dimRange *N >::Type > BaseType;

    public:
      DefaultLocalRestrictProlong ( const CombinedSpace< DiscreteFunctionSpace, N, PointBased > &space )
        : BaseType( space )
      {}
    };


    // DefaultLocalRestrictProlong ( specialization for CombinedSpace< DiscreteFunctionSpace, N, VariableBased > )
    template< class DiscreteFunctionSpace, int N >
    class DefaultLocalRestrictProlong< CombinedSpace< DiscreteFunctionSpace, N, VariableBased > >
      : public PowerLocalRestrictProlong< DiscreteFunctionSpace, N >
    {
      typedef DefaultLocalRestrictProlong< CombinedSpace< DiscreteFunctionSpace, N, VariableBased > > ThisType;
      typedef PowerLocalRestrictProlong< DiscreteFunctionSpace, N > BaseType;

    public:
      DefaultLocalRestrictProlong ( const CombinedSpace< DiscreteFunctionSpace, N, VariableBased > &space )
        : BaseType( space.containedSpace() )
      {}
    };

    /** @} **/

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_SPACE_COMBINEDSPACE_COMBINEDSPACE_HH
