/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_SPACE_COMBINEDSPACE_POWERLOCALRESTRICPROLONG_HH
#define DUNE_FEM_SPACE_COMBINEDSPACE_POWERLOCALRESTRICPROLONG_HH

#include <algorithm>

#include <dune/common/exceptions.hh>

#include <dune/fem/common/subvector.hh>
#include <dune/fem/function/localfunction/const.hh>
#include <dune/fem/function/localfunction/localfunction.hh>

#include <dune/fem/space/common/localrestrictprolong.hh>


namespace Dune
{

  namespace Fem
  {

    // PowerLocalRestricProlong
    // ------------------------

    template< class DiscreteFunctionSpace, int N >
    class PowerLocalRestrictProlong
    {
      typedef PowerLocalRestrictProlong< DiscreteFunctionSpace, N > ThisType;

      // type of contained DefaultLocalRestrictProlong
      typedef DefaultLocalRestrictProlong< DiscreteFunctionSpace > LocalRestrictProlongType;

    public:
      // type of DomainField
      typedef typename LocalRestrictProlongType::DomainFieldType DomainFieldType;

      PowerLocalRestrictProlong ( const DiscreteFunctionSpace &space )
        : localRestrictProlong_( space )
      {}

      void setFatherChildWeight ( const DomainFieldType &weight )
      {
        localRestrictProlong_.setFatherChildWeight( weight );
      }

      //! restrict data to father
      template< class LFFather, class LFSon, class LocalGeometry >
      void restrictLocal ( LFFather &lfFather, const LFSon &lfSon,
                           const LocalGeometry &geometryInFather, bool initialize ) const
      {
        typedef DenseSubVector< const typename LFSon::LocalDofVectorType > SubDofVectorTypeSon;
        typedef DenseSubVector< typename LFFather::LocalDofVectorType > SubDofVectorTypeFather;

        typedef typename LFSon::BasisFunctionSetType::ScalarBasisFunctionSetType SubSonBasisFunctionSetType;
        typedef typename LFFather::BasisFunctionSetType::ScalarBasisFunctionSetType SubFatherBasisFunctionSetType;

        SubFatherBasisFunctionSetType subFatherBasisFunctionSet = lfFather.basisFunctionSet().scalarBasisFunctionSet();
        SubSonBasisFunctionSetType subSonBasisFunctionSet = lfSon.basisFunctionSet().scalarBasisFunctionSet();

        for( std::size_t i = 0; i < N; ++i )
        {
          std::size_t sonBasisSetSize = subSonBasisFunctionSet.size();
          std::size_t fatherBasisSetsize = subFatherBasisFunctionSet.size();

          SubDofVectorTypeSon sonSubDofVector( lfSon.localDofVector(), sonBasisSetSize, sonBasisSetSize * i );
          SubDofVectorTypeFather fatherSubDofVector( lfFather.localDofVector(), fatherBasisSetsize, fatherBasisSetsize * i );

          BasicConstLocalFunction< SubSonBasisFunctionSetType, SubDofVectorTypeSon >
          subLFSon( subSonBasisFunctionSet, sonSubDofVector );
          LocalFunction< SubFatherBasisFunctionSetType, SubDofVectorTypeFather >
          subLFFather( subFatherBasisFunctionSet, fatherSubDofVector );

          localRestrictProlong_.restrictLocal( subLFFather, subLFSon, geometryInFather, initialize );
        }
      }


      template< class LFFather, class LFSon, class LocalGeometry >
      void prolongLocal ( const LFFather &lfFather, LFSon &lfSon,
                          const LocalGeometry &geometryInFather, bool initialize ) const
      {
        typedef DenseSubVector< typename LFSon::LocalDofVectorType > SubDofVectorTypeSon;
        typedef DenseSubVector< const typename LFFather::LocalDofVectorType > SubDofVectorTypeFather;

        typedef typename LFSon::BasisFunctionSetType::ScalarBasisFunctionSetType SubSonBasisFunctionSetType;
        typedef typename LFFather::BasisFunctionSetType::ScalarBasisFunctionSetType SubFatherBasisFunctionSetType;

        SubSonBasisFunctionSetType subSonBasisFunctionSet = lfSon.basisFunctionSet().scalarBasisFunctionSet();
        SubFatherBasisFunctionSetType subFatherBasisFunctionSet = lfFather.basisFunctionSet().scalarBasisFunctionSet();

        for( std::size_t i = 0; i < N; ++i )
        {
          std::size_t sonBasisSetSize = subSonBasisFunctionSet.size();
          std::size_t fatherBasisSetsize = subFatherBasisFunctionSet.size();

          SubDofVectorTypeSon sonSubDofVector( lfSon.localDofVector(), sonBasisSetSize, sonBasisSetSize * i );
          SubDofVectorTypeFather fatherSubDofVector( lfFather.localDofVector(), fatherBasisSetsize, fatherBasisSetsize * i );

          LocalFunction< SubSonBasisFunctionSetType, SubDofVectorTypeSon >
          subLFSon( subSonBasisFunctionSet, sonSubDofVector );
          BasicConstLocalFunction< SubFatherBasisFunctionSetType, SubDofVectorTypeFather >
          subLFFather( subFatherBasisFunctionSet, fatherSubDofVector );

          localRestrictProlong_.prolongLocal( subLFFather, subLFSon, geometryInFather, initialize );
        }
      }

      bool needCommunication () const
      {
        return localRestrictProlong_.needCommunication();
      }

    private:
      LocalRestrictProlongType localRestrictProlong_;
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_SPACE_COMBINEDSPACE_POWERLOCALRESTRICPROLONG_HH
