/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_SPACE_COMBINEDSPACE_POWERSPACE_HH
#define DUNE_FEM_SPACE_COMBINEDSPACE_POWERSPACE_HH

#include <algorithm>
#include <type_traits>

#include <dune/common/math.hh>
#include <dune/common/typetraits.hh>
#include <dune/grid/common/grid.hh>

#include <dune/fem/common/utility.hh>

#include <dune/fem/space/basisfunctionset/vectorial.hh>
#include <dune/fem/space/combinedspace/generic.hh>
#include <dune/fem/space/combinedspace/lagrangepointsetexporter.hh>
#include <dune/fem/space/combinedspace/powerlocalrestrictprolong.hh>
#include <dune/fem/space/combinedspace/powermapper.hh>
#include <dune/fem/space/common/defaultcommhandler.hh>

namespace Dune
{

  namespace Fem
  {

    // forward declaration

    template< class DiscreteFunctionSpace, int N >
    class PowerDiscreteFunctionSpace;


    // PowerDiscreteFunctionSpaceTraits
    // --------------------------------

    template< class DiscreteFunctionSpace, int N >
    struct PowerDiscreteFunctionSpaceTraits
    {
      // we need to store pointer to the spaces in the SpaceTuple, since space can not be copied.
      typedef DiscreteFunctionSpace *DiscreteFunctionSpaceTupleType;

      // helper struct to access contained sub spaces
      template< int >
      struct SubDiscreteFunctionSpace
      {
        // type of i-th sub space
        typedef DiscreteFunctionSpace Type;

        // type of i-th sub BlockMapper
        typedef typename Type::BlockMapperType BlockMapperType;

        // access to a const ref of the i-th subspace
        static const Type &subDiscreteFunctionSpace ( const DiscreteFunctionSpaceTupleType &tuple )
        {
          return (*tuple);
        }

        static BlockMapperType &subBlockMapper ( const DiscreteFunctionSpaceTupleType &tuple )
        {
          return subDiscreteFunctionSpace().blockMapper();
        }
      };

      // type of GridPart
      typedef typename DiscreteFunctionSpace::GridPartType GridPartType;
      typedef typename GridPartType::GridType GridType;
      typedef typename GridPartType::IndexSetType IndexSetType;
      typedef typename GridPartType::template Codim< 0 >::IteratorType IteratorType;
      typedef typename IteratorType::Entity EntityType;
      typedef typename GridPartType::IntersectionType IntersectionType;

      // type of this space
      typedef PowerDiscreteFunctionSpace< DiscreteFunctionSpace, N > DiscreteFunctionSpaceType;

      static const int codimension = DiscreteFunctionSpace::Traits::codimension;

      static_assert( DiscreteFunctionSpace::FunctionSpaceType::dimRange == 1,
                     "PowerDiscreteFunctionSpace only works for ContainedSpaces with dimRange = 1" );
      //! implementation of basefunction set
      typedef typename DiscreteFunctionSpace::BasisFunctionSetType ScalarBasisFunctionSetType;
      typedef typename ScalarBasisFunctionSetType::FunctionSpaceType::RangeFieldType RangeFieldType;
      typedef VectorialBasisFunctionSet< ScalarBasisFunctionSetType, FieldVector< RangeFieldType, N >, VerticalDofAlignment > BasisFunctionSetType;

      // type of block mapper
      typedef PowerMapper< GridPartType, typename DiscreteFunctionSpace::BlockMapperType, N > BlockMapperType;

      // in the most general case we will unroll all local blockings
      enum { localBlockSize = DiscreteFunctionSpace::localBlockSize };

      // type functionspace
      typedef typename BasisFunctionSetType::FunctionSpaceType FunctionSpaceType;

      static constexpr int polynomialOrder = DiscreteFunctionSpace::polynomialOrder;

      // review to make it work for all kind of combinations
      template< class DiscreteFunction,
                class Operation =
                  typename DiscreteFunctionSpace::template CommDataHandle< DiscreteFunctionSpace >::OperationType >
      struct CommDataHandle
      {
        //! type of data handle
        typedef typename DiscreteFunctionSpace::
          template CommDataHandle< DiscreteFunction, Operation >::Type Type;
        //! type of operatation to perform on scatter
        typedef typename DiscreteFunctionSpace::
          template CommDataHandle< DiscreteFunction, Operation >::OperationType OperationType;
      };

      // construct new instance of blockMapper
      static BlockMapperType *getBlockMapper ( const DiscreteFunctionSpaceTupleType &spaceTuple )
      {
        return new BlockMapperType( spaceTuple->gridPart(), spaceTuple->blockMapper() );
      }

      // delete instance of BlockMapper
      static void deleteBlockMapper ( BlockMapperType *blockMapper )
      {
        delete blockMapper;
      }

      // create Tuple of contained subspaces
      static DiscreteFunctionSpaceTupleType createSpaces ( GridPartType &gridPart, InterfaceType commInterface,
                                                           CommunicationDirection commDirection )
      {
        return new DiscreteFunctionSpace( gridPart, commInterface, commDirection );
      }

      // delete Tuple of contained subspaces
      static void deleteSpaces ( DiscreteFunctionSpaceTupleType &tuple )
      {
        delete tuple;
      }

      template< class Entity >
      static BasisFunctionSetType getBasisFunctionSet ( const Entity &entity, const DiscreteFunctionSpaceTupleType &tuple )
      {
        return BasisFunctionSetType( tuple->basisFunctionSet( entity ) );
      }

      static bool continuous ( const DiscreteFunctionSpaceTupleType &tuple )
      {
        return tuple->continuous();
      }

      static bool continuous ( const IntersectionType &intersection, const DiscreteFunctionSpaceTupleType &tuple )
      {
        return tuple->continuous( intersection );
      }
    };



    /** \addtogroup DiscreteFunctionSpace
     *
     *  Provides a DiscreteFunctionSpace combined from arbitrary number of DiscreteFunctionSpaces
     *  of different types into a single \ref Dune::Fem::DiscreteFunctionSpaceInterface ( U_h times V_h times .... ).
     *
     *  \note It is assumed that the each space is build upon the same gridpart
     */

    /** \class   DiscreteFunctionSpace
     *  \ingroup DiscreteFunctionSpace
     *  \brief    discrete function space
     */
    template< class DiscreteFunctionSpace, int N >
    class PowerDiscreteFunctionSpace
      : public GenericCombinedDiscreteFunctionSpace< PowerDiscreteFunctionSpaceTraits< DiscreteFunctionSpace, N > >,
        public CombinedSpaceHelper::LagrangePointSetExporter< DiscreteFunctionSpace >
    {
      typedef PowerDiscreteFunctionSpace< DiscreteFunctionSpace, N > ThisType;
      typedef GenericCombinedDiscreteFunctionSpace< PowerDiscreteFunctionSpaceTraits< DiscreteFunctionSpace, N > > BaseType;
      typedef CombinedSpaceHelper::LagrangePointSetExporter< DiscreteFunctionSpace > LagrangePointSetExporterType;

    public:
      typedef PowerDiscreteFunctionSpaceTraits< DiscreteFunctionSpace, N > Traits;
      //! extract grid informations, it is assumed the both spaces are living on the
      //! same gridPart
      typedef typename Traits::GridPartType GridPartType;

      //! type of contained discrete function space
      typedef DiscreteFunctionSpace ContainedDiscreteFunctionSpaceType;

      /** \brief constructor
       *
       *  \param[in]  gridPart       grid part for the Lagrange space
       *  \param[in]  commInterface  communication interface to use (optional)
       *  \param[in]  commDirection  communication direction to use (optional)
       */
      PowerDiscreteFunctionSpace ( GridPartType &gridPart,
                                   const InterfaceType commInterface = InteriorBorder_All_Interface,
                                   const CommunicationDirection commDirection = ForwardCommunication )
        : BaseType( gridPart, commInterface, commDirection ),
          LagrangePointSetExporterType( containedSpace() )
      {}

      // prohibit copy constructor and copy assignment
      PowerDiscreteFunctionSpace ( const ThisType & ) = delete;
      ThisType &operator= ( const ThisType & ) = delete;

      //! return tuple of const References to the contained sub spaces
      const ContainedDiscreteFunctionSpaceType &containedSpace () const
      {
        return BaseType::template subDiscreteFunctionSpace< 0 >();
      }
    };


    //! specialization of DifferentDiscreteFunctionSpace for PowerDiscreteFunctionSpace
    template< class DiscreteFunctionSpace, int  N, class NewFunctionSpace >
    struct DifferentDiscreteFunctionSpace< PowerDiscreteFunctionSpace< DiscreteFunctionSpace, N >, NewFunctionSpace >
    {
      typedef PowerDiscreteFunctionSpace< DiscreteFunctionSpace, NewFunctionSpace::dimRange > Type;
    };


    // DefaultLocalRestrictProlong
    // ---------------------------

    template< class DiscreteFunctionSpace, int N >
    class DefaultLocalRestrictProlong< PowerDiscreteFunctionSpace< DiscreteFunctionSpace, N > >
      : public PowerLocalRestrictProlong< DiscreteFunctionSpace, N >
    {
      typedef DefaultLocalRestrictProlong< PowerDiscreteFunctionSpace< DiscreteFunctionSpace, N > > ThisType;
      typedef PowerDiscreteFunctionSpace< DiscreteFunctionSpace, N > DiscreteFunctionSpacesType;
      typedef PowerLocalRestrictProlong< DiscreteFunctionSpace, N > BaseType;

    public:
      DefaultLocalRestrictProlong ( const DiscreteFunctionSpacesType &space )
        : BaseType( space.containedSpace() )
      {}

    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_SPACE_COMBINEDSPACE_POWERSPACE_HH
