/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_SPACE_COMMON_CAPABILITIES_HH
#define DUNE_FEM_SPACE_COMMON_CAPABILITIES_HH


namespace Dune
{

  namespace Fem
  {

    namespace Capabilities
    {

      /** \class hasFixedPolynomialOrder
       *
       *  \brief specialize with \b true if polynomial order does
       *         not depend on the grid (part) entity
       */
      template< class DiscreteFunctionSpace >
      struct hasFixedPolynomialOrder
      {
        static const bool v = false;
      };



      /** \class hasStaticPolynomialOrder
       *
       *  \brief specialize with \b true if polynomial order fixed
       *         and compile time static
       */
      template< class DiscreteFunctionSpace >
      struct hasStaticPolynomialOrder
      {
        static const bool v = false;
        static const int order = -1;
      };



      /** \class isContinuous
       *
       *  \brief specialize with \b true if space is always continuous
       */
      template< class DiscreteFunctionSpace >
      struct isContinuous
      {
        static const bool v = false;
      };



      /** \class isLocalized
       *
       *  \brief specialize with \b true if the space is localized, *
       *  i.e., the basis function set is based on a shape function set.
       *
       *  We require, that a localized space has a method
\code
  ShapeFunctionSetType shapeFunctionSet ( const EntityType &entity );
\endcode
       */
      template< class DiscreteFunctionSpace >
      struct isLocalized
      {
        static const bool v = false;
      };



      /** \class isParallel
       *
       *  \brief specialize with \b true if space can be used in parallel
       */
      template< class DiscreteFunctionSpace >
      struct isParallel
      {
        static const bool v = false;
      };



      /** \class isAdaptive
       *
       *  \brief specialize with \b true if space can be used with
       *         AdaptiveDiscreteFunction
       */
      template< class DiscreteFunctionSpace >
      struct isAdaptive
      {
        static const bool v = false;
      };



      /** \class threadSafe
       *
       *  \brief specialize with \b true if the space implementation
       *         is thread safe
       */
      template< class DiscreteFunctionSpace >
      struct threadSafe
      {
        static const bool v = false;
      };



      /** \class viewThreadSafe
       *
       * \brief specialize with \b true if the space implementation is
       *        thread safe, while it is not modified
       *
       */
      template< class DiscreteFunctionSpace >
      struct viewThreadSafe
      {
        static const bool v = false;
      };


      /** \class isHierarchic
       *
       *  \brief specialize with \b true if for a space the basis functions are
       *         sorted by the polynomial order, starting with the lowest order
       */
      template< class DiscreteFunctionSpace >
      struct isHierarchic
      {
        static const bool v = false;
      };



      // const specialization
      // --------------------

      template< class DiscreteFunctionSpace >
      struct hasFixedPolynomialOrder< const DiscreteFunctionSpace >
      {
        static const bool v = hasFixedPolynomialOrder< DiscreteFunctionSpace >::v;
      };

      template< class DiscreteFunctionSpace >
      struct hasStaticPolynomialOrder< const DiscreteFunctionSpace >
      {
        static const bool v = hasStaticPolynomialOrder< DiscreteFunctionSpace >::v;
        static const int order = hasStaticPolynomialOrder< DiscreteFunctionSpace >::order;
      };

      template< class DiscreteFunctionSpace >
      struct isContinuous < const DiscreteFunctionSpace >
      {
        static const bool v = isContinuous< DiscreteFunctionSpace >::v;
      };

      template< class DiscreteFunctionSpace >
      struct isLocalized< const DiscreteFunctionSpace >
      {
        static const bool v = isLocalized< DiscreteFunctionSpace >::v;
      };

      template< class DiscreteFunctionSpace >
      struct isParallel< const DiscreteFunctionSpace >
      {
        static const bool v = isParallel< DiscreteFunctionSpace >::v;
      };

      template< class DiscreteFunctionSpace >
      struct isAdaptive< const DiscreteFunctionSpace >
      {
        static const bool v = isAdaptive< DiscreteFunctionSpace >::v;
      };

      template< class DiscreteFunctionSpace >
      struct threadSafe< const DiscreteFunctionSpace >
      {
        static const bool v = threadSafe< DiscreteFunctionSpace >::v;
      };

      template< class DiscreteFunctionSpace >
      struct viewThreadSafe< const DiscreteFunctionSpace >
      {
        static const bool v = viewThreadSafe< DiscreteFunctionSpace >::v;
      };

      template< class DiscreteFunctionSpace >
      struct isHierarchic< const DiscreteFunctionSpace >
      {
        static const bool v = isHierarchic< DiscreteFunctionSpace >::v;
      };

    } // namespace Capabilities

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_SPACE_COMMON_CAPABILITIES_HH
