/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_FUNCTIONSPACEINTERFACE_HH
#define DUNE_FEM_FUNCTIONSPACEINTERFACE_HH

#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>

namespace Dune
{

  namespace Fem
  {

    /** \addtogroup FunctionSpace
     *
     *  This provides the interfaces for analytical function spaces.
     *  A function space is characterized by it's domain and range field type
     *  and the two finite dimensional vector spaces over these fields.
     *
     *  All corresponding types are given in a Traits class. In addition to the
     *  domain and range types also types for the Jacobian and the Hessian must
     *  be provieded.
     *
     *  \remarks The interface for analytical functions spaces is defined by the
     *           class FunctionSpaceInterface.
     */



    /** \class FunctionSpaceInterface
     *  \brief interface for an arbitrary function space
     *  \ingroup FunctionSpace
     *
     *  Base class for specific function spaces.
     *
     *  \interfaceclass
     */
    template< typename FunctionSpaceTraits >
    class FunctionSpaceInterface
    {
    public:
      /** \brief Dimensions of domain and range */
      enum
      {
        //! dimension of domain vector space
        dimDomain = FunctionSpaceTraits :: dimDomain,
        //! dimension of range vector space
        dimRange  = FunctionSpaceTraits :: dimRange
      };

      // for compatibility with GrapDataDisplay (see dune-grid), we also export
      // the following two values:
      enum
      {
        DimDomain = dimDomain,
        DimRange = dimRange
      };

      /** \brief Intrinsic type used for values in the domain field (usually a double) */
      typedef typename FunctionSpaceTraits::DomainFieldType DomainFieldType;

      /** \brief Intrinsic type used for values in the range field (usually a double) */
      typedef typename FunctionSpaceTraits::RangeFieldType RangeFieldType;

      /** \brief Type of domain vector (using type of domain field)
          has a Dune::FieldVector type interface */
      typedef typename FunctionSpaceTraits::DomainType DomainType;

      /** \brief Type of range vector (using type of range field)
          has a Dune::FieldVector type interface */
      typedef typename FunctionSpaceTraits::RangeType RangeType;

      /** \brief Intrinsic type used for the jacobian values
          has a Dune::FieldMatrix type interface */
      typedef typename FunctionSpaceTraits::LinearMappingType JacobianRangeType;

      /** \brief Intrinsic type used for the hessian values
          has a Dune::FieldMatrix type interface */
      typedef FieldVector< FieldMatrix< RangeFieldType, dimDomain, dimDomain >, dimRange > HessianRangeType;

      /** \brief corresponding scalar function space */
      typedef typename FunctionSpaceTraits :: ScalarFunctionSpaceType
        ScalarFunctionSpaceType;
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_FUNCTIONSPACEINTERFACE_HH
