/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_SPACE_INTERPOLATION_HH
#define DUNE_FEM_SPACE_INTERPOLATION_HH

#include <dune/fem/misc/iteratorprovider.hh>

#warning "<dune/fem/space/common/interpolation.hh> is deprecated. Use <dune/fem/space/common/interpolate.hh> instead."
#warning "Replace 'Dune::Fem::Interpolation< DiscreteFunction >::apply( u, v )' by 'interpolate( u, v )'."

namespace Dune
{

  namespace Fem
  {

    // Interpolation
    // -------------

    /** \class   Interpolation
        \ingroup DiscreteFunctionSpace
        \brief   native interpolation of a discrete function space

        By definition of its degrees of freedom, each discrete function space
        has a native interpolation, which can be computed very quickly.

        For example, the native interpolation of a Lagrange discrete function
        space is the evaluation in its Lagrange points.
        An orthonormal DG space would instead perform an \f$L^2\f$-Projection.

        The actual implementation must locally be provided by the discrete
        function space through the method
        \code
template< class LocalFunction, class LocalDofVector >
void interpolate ( const LocalFunction &f, LocalDofVector &dofs ) const;
        \endcode

        \tparam  DiscreteFunction  type of discrete function to interpolate to
     */
    template< class DiscreteFunction,
              class IteratorProvider = Fem::IteratorProvider< typename DiscreteFunction::DiscreteFunctionSpaceType > >
    struct Interpolation
    {
      typedef DiscreteFunction DiscreteFunctionType;
      typedef typename DiscreteFunctionType::DiscreteFunctionSpaceType DiscreteFunctionSpaceType;

      /** \brief interpolate a grid function

          \param[in]   u  grid function to interpolate
          \param[out]  v  discrete function to represent the interpolation
       */
      template< class GridFunction >
      void operator() ( const GridFunction &u, DiscreteFunctionType &v )
      {
        apply( u, v );
      }

      /** \brief interpolate a grid function

          \param[in]   u  grid function to interpolate
          \param[out]  v  discrete function to represent the interpolation
       */
      template< class GridFunction >
      static void apply ( const GridFunction &u, DiscreteFunctionType &v )
      {
        typedef typename IteratorProvider::IteratorType IteratorType;
        typedef typename DiscreteFunctionSpaceType::EntityType EntityType;
        typedef typename DiscreteFunctionType::LocalFunctionType LocalFunctionType;

        const DiscreteFunctionSpaceType &space = v.space();
        IteratorProvider iteratorProvider( space );

        const IteratorType end = iteratorProvider.end();
        for( IteratorType it = iteratorProvider.begin(); it != end; ++it )
        {
          const EntityType &entity = *it;
          const auto interpolation = space.interpolation( entity );

          LocalFunctionType vLocal = v.localFunction( entity );
          interpolation( u.localFunction( entity ), vLocal.localDofVector() );
        }
      }
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_SPACE_INTERPOLATION_HH
