/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_SPACE_DISCONTINUOUSGALERKIN_BASISFUNCTIONSETS_HH
#define DUNE_FEM_SPACE_DISCONTINUOUSGALERKIN_BASISFUNCTIONSETS_HH

#include <utility>

#include <dune/common/documentation.hh>

#include <dune/fem/space/basisfunctionset/default.hh>

namespace Dune
{

  namespace Fem
  {

    // BasisFunctionSets
    // -----------------

    /** \class BasisFunctionSets
     *
     *  \brief interface class representing a family of basis function sets
     */
    class BasisFunctionSets
    {
    public:
      /** \brief basis function set */
      typedef ImplementationDefined BasisFunctionSetType;
      /** \brief entity type */
      typedef ImplementationDefined EntityType;

      /** \name Move construction/assignment
       *  \{
       */

      /** \brief move constructor */
      BasisFunctionSets ( BasisFunctionSets && );

      /** \} */

      /** \name Deleted methods
       *  \{
       */

      /** \brief copy constructor */
      BasisFunctionSets ( const BasisFunctionSets & ) = delete;

      /** \brief assignment constructor */
      BasisFunctionSets &operator= ( const BasisFunctionSets & ) = delete;

      /** \} */

      /** \name Public member methods
       *  \{
       */

      /** \brief return maximum order */
      int order () const;

      /** \brief return order for given grid part entity */
      int order ( const EntityType &entity ) const;

      /** \brief return basis function set for given entity */
      BasisFunctionSetType basisFunctionSet ( const EntityType &entity ) const;

      /** \} */
    };



    // DefaultBasisFunctionSets
    // ------------------------

    /** \class DefaultBasisFunctionSets
     *
     *  \brief generate a set of default basis function sets from given set of
     *         shape function sets
     *
     *  \tparam  GridPart  grid part type
     *  \tparam  ShapeFunctionSets  shape function sets type
     */
    template< class GridPart, class ShapeFunctionSets >
    class DefaultBasisFunctionSets
    {
      typedef DefaultBasisFunctionSets< GridPart, ShapeFunctionSets > ThisType;

    public:
      /** \brief grid part type */
      typedef GridPart GridPartType;

      /** \brief shape function sets type */
      typedef ShapeFunctionSets ShapeFunctionSetsType;
      /** \brief shape function set type */
      typedef typename ShapeFunctionSetsType::ShapeFunctionSetType ShapeFunctionSetType;

    private:
      static const int dimension = GridPartType::dimension;
      static const int mydimension = ShapeFunctionSetType::FunctionSpaceType::dimDomain;
      static const int codimension = dimension - mydimension;

    public:
      /** \copydoc Dune::Fem::BasisFunctionSets::EntityType */
      typedef typename GridPartType::template Codim< codimension >::EntityType EntityType;

      /** \copydoc Dune::Fem::BasisFunctionSets::EntityType */
      typedef Dune::Fem::DefaultBasisFunctionSet< EntityType, ShapeFunctionSetType > BasisFunctionSetType;

    public:
      /** \name Construction
       *  \{
       */

      explicit DefaultBasisFunctionSets ( ShapeFunctionSetsType &&shapeFunctionSets )
        : shapeFunctionSets_( std::forward< ShapeFunctionSetsType >( shapeFunctionSets ) )
      {}

      template< class... Args >
      explicit DefaultBasisFunctionSets ( Args &&... args )
        : shapeFunctionSets_( std::forward< Args >( args )... )
      {}

      /** \} */

      /** \name Copying and assignment
       *  \{
       */

      DefaultBasisFunctionSets ( const ThisType & ) = delete;

      DefaultBasisFunctionSets ( ThisType & ) = delete;

      DefaultBasisFunctionSets ( ThisType &&other )
        : shapeFunctionSets_( std::move( other.shapeFunctionSets_ ) )
      {}

      DefaultBasisFunctionSets &operator= ( const ThisType & ) = delete;

      /** \} */

      /** \name Public member methods
       *  \{
       */

      /** \copydoc Dune::Fem::BasisFunctionSets::order */
      int order () const { return shapeFunctionSets_.order(); }

      /** \copydoc Dune::Fem::BasisFunctionSets::order */
      int order ( const EntityType &entity ) const
      {
        return shapeFunctionSets_.order( entity.type() );
      }

      /** \copydoc Dune::Fem::BasisFunctionSets::basisFunctionSet */
      BasisFunctionSetType basisFunctionSet ( const EntityType &entity ) const
      {
        return BasisFunctionSetType( entity, shapeFunctionSet( entity.type() ) );
      }

      /** \} */

    private:
      ShapeFunctionSetType shapeFunctionSet ( const Dune::GeometryType &type ) const
      {
        return shapeFunctionSets_.shapeFunctionSet( type );
      }

      ShapeFunctionSetsType shapeFunctionSets_;
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_SPACE_DISCONTINUOUSGALERKIN_BASISFUNCTIONSETS_HH
