/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_SPACE_DISCONTINUOUSGALERKIN_INTERPOLATION_HH
#define DUNE_FEM_SPACE_DISCONTINUOUSGALERKIN_INTERPOLATION_HH

#include <type_traits>
#include <utility>

#include <dune/grid/common/capabilities.hh>

#include <dune/fem/quadrature/cachingquadrature.hh>
#include <dune/fem/operator/projection/local/l2projection.hh>
#include <dune/fem/operator/projection/local/riesz.hh>

namespace Dune
{

  namespace Fem
  {

    // Internal forward declaration
    // ----------------------------

    template< class GridPart, class BasisFunctionSet,
              class Quadrature = CachingQuadrature< GridPart, BasisFunctionSet::EntityType::codimension > >
    class DiscontinuousGalerkinLocalL2Projection;



    // DiscontinuousGalerkinLocalL2Projection
    // --------------------------------------

    template< class GridPart, class BasisFunctionSet, class Quadrature >
    class DiscontinuousGalerkinLocalL2Projection
    : public LocalL2Projection< BasisFunctionSet, DiscontinuousGalerkinLocalL2Projection< GridPart, BasisFunctionSet, Quadrature > >
    {
      typedef DiscontinuousGalerkinLocalL2Projection< GridPart, BasisFunctionSet, Quadrature > ThisType;
      typedef LocalL2Projection< BasisFunctionSet, DiscontinuousGalerkinLocalL2Projection< GridPart, BasisFunctionSet, Quadrature > > BaseType;

    public:
      /** \copydoc Dune::Fem::LocalL2Projection::BasisFunctionSetType */
      typedef typename BaseType::BasisFunctionSetType BasisFunctionSetType;

    private:
      typedef GridPart GridPartType;
      typedef typename GridPartType::GridType GridType;
      static const bool cartesian = Dune::Capabilities::isCartesian< GridType >::v;

      typedef typename std::conditional< cartesian,
          OrthonormalLocalRieszProjection< BasisFunctionSetType >,
          DenseLocalRieszProjection< BasisFunctionSetType, Quadrature >
        >::type LocalRieszProjectionType;

      typedef DefaultLocalL2Projection< LocalRieszProjectionType, Quadrature > Implementation;

    public:
      /** \name Construction
       *  \{
       */

      explicit DiscontinuousGalerkinLocalL2Projection ( const BasisFunctionSetType &basisFunctionSet )
        : impl_( LocalRieszProjectionType( basisFunctionSet ) )
      {}

      explicit DiscontinuousGalerkinLocalL2Projection ( BasisFunctionSetType &&basisFunctionSet )
        : impl_( LocalRieszProjectionType( std::forward< BasisFunctionSetType >( basisFunctionSet ) ) )
      {}

      /** \} */

      /** \name Copying and assignment
       *  \{
       */

      DiscontinuousGalerkinLocalL2Projection ( const ThisType & ) = default;

      DiscontinuousGalerkinLocalL2Projection ( ThisType &&other )
        : impl_( std::move( other.impl_ ) )
      {}

      ThisType &operator= ( const ThisType & ) = default;

      ThisType &operator= ( ThisType &&other )
      {
        impl_ = std::move( other.impl_ );
        return *this;
      }

      /** \} */

      /** \name Public member methods
       *  \{
       */

      /** \copydoc Dune::Fem::LocalL2Projection::basisFunctionSet */
      BasisFunctionSet basisFunctionSet () const
      {
        return impl_.basisFunctionSet();
      }

      /** \copydoc Dune::Fem::LocalL2Projection::apply */
      template< class LocalFunction, class LocalDofVector >
      void apply ( const LocalFunction &localFunction, LocalDofVector &localDofVector ) const
      {
        impl_( localFunction, localDofVector );
      }

      /** \} */

    private:
      Implementation impl_;
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_SPACE_DISCONTINUOUSGALERKIN_INTERPOLATION_HH
