/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_SPACE_FINITEVOLUME_INTERPOLATION_HH
#define DUNE_FEM_SPACE_FINITEVOLUME_INTERPOLATION_HH

#include <functional>

#include <dune/fem/function/localfunction/average.hh>

#include "basisfunctionset.hh"

namespace Dune
{

  namespace Fem
  {

    // FiniteVolumeLocalInterpolation
    // ------------------------------

    template< class GridPart, class Range >
    class FiniteVolumeLocalInterpolation
    {
      typedef FiniteVolumeLocalInterpolation< GridPart, Range > ThisType;

    public:
      /** \brief grid part type */
      typedef GridPart GridPartType;
      /** \brief entity type */
      typedef typename GridPartType::template Codim< 0 >::EntityType EntityType;

      /** \brief basis function set type */
      typedef FiniteVolumeBasisFunctionSet< EntityType, Range > BasisFunctionSetType;

      /** \name Construction
       * \{
       */

      explicit FiniteVolumeLocalInterpolation ( const EntityType &entity )
        : entity_( entity )
      {}

      /** \} */

      /** \name Copying and assignment
       * \{
       */

      FiniteVolumeLocalInterpolation ( const ThisType & ) = default;

      FiniteVolumeLocalInterpolation &operator= ( const ThisType & ) = default;

      /** \} */

      /** \name Public member methods
       *  \{
       */

      /** \brief return basis function set */
      BasisFunctionSetType basisFunctionSet () const
      {
        return BasisFunctionSetType( entity() );
      }

      /** \brief interpolate local function */
      template< class LocalFunction, class LocalDofVector >
      void operator() ( const LocalFunction &localFunction, LocalDofVector &localDofVector ) const
      {
        apply( localFunction, localDofVector );
      }

      /** \brief interpolate local function */
      template< class LocalFunction, class LocalDofVector >
      void apply ( const LocalFunction &localFunction, LocalDofVector &localDofVector ) const
      {
        Range value;
        LocalAverage< LocalFunction, GridPartType >::apply( localFunction, value );
        for( int i = 0; i < Range::dimension; ++i )
          localDofVector[ i ] = value[ i ];
      }

      /** \} */

    private:
      const EntityType &entity () const { return entity_.get(); }

      std::reference_wrapper< const EntityType > entity_;
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_SPACE_FINITEVOLUME_INTERPOLATION_HH
