/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_SPACE_LAGRANGE_INTERPOLATION_HH
#define DUNE_FEM_SPACE_LAGRANGE_INTERPOLATION_HH

#include <cstddef>

#include <utility>

#include "lagrangepoints.hh"

namespace Dune
{

  namespace Fem
  {

    // LagrangeLocalInterpolation
    // --------------------------

    template< class GridPart, int order, class BasisFunctionSet >
    class LagrangeLocalInterpolation
    {
      typedef LagrangeLocalInterpolation< GridPart, order, BasisFunctionSet > ThisType;

    public:
      /** \brief basis function set type */
      typedef BasisFunctionSet BasisFunctionSetType;
      /** \brief point set type */
      typedef LagrangePointSet< GridPart, order > LagrangePointSetType;

    private:
      typedef typename BasisFunctionSetType::FunctionSpaceType FunctionSpaceType;

    public:
      /** \name Construction
       *  \{
       */

      LagrangeLocalInterpolation ( const LagrangePointSetType &pointSet,
                                   const BasisFunctionSetType &basisFunctionSet )
        : pointSet_( pointSet ),
          basisFunctionSet_( basisFunctionSet )
      {}

      LagrangeLocalInterpolation ( const LagrangePointSetType &pointSet,
                                   BasisFunctionSetType &&basisFunctionSet )
        : pointSet_( pointSet ),
          basisFunctionSet_( std::forward< BasisFunctionSetType >( basisFunctionSet ) )
      {}

      /** \} */

      /** \name Copying and assignment
       *  \{
       */

      /** \brief copy constructor */
      LagrangeLocalInterpolation ( const ThisType & ) = default;

      /** \brief move constructor */
      LagrangeLocalInterpolation ( ThisType &&other )
        : pointSet_( std::move( other.pointSet_ ) ),
          basisFunctionSet_( std::move( other.basisFunctionSet_ ) )
      {}

      /** \brief assignment operator */
      LagrangeLocalInterpolation &operator= ( const ThisType & ) = default;

      /** \brief move assignment operator */
      LagrangeLocalInterpolation &operator= ( ThisType &&other )
      {
        pointSet_ = std::move( other.pointSet_ );
        basisFunctionSet_ = std::move( other.basisFunctionSet_ );
        return *this;
      }

      /** \} */

      /** \name Public member methods
       *  \{
       */

      /** \brief return basis function set */
      BasisFunctionSetType basisFunctionSet () const
      {
        return basisFunctionSet_;
      }

      /** \brief apply interpolation */
      template< class LocalFunction, class LocalDofVector >
      void operator() ( const LocalFunction &localFunction, LocalDofVector &localDofVector ) const
      {
        apply( localFunction, localDofVector );
      }

      /** \brief apply interpolation */
      template< class LocalFunction, class LocalDofVector >
      void apply ( const LocalFunction &localFunction, LocalDofVector &localDofVector ) const
      {
        const LagrangePointSetType &pointSet = this->pointSet();

        int k = 0;
        const std::size_t nop = pointSet.nop();
        for( std::size_t pt = 0; pt < nop; ++pt )
        {
          typename FunctionSpaceType::RangeType phi;
          localFunction.evaluate( pointSet[ pt ], phi );
          for( int i = 0; i < FunctionSpaceType::dimRange; ++i )
            localDofVector[ k++ ] = phi[ i ];
        }
      }

      /** \} */

    private:
      const LagrangePointSetType &pointSet () const { return pointSet_.get(); }

      std::reference_wrapper< const LagrangePointSetType > pointSet_;
      BasisFunctionSetType basisFunctionSet_;
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_SPACE_LAGRANGE_INTERPOLATION_HH
