/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_SPACE_RANNACHERTUREK_LOCALINTERPOLATION_HH
#define DUNE_FEM_SPACE_RANNACHERTUREK_LOCALINTERPOLATION_HH

#include <cstddef>

#include <utility>
#include <vector>

#include <dune/common/fvector.hh>


namespace Dune
{

  namespace Fem
  {

    // RannacherTurekLocalInterpolation
    // --------------------------------

    template< class BasisFunctionSet, class LocalInterpolation >
    class RannacherTurekLocalInterpolation
    {
      typedef RannacherTurekLocalInterpolation< BasisFunctionSet, LocalInterpolation > ThisType;

    public:
      typedef BasisFunctionSet BasisFunctionSetType;
      typedef LocalInterpolation LocalInterpolationType;

    private:
      typedef typename BasisFunctionSetType::FunctionSpaceType FunctionSpaceType;

      typedef typename FunctionSpaceType::RangeType RangeType;
      typedef typename FunctionSpaceType::RangeFieldType RangeFieldType;
      static const int dimRange = FunctionSpaceType::dimRange;

      typedef std::size_t size_type;

      template< class LocalFunction >
      struct LocalFunctionWrapper;

    public:
      explicit RannacherTurekLocalInterpolation ( const BasisFunctionSetType &basisFunctionSet,
                                                  const LocalInterpolationType &localInterpolation = LocalInterpolationType() )
        : basisFunctionSet_( basisFunctionSet ),
          localInterpolation_( localInterpolation )
      {}

      RannacherTurekLocalInterpolation ( const ThisType & ) = default;

      RannacherTurekLocalInterpolation ( ThisType &&other )
        : basisFunctionSet_( std::move( other.basisFunctionSet_ ) ),
          localInterpolation_( std::move( other.localInterpolation_ ) )
      {}

      RannacherTurekLocalInterpolation &operator= ( const ThisType & ) = default;

      RannacherTurekLocalInterpolation &operator= ( ThisType &&other )
      {
        basisFunctionSet_ = std::move( other.basisFunctionSet_ );
        localInterpolation_ = std::move( other.localInterpolation_ );
        return *this;
      }

      BasisFunctionSetType basisFunctionSet () const
      {
        return basisFunctionSet_;
      }

      template< class LocalFunction, class LocalDofVector >
      void operator() ( const LocalFunction &localFunction, LocalDofVector &localDofVector ) const
      {
        apply( localFunction, localDofVector );
      }

      template< class LocalFunction, class LocalDofVector >
      void apply ( const LocalFunction &localFunction, LocalDofVector &localDofVector ) const;

    protected:
      const LocalInterpolationType &localInterpolation () const { return localInterpolation_; }

    private:
      BasisFunctionSet basisFunctionSet_;
      LocalInterpolationType localInterpolation_;
    };



    // Implementation of RannacherTurekLocalInterpolation::LocalFunctionWrapper
    // ------------------------------------------------------------------------

    template< class LocalInterpolation, class RangeVector >
    template< class LocalFunction >
    struct RannacherTurekLocalInterpolation< LocalInterpolation, RangeVector >::LocalFunctionWrapper
    {
      LocalFunctionWrapper ( const LocalFunction &localFunction, size_type component )
      : localFunction_( localFunction ),
        component_( component )
      {}

      template< class RangeFieldType >
      void evaluate ( const typename LocalFunction::DomainType &x, Dune::FieldVector< RangeFieldType, 1 > &y ) const
      {
        typename LocalFunction::RangeType z;
        localFunction().evaluate( x, z );
        y = z[ component() ];
      }

    private:
      const LocalFunction &localFunction () const { return localFunction_; }

      size_type component () const { return component_; }

      const LocalFunction &localFunction_;
      size_type component_;
    };



    // Implementation of RannacherTurekLocalInterpolation
    // --------------------------------------------------


    template< class LocalInterpolation, class RangeVector >
    template< class LocalFunction, class LocalDofVector >
    inline void RannacherTurekLocalInterpolation< LocalInterpolation, RangeVector >
      ::apply ( const LocalFunction &localFunction, LocalDofVector &localDofVector ) const
    {
      std::vector< RangeFieldType > phi;

      for( int k = 0; k < dimRange; ++k )
      {
        LocalFunctionWrapper< LocalFunction > localFunctionWrapper( localFunction, k );
        localInterpolation().interpolate( localFunctionWrapper, phi );

        const size_type size = phi.size();
        for( size_type i = 0; i < size; ++i )
          localDofVector[ i*dimRange + k ] = phi[ i ];
      }
    }

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_SPACE_RANNACHERTUREK_LOCALINTERPOLATION_HH
