/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_SHAPEFUNCTIONSET_SHAPEFUNCTIONSET_HH
#define DUNE_FEM_SHAPEFUNCTIONSET_SHAPEFUNCTIONSET_HH

//- C++ includes
#include <cstddef>

/**
  @file
  @brief Interface for shape function sets
*/


namespace Dune
{

  namespace Fem
  {

    // ShapeFunctionSet
    // ----------------

    /**
     * \addtogroup ShapeFunction
     * \brief Interface class for shape function sets
     *
     * This class cannot be used itself, it is for documentation purposes
     * only.
     *
     * \note Constructor signatures are explicitly not specified by this
     *       interface.
     */
    template< class FunctionSpace >
    class ShapeFunctionSet
    {
    public:
      //! \brief function space type
      typedef FunctionSpace FunctionSpaceType;

      //! \brief domain type
      typedef typename FunctionSpaceType::DomainType DomainType;
      //! \brief range type
      typedef typename FunctionSpaceType::RangeType RangeType;
      //! \brief jacobian range type
      typedef typename FunctionSpaceType::JacobianRangeType JacobianRangeType;
      //! \brief hessian range type
      typedef typename FunctionSpaceType::HessianRangeType HessianRangeType;

      //! \brief return order of shape functions
      int order () const;

      //! \brief return number of shape functions
      std::size_t size () const;

      /**
         \brief evalute each shape function

          \param[in]  x        coordinate or quadrature point
          \param[in]  functor  functor call for evaluating each shape function

          The functor has to be a copyable object satisfying the following
          interface:
          \code
          struct Functor
          {
            template< class Value >
            void operator() ( const int shapeFunction, const Value &value );
          };
          \endcode
       */
      template< class Point, class Functor >
      void evaluateEach ( const Point &x, Functor functor ) const;

      /**
         \brief evalute jacobian of each shape function

          \param[in]  x        coordinate or quadrature point
          \param[in]  functor  functor call for evaluating the jacobian of each shape function

          The functor has to be a copyable object satisfying the following
          interface:
          \code
          struct Functor
          {
            template< class Jacobian >
            void operator() ( const int shapeFunction, const Jacobian &jacobian );
          };
          \endcode
       */
      template< class Point, class Functor >
      void jacobianEach ( const Point &x, Functor functor ) const;

      /**
         \brief evalute hessian of each shape function

          \param[in]  x        coordinate or quadrature point
          \param[in]  functor  functor call for evaluating the hessian of each shape function

          The functor has to be a copyable object satisfying the following
          interface:
          \code
          struct Functor
          {
            template< class Hessian >
            void operator() ( const int shapeFunction, const Hessian &hessian );
          };
          \endcode
       */
      template< class Point, class Functor >
      void hessianEach ( const Point &x, Functor functor ) const;
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_SHAPEFUNCTIONSET_SHAPEFUNCTIONSET_HH
