/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_STLARRAY_HH
#define DUNE_FEM_STLARRAY_HH

#include <vector>

#include <dune/fem/storage/array.hh>

namespace Dune
{
  namespace Fem
  {

    template< class ElementImp, class RealImp >
    class STLArrayIterator
    {
    public:
      typedef ElementImp ElementType;

    private:
      typedef STLArrayIterator< ElementType, RealImp > ThisType;

    protected:
      RealImp realIterator_;

    public:
      inline explicit STLArrayIterator ( RealImp realIterator )
      : realIterator_( realIterator )
      {
      }

      inline STLArrayIterator(  const ThisType &other )
      : realIterator_( other.realIterator )
      {
      }

      inline ThisType &operator= ( const ThisType &other )
      {
        realIterator_ = other.realIterator_;
      }

      inline ElementType &operator* ()
      {
        return *realIterator_;
      }

      inline ThisType &operator++ ()
      {
        ++realIterator_;
        return *this;
      }

      inline bool operator== ( const ThisType &other ) const
      {
        return realIterator_ == other.realIterator_;
      }

      inline bool operator!= ( const ThisType &other ) const
      {
        return realIterator_ != other.realIterator_;
      }
    };


    template< class ElementImp >
    class STLArray;


    template< class ElementImp >
    struct STLArrayTraits
    {
      typedef ElementImp ElementType;

      typedef STLArray< ElementType > ArrayType;

      typedef std :: vector< ElementType > stdVectorType;

      typedef STLArrayIterator< ElementType, typename stdVectorType :: iterator >
        IteratorType;
      typedef STLArrayIterator< const ElementType, typename stdVectorType :: const_iterator >
        ConstIteratorType;
    };



    template< class ElementImp >
    class STLArray
    : public ArrayInterface< STLArrayTraits< ElementImp > >
    {
    public:
      typedef ElementImp ElementType;

      typedef STLArrayTraits< ElementType > Traits;

    private:
      typedef STLArray< ElementType > ThisType;
      typedef ArrayInterface< Traits > BaseType;

      typedef typename Traits :: stdVectorType stdVectorType;

    public:
      typedef typename Traits :: IteratorType IteratorType;
      typedef typename Traits :: ConstIteratorType ConstIteratorType;

    protected:
      stdVectorType vector_;

    public:
      inline explicit STLArray ( unsigned int size = 0 )
      : vector_( size )
      {
      }

      inline STLArray ( unsigned int size,
                        const ElementType &element )
      : vector_( size, element )
      {
      }

      inline STLArray ( const ThisType &other )
      : vector_( other.vector_ )
      {
      }

      inline const ElementType &operator[] ( unsigned int index ) const
      {
        return vector_[ index ];
      }

      inline ElementType &operator[] ( unsigned int index )
      {
        return vector_[ index ];
      }

      //! fill the array with copies of an element
      inline ThisType &assign ( const ElementType &element )
      {
        vector_.assign( size(), element );
        return *this;
      }

      //! copy another array to this one
      template< class T >
      inline ThisType &assign( const ArrayInterface< T > &other )
      {
        const unsigned int size = other.size();
        resize( size );
        for( unsigned int i = 0; i < size; ++i )
          vector_[ i ] = other[ i ];
        return *this;
      }

      inline void append ( const ElementType &element )
      {
        vector_.push_back( element );
      }

      template< class T >
      inline void append ( const ArrayInterface< T > &array )
      {
        const unsigned int arraySize = array.size();
        for( unsigned int i = 0; i < arraySize; ++i )
          append( array[ i ] );
      }

      inline void resize ( unsigned int newSize )
      {
        vector_.resize( newSize );
      }

      //! obtain begin iterator
      inline ConstIteratorType begin () const
      {
        return ConstIteratorType( vector_.begin() );
      }

      //! obtain begin iterator
      inline IteratorType begin ()
      {
        return IteratorType( vector_.begin() );
      }

      //! obtain end iterator
      inline ConstIteratorType end () const
      {
        return ConstIteratorType( vector_.end() );
      }

      //! obtain end iterator
      inline IteratorType end ()
      {
        return IteratorType( vector_.end() );
      }

      inline unsigned int size () const
      {
        return vector_.size();
      }
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_STLARRAY_HH
