/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_TEST_EXACTSOLUTION_HH
#define DUNE_FEM_TEST_EXACTSOLUTION_HH

#include <complex>
// dune-fem includes
#include <dune/fem/function/common/function.hh>


namespace Dune
{

  namespace Fem
  {

    // ExactSolution
    // -------------

    template< class FunctionSpaceImp >
    class ExactSolution
    : public Fem::Function< FunctionSpaceImp, ExactSolution< FunctionSpaceImp > >
    {
      typedef ExactSolution< FunctionSpaceImp > ThisType;
      typedef Fem::Function< FunctionSpaceImp, ThisType > BaseType;

    public:
      typedef FunctionSpaceImp FunctionSpaceType;

      typedef typename FunctionSpaceType::DomainFieldType DomainFieldType;
      typedef typename FunctionSpaceType::RangeFieldType RangeFieldType;

      typedef typename FunctionSpaceType::DomainType DomainType;
      typedef typename FunctionSpaceType::RangeType RangeType;
      typedef typename FunctionSpaceType::JacobianRangeType JacobianRangeType;
      typedef typename FunctionSpaceType::HessianRangeType HessianRangeType;

    public:
      void evaluate( const DomainType &x, RangeType &phi ) const
      {
        phi = 1;
        for(int r = 0; r < RangeType :: dimension; ++r )
          for( int i = 0; i < DomainType :: dimension; ++i )
            phi[ r ] += pow(sin( M_PI * x[ i ] ),double(r+1));
#if defined USE_COMPLEX
        phi *= std::complex<double>( 1 , -2. );
#endif
      }

      void evaluate( const DomainType &x, RangeFieldType t, RangeType &phi ) const
      {
        evaluate( x, phi );
      }

      void jacobian( const DomainType &x, JacobianRangeType &Dphi ) const
      {
        Dphi = 1;
        for (int r = 0; r < RangeType :: dimension; ++r)
          for( int i = 0; i < DomainType :: dimension; ++i )
            for( int j = 0; j < DomainType :: dimension; ++j )
              Dphi[ r ][ j ] += double(r+1)*pow(sin( M_PI * x[ i ] ),double(r))*
                ((i != j) ? 0 : M_PI * cos( M_PI * x[ i ] ));
#if defined USE_COMPLEX
        Dphi *= std::complex<double>( 1 , -2. );
#endif
      }

      void jacobian( const DomainType &x, RangeFieldType t, JacobianRangeType &Dphi ) const
      {
        jacobian( x, Dphi );
      }

      void hessian( const DomainType &x, HessianRangeType &H ) const
      {
        for (int r = 0; r < RangeType :: dimension; ++r)
        {
          H[ r ] = RangeFieldType ( 0 );
          for( int i = 0; i < DomainType :: dimension; ++i )
          {
            const double s =  sin( M_PI * x[ i ] );
            H[ r ][ i ][ i ] = M_PI * M_PI * (r+1) *
              ( r * std::pow( s, double( r-1 ) )  - ( r+1 ) * std::pow( s, double( r+1 ) ) );
           }
        }
#if defined USE_COMPLEX
        H *= std::complex<double>( 1 , -2. );
#endif
      }

      void hessian( const DomainType &x, RangeFieldType t, HessianRangeType &H ) const
      {
        hessian( x, H );
      }
    };

  } // namespace Fem

} // namespace Dune

#endif // #ifndef DUNE_FEM_TEST_EXACTSOLUTION_HH
